﻿using System.Collections.Generic ;

namespace MCScheduling.ClairvoyantEDF
{
    /**
     * A priority queue implementation based on a binary heap
     * structure, which guarantees that all the operations define
     * has logarithmic time complexity.
     * 
     * @author Tomas Tunys
     */
    class PriorityQueue<T>
    {
        /** The binary heap structure represented by an array. */
        private List<T> queue ;

        /** The number of elements in the queue. */
        private int count ;

        /** The comparer used to determine the ordering of the elements. */
        private IComparer<T> comparer ;

        /**
         * Creates a new priority queue with the default comparer.
         */
        public PriorityQueue() : this(Comparer<T>.Default)
        {
        }

        /**
         * Creates a new priority queue with the specified comparer.
         * 
         * @param comparer    The comparer used to determine
         *                    the ordering of the elements 
         *                    within the queue.
         *                    
         */
        public PriorityQueue(IComparer<T> comparer)
        {
            this.comparer = comparer ;
            this.queue    = new List<T>() ;
        }

        /**
         * Swaps the content of the two elements specified 
         * by their position in the queue.
         * 
         * @param i The first element's position within the queue.
         * @param j The second element's position within the queue.
         */
        private void SwapQueueElements(int i, int j)
        {
           T t = queue[i] ; queue[i] = queue[j] ; queue[j] = t ;
        }

        /**
         * Fixes the heap upwards from the i-th element
         * in the heap.
         * 
         * @param The position of the element in the heap
         *        to start its reconstruction.
         */
        private void FixUp(int i)
        {
            int j = (i - 1) / 2 ;

            while (i > 0 && comparer.Compare(queue[i], queue[j]) < 0)
            {
                SwapQueueElements(i, j) ;

                i = j ;
                j = (i - 1) / 2 ;
            }
        }

        /**
         * Fixes the heap downwards from the i-th element
         * in the heap.
         * 
         * @param The position of the element in the heap
         *        to start its reconstruction.
         */
        private void FixDown(int i)
        {
            int j = 2 * i + 1 ;

            while (j < count)
            {
                if (j + 1 < count && comparer.Compare(queue[j], queue[j + 1]) > 0)
                {
                    ++j ;
                }

                if (comparer.Compare(queue[i], queue[j]) > 0)
                {
                    SwapQueueElements(i, j) ;
                }
                else
                {
                    break ;
                }

                i = j ;
                j = 2 * i + 1 ;
            }
        }

        /**
         * Inserts the specified element into the queue. This increases
         * the size of the queue by .
         * 
         * @param item The element to be inserted into the queue.
         */
        public void Add(T item)
        {
            /* To allow inflation of the queue array. */
            queue.Add(item) ;

            FixUp(count++) ;
        }

        /**
         * Removes the element with the minimum key value.
         */
        public void RemoveMinimum()
        {
            if (count != 0)
            {
                queue[0] = queue[--count] ;

                /* To allow deflation of the queue array. */
                queue.RemoveAt(count) ;

                FixDown(0) ;
            }
        }

        /**
         * Returns an element with the minimum key value.
         * 
         * @return An element with the minimum key value.
         * 
         * @throws InvalidOperationException When the queue is empty.
         */
        public T GetMinimum()
        {
            if (count == 0)
            {
                throw new System.InvalidOperationException("The priority queue is empty!") ;
            }

            return queue[0] ;
        }

        /**
         * Empties the queue.
         */
        public void Clear()
        {
            queue.Clear() ;
            count = 0 ;
        }

        /**
         * Returns the number of items in the queue.
         * 
         * @return The number of items in the queue.
         */
        public int Count
        {
            get { return count ; }
        }

        /**
         * Returns read-only list of the elements 
         * in the queue.
         * 
         * @return A read-only list of elements in the queue.
         */
        public System.Collections.ObjectModel.ReadOnlyCollection<T> Items()
        {
            return queue.AsReadOnly() ;
        }
    }
}
