﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MCScheduling.MixedCriticality ;

namespace MCScheduling.ClairvoyantEDF
{
    /**
     * The wrapper of mixed-criticality job to be used
     * with priority queues and criticality queue of the
     * clairvoyant EDF algorithm.
     * 
     * @author Tomas Tunys
     */
    public class Task
    {
        /** The mixed-criticality job this task represents. */
        private CMixedCriticalityJob job ;

        /** The release time of this task. */
        public int ReleaseTime { get ; set ; }

        /** The deadline of this task. */
        public int Deadline { get { return (int)job.getDeadline() ; } }

        /** The start time of this task. */
        public int StartTime { get { return (int)job.getStartTime() ; } set { job.setStartTime(value) ; } }

        /** The minimum start time of this task. */
        public int MinimumStartTime { get ; set ; }

        /** The maximum start time of this task. */
        public int MaximumStartTime { get ; set ; }
        
        /** The criticality of this task. */
        public int Criticality { get { return job.getCriticality() ; } }
        
        /** The execution times of this task for each critical level. */
        public ExecutionTimeArray ExecutionTime ;

        /**
         * The execution times of the task. The class is used for indexing
         * capability.
         * 
         * @author Tomas Tunys
         */
        public class ExecutionTimeArray
        {
            /** The mixed-criticality job having these WCETs. */
            private CMixedCriticalityJob job ;

            /**
             * Returns the jobs execution time at the specified
             * criticality level.
             * 
             * @return The jobs execution time at the specified
             * criticality level.
             */
            public int this[int criticality] { get { return (int)job.getExecutionTime(criticality) ; } }

            /**
             * Creates a new ExecutionTime instance for the specified
             * job.
             * 
             * @param job The job whose execution times this instance 
             *            represents.
             */
            public ExecutionTimeArray(CMixedCriticalityJob job)
            {
                this.job = job ;
            }
        }

        /**
         * Creates a new Task instance for the specified
         * mixed-criticality job.
         * 
         * @param job The mixed-criticality job that is represented
         *            by this instance.
         */
        public Task(CMixedCriticalityJob job)
        {
            this.job         = job ;
            ReleaseTime      = (int)job.getReleaseTime() ;
            MinimumStartTime = (int)job.getReleaseTime() ;
            MaximumStartTime = (int)(job.getDeadline() - job.getLargestExecutionTime()) ;
            ExecutionTime    = new ExecutionTimeArray(job) ;
        }    
    }
}