﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A base class for implementation of a crossover genetic operator.
     */
    public abstract class CBaseCrossoverOperator : CGeneticOperator, ICrossoverOperator
    {
        /** The crossover rate for this operator. */
        private double m_crossoverRate ;

        /**
         * Instantiates the base crossover operator.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this operator.
         * @param crossoverRate The crossover rate for this operator.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         * 
         * @throws ArgumentOutOfRangeException If crossoverRate is not in
         *                                    0.0 and 1.0 range.
         */
        protected CBaseCrossoverOperator(CConfiguration configuration, double crossoverRate) : base(configuration)
        {
            setCrossoverRate(crossoverRate) ;
        }

        /**
         * Returns the mutation rate used by this
         * operator.
         * 
         * @return The mutation rate.
         */
        public double getCrossoverRate()
        {
            return m_crossoverRate ;
        }

        /**
         * Sets the mutation rate for this operator. The crossoverRate
         * is expected to be decimal value between 0.0 and 1.0 inclusive.
         * 
         * @param crossoverRate The mutation rate to be set.
         * 
         * @throws ArgumentOutOfRangeException If crossoverRate is not in
         *                                    0.0 and 1.0 range.
         */
        public void setCrossoverRate(double crossoverRate)
        {
            if (crossoverRate < 0.0 || crossoverRate > 1.0)
            {
                throw new ArgumentOutOfRangeException("The crossover rate must be decimal value between 0 and 1!") ;
            }
            m_crossoverRate = crossoverRate ;
        }

        /**
         * Determines whether the crossover should be performed.
         * 
         * @return True if the crossover should be performed
         *         by this operator, otherwise, false.
         */
        public bool shouldPerformCrossover()
        {
            return (getConfiguration().getRandomGenerator().NextDouble() < m_crossoverRate) ;
        }

        /**
         * Crossovers the specified couple of chromosomes in a operator-specific way
         * forming a new chromosomes, who should be added to the specified population.
         * 
         * @param daddy      The first chromosome entering into the crossover.
         * @param mommy      The second chromosome entering into the crossover.
         * @param population The population for the next generation. It may
         *                   contain some chromosomes that have been already chosen
         *                   for the next generation.
         */
        public abstract void apply(IChromosome daddy, IChromosome mommy, CPopulation population) ;

#if DEBUG
        protected void printCrossoverDebugInfo(string crossoverInfo, IChromosome daddy, IChromosome mommy, IChromosome son, IChromosome daughter)
        {
            /* Prints the debug information. */
            Console.WriteLine(crossoverInfo) ;

            Console.WriteLine("Parents:") ;

            int numOfDigits = (int)Math.Ceiling(Math.Log10(daddy.getSize())) ;

            /* Writes out daddy to debug window. */
            printOutGenome(daddy, numOfDigits, true) ;

            /* Prints out mommy to debug window. */
            printOutGenome(mommy, numOfDigits, false) ;

            Console.WriteLine((daughter != null ? "Children:" : "Child:")) ;            

            /* Writes out son to debug window. */
            printOutGenome(son, numOfDigits, false) ;

            /* Prints out daughter to debug window. */
            if (daughter != null)
            {
                printOutGenome(daughter, numOfDigits, false) ;
            }
        }
#endif
    }
}
