﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A base class for implementation of a mutation genetic operator.
     */
    public abstract class CBaseMutationOperator : CGeneticOperator, IMutationOperator
    {
        /** The mutation rate for this mutation operator. */
        private double m_mutationRate ;

        /**
         * Instantiates the base mutation operator.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this operator.
         * @param mutationRate The mutation rate for this operator.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         * 
         * @throws ArgumentOutOfRangeException If mutationRate is not in
         *                                    0.0 and 1.0 range.
         */
        protected CBaseMutationOperator(CConfiguration configuration, double mutationRate) : base(configuration)
        {
            setMutationRate(mutationRate) ;
        }

        /**
         * Returns the mutation rate used by this
         * operator.
         * 
         * @return The mutation rate.
         */
        public double getMutationRate()
        {
            return m_mutationRate ;
        }

        /**
         * Sets the mutation rate for this operator. The mutationRate
         * is expected to be decimal value between 0.0 and 1.0 inclusive.
         * 
         * @param mutationRate The mutation rate to be set.
         * 
         * @throws ArgumentOutOfRangeException If mutationRate is not in
         *                                    0.0 and 1.0 range.
         */
        public void setMutationRate(double mutationRate)
        {
            if (mutationRate < 0.0 || mutationRate > 1.0)
            {
                throw new ArgumentOutOfRangeException("The mutation rate must be decimal value between 0 and 1!") ;
            }
            m_mutationRate = mutationRate ;
        }

        /**
         * Determines whether the mutation should be
         * performed.
         * 
         * @return True if the mutation should be performed
         *         by this operator, otherwise, false.
         */
        public bool shouldPerformMutation()
        {
            return (getConfiguration().getRandomGenerator().NextDouble() < m_mutationRate) ;
        }

        /**
         * Mutates the specified chromosome in a operator-specific way forming 
         * a new chromosome, which will then go to the next generation.
         * 
         * @param chromosome The chromosome to be mutated.
         */
        public abstract void apply(IChromosome chromosome) ;
    }
}
