﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * In case one wants the selection preasure to vary 
     * during the evolution he uses Bolzmann scaling. 
     * This scaling method works like this: First,
     * a temperature is set and a size of a step 
     * in which will be decreased each call, then
     * average value of e^(fitness/temperature) is
     * calculated, where e is the Euler's number and
     * fitness is a fitness score of an individual in
     * a population, afterwards, the fitness score
     * for each individual is assigned according to
     * newFitness = e^(oldFitness/temperature) / average.
     */
    class CBoltzmannScaling : IFitnessScaler, IConfigurationDependent
    {
        /** The actual temperature for scaling. */
        double m_temperature ;

        /** The starting temperature. */
        double m_startingTemperature ;

        /** The value the temperature cannot drop below. */
        double m_minimumTemperature ;

        /** The step in which the temperature is decreased. */
        double m_coolingStep ;

        /**
         * Initializes the boltzmann scaling method.
         * 
         * @param startTemperature   The starting temperature.
         * @param minimumTemperature The minimum value the temperature
         *                           may cool down to. It must be smaller
         *                           then start temperature.
         * @param coolingStep        The step in which the temperature 
         *                           falls down.
         *                            
         * @throws ArgumentException If the starting temperature is smaller
         *                           then the minimum temperature.
         */                   
        public CBoltzmannScaling(double startTemperature, double minimumTemperature, double coolingStep)
        {
            if (startTemperature < minimumTemperature)
            {
                throw new ArgumentException("The minimum temperature must be smaller then starting temperature!") ;
            }

            m_startingTemperature = startTemperature ;
            m_minimumTemperature  = minimumTemperature ;
            m_coolingStep         = coolingStep ;
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this scaler.
         */
        public void notify(CConfiguration configuration)
        {
            m_temperature = m_startingTemperature ;
        }

        /**
         * Returns the minimum temperature.
         * 
         * @return The minimum temperature.
         */
        public double getMinimumTemperature()
        {
            return m_minimumTemperature ;
        }

        /**
         * Returns the starting temperature.
         * 
         * @return The starting temperature.
         */
        public double getStartTemperature()
        {
            return m_temperature ;
        }

        /**
         * Returns the step size of the temperature
         * dropping.
         * 
         * @return The step size of temperature's
         *         decrease.
         */
        public double getCoolingStepSize()
        {
            return m_coolingStep ;
        }

        /**
         * Scales the fitness scores of the individuals 
         * within the specified population.
         * <p>
         * For more information look at the description 
         * of the CBoltzmannScaling class.
         * </p>
         * 
         * @param population The population of individuals
         *                   to be scaled.
         */
        public void scale(CPopulation population)
        {
            double boltzmannAverage = 0.0 ;

            for (int i = 0 ; i < population.getSize() ; i++)
            {
                boltzmannAverage += Math.Exp(population[i].getFitnessValue() / m_temperature) ;

                /* Temporarily saves e^(oldFitness/temperature) as new fitness. */
                population[i].setFitnessValue(boltzmannAverage) ;
            }

            /* The average e^(fitness / temperature) value. */
            boltzmannAverage /= population.getSize() ;

            /* Scales the fitness score of each individual within the population. */
            for (int i = 0 ; i < population.getSize() ; i++)
            {
                population[i].setFitnessValue(population[i].getFitnessValue() / boltzmannAverage) ;
            }

            /* Reduces the temperature a little each call to scale. */
            m_temperature -= m_coolingStep ;

            /* Ensures the temperature does not drop below minimum value. */
            if (m_temperature < m_minimumTemperature)
            {
                m_temperature = m_minimumTemperature ;
            }
        }
    }
}