﻿using System ;
using System.IO ;
using System.Linq ;
using System.Text ;
using System.Collections.Generic ;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The configuration represents a setting of a genetic algorithm,
     * i.e. what genetic operators it uses (selectors, crossover, 
     * mutation), what is the population size, chromosome size, and
     * mainly what fitness function evaluates chromosomes.
     */
    public class CConfiguration
    {
        /**
         * The user-specific name given to the configuration.
         */
        private String m_configurationName ;

        /**
         * The fitness evaluater that is used to determine the quality
         * of the chromosomes in the population created during
         * evolution.
         */
        private IFitnessEvaluator m_fitnessEvaluator ;

        /**
         * The fitness comparer that is used to determine the ordering
         * of the fitness values.
         */
        private IFitnessComparer m_fitnessComparer ;

        /**
         * The fitness scaler that may be used to scale the "raw"
         * fitness score values.
         */
        private IFitnessScaler m_fitnessScaler ;

        /**
         * The gene distance measurer that is used to determine
         * the distance between pair of genes.
         */
        IGeneDistanceMeasurer m_geneDistanceMeasurer ;

        /**
         * The number of chromosomes in the genotype's population.
         */
        private int m_populationSize ;

        /**
         * The size of a chromosome that will be used by a genetic
         * algorithm according to this configuraiton.
         */
        private int m_chromosomeSize ;

        /**
         * The maximum allowed difference of two decimals in which are
         * considered as being the same, i.e. x == y iff abs(x - y) is 
         * less or equal to this value.
         * <p>
         * The default value is 10e-6.
         * </p>
         */
        private double m_tolerance = 10e-6 ;

        /**
         * The number of fittest chromosomes to be always copied
         * into the mating pool.
         */
        private int m_numberOfFittestToCopy ;

        /**
         * The algorithm perfoming the evolution.
         */
        private CBaseBreeder m_breeder ;

        /**
         * The pseudo-random number generator that is used for the generation 
         * of any random numbers needed during the evolution processes.
         */
        private IRandomizer m_randomGenerator ;

        /**
         * The chromosome pool that may be used by the operators to acquire
         * chromosome instances.
         */
        private IChromosomePool m_chromosomePool ;

        /**
         * A natural selector that is used to pick an individual for reproduction.
         */
        private ISelectionOperator m_naturalSelector ;

        /**
         * A list of crossover operators that are used during the reproduction
         * phase of a genetic algoritm. The operators are executed in the order
         * they are stored in the list.
         */
        private List<ICrossoverOperator> m_crossoverOperators = new List<ICrossoverOperator>() ;

        /**
         * A list of mutation operators that are used during the reproduction
         * phase of a genetic algoritm. The operators are executed in the order
         * they are stored in the list.
         */
        private List<IMutationOperator> m_mutationOperators = new List<IMutationOperator>() ;

        /**
         * The list of genetic algorithm components that depend
         * on this configuration.
         */
        List<IConfigurationDependent> m_dependentComponents = new List<IConfigurationDependent>() ;

        #region >>> CConfiguration Constructors <<<

        /**
         * Creates a default configuration with the specified name.
         * 
         * @param configurationName The user-specific name for the configuration.
         */
        public CConfiguration(String configurationName) 
        {
            setName(configurationName) ;
            setBreeder(new CDefaultBreeder(this)) ;
            setRandomGenerator(new CMersenneTwister((ulong)System.DateTime.Now.Ticks));
            setFitnessComparer(new CGreaterFitterComparer(this)) ;
            setChromosomePool(new CChromosomePool()) ;
        }

        /**
         * Reads in the configuration from the specified file.
         *
         * @param configurationFilePath The file from which the configuration
         *                              will be created.
         * @param ignore                An ignored parameter.
         */
        public CConfiguration(String configurationFilePath, int ignore)
        {
            setBreeder(new CDefaultBreeder(this)) ;
            setRandomGenerator(new CMersenneTwister((ulong)System.DateTime.Now.Ticks));
            setFitnessComparer(new CGreaterFitterComparer(this)) ;
            setChromosomePool(new CChromosomePool()) ;

            deserialize(configurationFilePath) ;
        }

        #endregion >>> CConfiguration Constructors <<<

        #region >>> Configuration Name Property <<<

        /**
         * Sets the name for this configuration.
         * 
         * @param configurationName The user-specific name for the configuration.
         */
        public void setName(String configurationName)
        {
            m_configurationName = configurationName ;
        }

        /**
         * Returns the name of the configuration.
         * 
         * @return The user-specific name of the configuration.
         */
        public String getName()
        {
            return m_configurationName ;
        }

        #endregion >>> Configuration Name Property <<<

        #region >>> Fitness Evaluator Setting <<<

       /**
        * Sets the fitness evaluator that will be used by this genetic algorithm.
        * The fitness evaluator determines for each of the chromosome in the population
        * how good he is in solving the given problem, this value is so called
        * fitness.
        *
        * @param fitnessEvaluator The fitness evaluator to be set up.
        *
        * @throws ArgumentNullException If the fitnessEvaluator is null.
        */
        public void setFitnessEvaluator(IFitnessEvaluator fitnessEvaluator)
        {
            if (fitnessEvaluator == null) 
            {
                throw new ArgumentNullException("The fitness evaluator must not be null!") ;
            }
            else
            {
                if (!Object.ReferenceEquals(fitnessEvaluator, m_fitnessEvaluator))
                {
                    removeConfigurationDependentComponent(m_fitnessEvaluator) ;

                    m_fitnessEvaluator = fitnessEvaluator ;

                    addConfigurationDependentComponent(m_fitnessEvaluator) ;
                }
            }
        }

        /**
         * Returns the fitness comparer used in this configuration.
         *
         * @return The fitness comparer used in this configuration.
         */
        public IFitnessEvaluator getFitnessEvaluator()
        {
            return m_fitnessEvaluator ;
        }

        #endregion >>> Fitness Evaluator Setting <<<

        #region >>> Fitness Comparer Setting <<<

        /**
         * Sets the fitness comparer for this configuration. The fitness 
         * comparer determines the ordering of the fitness values.
         * 
         * @param fitnessComparer The fitness comparer to be set up.
         * 
         * @throws ArgumentNullException If fitnessComparer is null.
         */
        public void setFitnessComparer(IFitnessComparer fitnessComparer) 
        {
            if (fitnessComparer == null)
            {
                throw new ArgumentNullException("The fitness comparer must not be null!") ;
            }
            else
            {
                if (!Object.ReferenceEquals(fitnessComparer, m_fitnessComparer))
                {
                    removeConfigurationDependentComponent(m_fitnessComparer) ;

                    m_fitnessComparer = fitnessComparer ;

                    addConfigurationDependentComponent(m_fitnessComparer) ;
                }
            }
        }

        /**
         * Returns the fitness comparer that is used in this configuration.
         * 
         * @return the fitness comparer that is used in this configuration.
         */
        public IFitnessComparer getFitnessComparer()
        {
            return m_fitnessComparer ;
        }

        #endregion >>> Fitness Comparer Setting <<<

        #region >>> Fitness Scaler Setting <<<

        /**
         * Sets the fitness scaler for this configuration. The fitness 
         * scaler is used to scale raw fitness scores.
         * 
         * @param fitnessScaler The fitness scaler to be set up.
         * 
         * @throws ArgumentNullException If fitnessScaler is null.
         */
        public void setFitnessScaler(IFitnessScaler fitnessScaler) 
        {
            if (fitnessScaler == null)
            {
                throw new ArgumentNullException("The fitness scaler must not be null!") ;
            }
            else
            {
                if (!Object.ReferenceEquals(fitnessScaler, m_fitnessScaler))
                {
                    removeConfigurationDependentComponent(m_fitnessScaler) ;

                    m_fitnessScaler = fitnessScaler ;

                    addConfigurationDependentComponent(m_fitnessScaler) ;
                }
            }
        }

        /**
         * Returns the fitness scaler that is used in this configuration.
         * 
         * @return the fitness scaler that is used in this configuration.
         */
        public IFitnessScaler getFitnessScaler()
        {
            return m_fitnessScaler ;
        }

        #endregion >>> Fitness Scaler Setting <<<

        #region >>> Gene Distance Measurer Setting <<< 

        /**
         * Sets the gene distance measurer for this configuration.
         * The gene distance measurer is used to determine
         * the distance between pair of genes.
         * 
         * @param geneDistanceMeasurer The gene distance measurer to be set up.
         * 
         * @throws ArgumentNullException If gene distance measurer is null.
         */
        public void setGeneDistanceMeasurer(IGeneDistanceMeasurer geneDistanceMeasurer) 
        {
            if (geneDistanceMeasurer == null)
            {
                throw new ArgumentNullException("The gene distance measurer must not be null!") ;
            }
            else
            {
                if (!Object.ReferenceEquals(geneDistanceMeasurer, m_geneDistanceMeasurer))
                {
                    removeConfigurationDependentComponent(m_geneDistanceMeasurer) ;

                    m_geneDistanceMeasurer = geneDistanceMeasurer ;

                    addConfigurationDependentComponent(m_geneDistanceMeasurer) ;
                }
            }
        }

        /**
         * Returns the gene distance measurer that is used in this configuration.
         * 
         * @return the gene distance measurer that is used in this configuration.
         */
        public IGeneDistanceMeasurer getGeneDistanceMeasurer()
        {
            return m_geneDistanceMeasurer ;
        }

        #endregion >>> Gene Distance Measurer Setting <<< 

        #region >>> Breeder Setting <<<

        /**
         * Sets the breeder to be used in this configuration.
         * 
         * @param breeder The breeder to be set up for this configuration.
         *
         * @throws ArgumentNullException If breeder is null.
         */
        public void setBreeder(CBaseBreeder breeder) 
        {
            if (breeder == null)
            {
                throw new ArgumentNullException("The breeder must not be null!") ;
            }
            else
            {
                if (!Object.ReferenceEquals(breeder, m_breeder))
                {
                    removeConfigurationDependentComponent(m_breeder) ;

                    m_breeder = breeder ;

                    addConfigurationDependentComponent(m_breeder) ;
                }
            }
        }

        /**
         * Returns the breeder that is used by this configuration.
         * 
         * @return The breeder that is used by this configuration.
         */
        public CBaseBreeder getBreeder() 
        {
            return m_breeder ;
        }

        #endregion >>> Breeder Setting <<<

        #region >>> Selection Operator Setting <<<

        /**
         * Sets the natural selector to be used for this genetic algorithm.
         * For more information about natural selectors see description of
         * the INaturalSelector public interface.
         *
         * @param naturalSelector The natural selector to be set up.
         *
         * @throws ArgumentNullException If the naturalSelector is null.
         */
        public void setNaturalSelector(ISelectionOperator naturalSelector)
        {
            if (naturalSelector == null)
            {
                throw new ArgumentNullException("The natural selector must not be null!") ;
            }
            else
            {
                m_naturalSelector = naturalSelector ;

                addConfigurationDependentComponent(naturalSelector) ;
            }
        }

        /**
        * Returns the natural selector that is used in this configuration.
        *
        * @return The natural selector used in thie configuration.
        */
        public ISelectionOperator getNaturalSelector()
        {
            return m_naturalSelector ;
        }

        #endregion >>> Selection Operator Setting <<<

        #region >>> Random Generator Setting <<<

        /**
         * Sets the pseudo-random number generator that will be used 
         * by this algorithm.
         *
         * @param randomGenerator The random number generator to be used.
         *
         * @throws ArgumentNullException If randomGenerator is null.
         */
        public void setRandomGenerator(IRandomizer randomGenerator)
        {
            if (randomGenerator == null)
            {
                throw new ArgumentNullException("The RandomGenerator instance may not be null.") ;
            }
            else
            {
                m_randomGenerator = randomGenerator ;
            }
        }

        /**
         * Returns the pseudo-random number generator that is used 
         * by this configuration.
         *
         * @return The random number generator.
         */
        public IRandomizer getRandomGenerator()
        {
            return m_randomGenerator ;
        }

        #endregion >>> Random Generator Setting <<<

        #region >>> Tolerance Property <<<

        /**
         * Returns the maximum allowed difference of two decimals
         * in which are considered as being the same.
         * 
         * @return A tolerance.
         */
        public double getTolerance()
        {
            return m_tolerance ;
        }

        /**
         * Sets the maximum allowed difference of two decimals
         * in which are considered as being the same.
         * 
         * @param tolerance A new tolerance value.
         */
        public void setTolerance(double tolerance)
        {
            m_tolerance = tolerance ;
        }

        #endregion >>> Tolerance Property <<<

        #region >>> Elitism - Number of Fittest To Copy - Property <<<

        /**
         * Sets the number of fittest chromosomes that will be
         * always copied into the mating pool.
         * 
         * @param numberOfFittestToCopy The number of chromosomes
         *                              to be copied.
         */
        public void setNumberOfFittestToCopy(int numberOfFittestToCopy)
        {
            m_numberOfFittestToCopy = numberOfFittestToCopy ;
        }

        /**
         * Returns the number of fittest chromosomes that is 
         * being always copied into the mating pool.
         * 
         * @return A number of chromosomes being copied.
         */
        public int getNumberOfFittestToCopy()
        {
            return m_numberOfFittestToCopy ;
        }

        #endregion >>> Elitism - Number of Fittest To Copy - Property <<<

        #region >>> Crossover Operator Manipulation Methods <<<

        /**
         * Adds a crossover operator that is used by this genetic algorithm. 
         * These operators are used in the reproduction phase of the evolution
         * process; in the order in which they were added by this method.
         *
         * @param crossoverOperator The crossover operator to be added.
         *
         * @throws ArgumentNullException If the crossoverOperator is null.
         */
        public void addCrossoverOperator(ICrossoverOperator crossoverOperator)
        {
            if (crossoverOperator == null) 
            {
                throw new ArgumentNullException("The crossover operator must not be null!") ;
            }
            else
            {
                m_crossoverOperators.Add(crossoverOperator) ;

                addConfigurationDependentComponent(crossoverOperator) ;
            }
        }

        /**
         * Removes the specified crossover operator from the list of genetic
         * operators used by this genetic algorithm. If the operator is not
         * present in the list, nothing happens.
         *
         * @param crossoverOperator The crossover operator to be removed.
         *
         * @throws ArgumentNullException If the crossoverOperator is null.
         */
        public void removeCrossoverOperator(ICrossoverOperator crossoverOperator)
        {
            if (crossoverOperator == null) 
            {
                throw new ArgumentNullException("The crossover operator must not be null!") ;
            }
            else
            {
                m_crossoverOperators.Remove(crossoverOperator) ;

                removeConfigurationDependentComponent(crossoverOperator) ;
            }
        }

        /**
         * Removes all the crossover operators registred with this configuration.
         */
        public void removeAllCrossoverOperators()
        {
            foreach (ICrossoverOperator crossoverOperator in m_crossoverOperators)
            {
                removeConfigurationDependentComponent(crossoverOperator) ;
            }

            m_crossoverOperators.Clear() ;
        }

        /**
         * Returns the crossover operators setup in this configuration.
         * The returned list should NOT be modified.
         *
         * @return The list of crossover operators.
         */
        public List<ICrossoverOperator> getCrossoverOperators() 
        {
            return m_crossoverOperators ;
        }

        #endregion >>> Crossover Operator Manipulation Methods <<<

        #region >>> Mutation Operator Manipulation Methods <<<

        /**
         * Adds a mutation operator that is used by this genetic algorithm. 
         * These operators are used in the reproduction phase of the evolution
         * process; in the order in which they were added by this method.
         *
         * @param mutationOperator The crossover operator to be added.
         *
         * @throws ArgumentNullException If the mutationOperator is null.
         */
        public void addMutationOperator(IMutationOperator mutationOperator)
        {
            if (mutationOperator == null) 
            {
                throw new ArgumentNullException("The mutation operator must not be null!") ;
            }
            else
            {
                m_mutationOperators.Add(mutationOperator) ;

                addConfigurationDependentComponent(mutationOperator) ;
            }
        }

        /**
         * Removes the specified mutation operator from the list of genetic
         * operators used by this genetic algorithm. If the operator is not
         * present in the list, nothing happens.
         *
         * @param mutationOperator The mutation operator to be removed.
         *
         * @throws ArgumentNullException If the mutationOperator is null.
         */
        public void removeMutationOperator(IMutationOperator mutationOperator)
        {
            if (mutationOperator == null) 
            {
                throw new ArgumentNullException("The mutation operator must not be null!") ;
            }
            else
            {
                m_mutationOperators.Remove(mutationOperator) ;

                removeConfigurationDependentComponent(mutationOperator) ;
            }
        }

        /**
         * Removes all the mutation operators registred with this configuration.
         */
        public void removeAllMutationOperators()
        {
            foreach (IMutationOperator mutationOperator in m_mutationOperators)
            {
                removeConfigurationDependentComponent(mutationOperator) ;
            }

            m_mutationOperators.Clear() ;
        }

        /**
         * Returns the mutation operators setup in this configuration.
         * The returned list should NOT be modified.
         *
         * @return The list of mutation operators.
         */
        public List<IMutationOperator> getMutationOperators() 
        {
            return m_mutationOperators ;
        }

        #endregion >>> Mutation Operator Manipulation Methods <<<

        #region >>> Dependent Components Manipulation Methods <<<

        /**
         * Adds the specified component to the list of dependent components
         * of this configuration.
         * 
         * @param component The dependent component to add.
         */
        private void addConfigurationDependentComponent(object component)
        {
            if (component != null && component is IConfigurationDependent)
            {
                m_dependentComponents.Add((IConfigurationDependent)component) ;
            }
        }

        /**
         * Removes the specified component from the list of dependent components
         * of this configuration.
         * 
         * @param component The dependent component to remove.
         */
        private void removeConfigurationDependentComponent(object component)
        {
            if (component != null && component is IConfigurationDependent)
            {
                m_dependentComponents.Remove((IConfigurationDependent)component) ;
            }
        }

        /**
         * Notifies the dependent components registered with this configuration
         * before the genetic algorithm that uses them is executed.
         */
        public void notifyConfigurationDependentComponents()
        {
            foreach (IConfigurationDependent component in m_dependentComponents)
            {
                component.notify(this) ;
            }
        }

        #endregion >>> Dependent Components Manipulation Methods <<<

        #region >>> Population Size Property <<<

        /**
         * Sets up the population size that will be used by this 
         * genetic algorithm. The population size represents 
         * the number of chromosomes contained within a genotype.
         * 
         * @param populationSize The population size value to be set.
         *
         * @throws ArgumentOutOfRangeException If the population size is
         *                                     less then 1.
         */
        public void setPopulationSize(int populationSize)
        {
            if (populationSize < 1)
            {
                throw new ArgumentOutOfRangeException("The population size must be > 0!") ;
            }
            else
            {
                m_populationSize = populationSize ;
            }
        }

        /**
         * Returns the population size.
         *
         * @return The size of the population.
         */
        public int getPopulationSize()
        {
            return m_populationSize ;
        }

        #endregion >>> Population Size Property <<<

        #region >>> Chromosome Size Property <<<

        /**
         * Sets the size of a chromosome that will be used by this 
         * genetic algorithm.
         * 
         * @param chromosomeSize A chromosome size.
         */
        public void setChromosomeSize(int chromosomeSize)
        {
            m_chromosomeSize = chromosomeSize ;
        }

        /**
         * Returns the size of a chromosome that will be used by this
         * genetic algorithm.
         *
         * @return The chromosome size used in this configuration.
         */
        public int getChromosomeSize()
        {
            return m_chromosomeSize ;
        }

        #endregion >>> Chromosome Size Property <<<

        #region >>> The Chromosome Pool Setting <<<

        /**
         * Sets the chromosome pool that will be used by this 
         * configuration.
         * 
         * @param chromosomePool The chromosome pool to be used.
         */
        public void setChromosomePool(IChromosomePool chromosomePool)
        {
            m_chromosomePool = chromosomePool ;
        }

        /**
         * Returns the chromosome pool used by this configuration.
         * 
         * @return The chromosome pool used by this configuration.
         */
        public IChromosomePool getChromosomePool()
        {
            return m_chromosomePool ;
        }

        #endregion >>> The Chromosome Pool Setting <<<

        #region >>> Serialization Methods <<<

        /**
         * Serializes this configuration into the specified
         * file.
         * 
         * @param filePath The path to a file where this configuration
         *                 is going to be serialized.
         */ 
        public void serialize(string filePath)
        {
            using (StreamWriter fileWriter = new StreamWriter(filePath))
            {
                try
                {
                    fileWriter.Write(getName()) ;
                    fileWriter.WriteLine() ;
                    fileWriter.Write(getPopulationSize()) ;
                    fileWriter.WriteLine() ;
                    fileWriter.Write(getNumberOfFittestToCopy()) ;
                    fileWriter.WriteLine() ;
                    fileWriter.Write(CFitnessEvaluatorSerializer.serialize(getFitnessEvaluator())) ;
                    fileWriter.WriteLine() ;
                    fileWriter.Write(CFitnessScalerSerializer.serialize(getFitnessScaler())) ;
                    fileWriter.WriteLine() ;
                    fileWriter.Write(CSelectorOperatorSerializer.serialize(getNaturalSelector())) ;
                    fileWriter.WriteLine() ;
            
                    foreach (ICrossoverOperator crossoverOperator in getCrossoverOperators())
                    {
                        fileWriter.Write(CCrossoverOperatorSerializer.serialize(crossoverOperator)) ;
                        fileWriter.WriteLine() ;
                    }

                    foreach (IMutationOperator mutationOperator in getMutationOperators())
                    {
                        fileWriter.Write(CMutationOperatorSerializer.serialize(mutationOperator)) ;
                        fileWriter.WriteLine() ;
                    }
                }
                catch (Exception exception)
                {
                    fileWriter.WriteLine("Error: " + exception.Message) ;
                }
            }
        }

        /**
         * Deserializes the configuration in the specified 
         * file and sets the fields of this configuration
         * according to it.
         * 
         * @param filePath The path where the configuration 
         *                 is going to be deserialized from.
         */
        public void deserialize(string filePath)
        {
            using (StreamReader fileReader = new StreamReader(filePath))
            {
                setName(fileReader.ReadLine().Trim()) ;
                setPopulationSize(int.Parse(fileReader.ReadLine().Trim())) ;
                setNumberOfFittestToCopy(int.Parse(fileReader.ReadLine().Trim())) ;
                setFitnessEvaluator(CFitnessEvaluatorSerializer.deserialize(fileReader.ReadLine().Trim())) ;
                setFitnessScaler(CFitnessScalerSerializer.deserialize(fileReader.ReadLine().Trim())) ;
                setNaturalSelector(CSelectorOperatorSerializer.deserialize(fileReader.ReadLine().Trim(), this)) ;

                removeAllCrossoverOperators() ;

                string line = null ;

                try
                {
                    /* Technically does not read till the end of file, because it should be
                     * break-ed by an exception once a mutation operator is read.
                     */
                    while (!fileReader.EndOfStream)
                    {
                        line = fileReader.ReadLine().Trim() ;
                        addCrossoverOperator(CCrossoverOperatorSerializer.deserialize(line, this)) ;
                    }
                }
                catch (ArgumentException)
                {
                    if (getCrossoverOperators().Count == 0)
                    {
                        throw new ArgumentException("The configuration file does not contain any crossover operators!") ;
                    }
                }

                removeAllMutationOperators() ;

                try
                {
                    addMutationOperator(CMutationOperatorSerializer.deserialize(line, this)) ;

                    while (!fileReader.EndOfStream)
                    {
                        line = fileReader.ReadLine().Trim() ;
                        addMutationOperator(CMutationOperatorSerializer.deserialize(line, this)) ;   
                    }
                }
                catch (ArgumentException exception)
                {
                    if (getMutationOperators().Count == 0)
                    {
                        throw new ArgumentException("The configuration file does not contain any mutation operators!") ;
                    }
                    else
                    {
                        throw exception ;
                    }
                }
            }
        }

        #endregion >>> Serialization Methods <<<

        #region >>> The Default Fitness Comparer Definition <<<

        /**
         * The default fitness comparer. The greater the fitness score
         * of an individual the fitter its status.
         */
        public class CGreaterFitterComparer : IFitnessComparer
        {
            /** The configuration that uses this comparer. */
            private CConfiguration m_configuration ;

            /**
             * Creates the default fitness comparer. The configuration
             * is used for its <em>tolerance</em> property.
             * 
             * @param configuration The configuration of the genetic algorithm 
             *                      that uses this comparer.
             */
            public CGreaterFitterComparer(CConfiguration configuration)
            {
                m_configuration = configuration ;
            }

            /**
             * Compares the two specified individuals and returns -1, 
             * if y is fitter; 1 if x is fitter, and 0 otherway. The
             * fitter individual is that with greater fitness score.
             * 
             * @param x The first individual to be compared.
             * @param y The second individual to be compared.
             * 
             * @return -1, if y is fitter; 1 if x is fitter, and 0 otherway.
             */
            public int Compare(IChromosome x, IChromosome y)
            {
                double d = (x.getFitnessValue() - y.getFitnessValue()) ;

                if (d < -m_configuration.getTolerance())
                {
                    return -1 ;
                }
                else if (d > m_configuration.getTolerance())
                {
                    return 1 ;
                }
                else 
                {
                    return 0 ;
                }                
            }
        }

        #endregion >>> The Default Fitness Comparer Definition <<<
    }
}