﻿using System ;
using System.Linq ;
using System.Text ;
using System.Globalization ;
using System.Collections.Generic ;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A serializer for crossover operators.
     */
    class CCrossoverOperatorSerializer
    {
        /**
         * Returns the System.String representing the serialization
         * of the specified crossover operator.
         * 
         * @param crossover The operator to be serialized.
         * 
         * @return The string representing the serialization of the 
         *         specified crossover operator.
         * 
         * @throws ArgumentException If crossover is null or unknown
         *                           by this serializer.
         */
        public static string serialize(ICrossoverOperator crossover)
        {
            if (crossover is COrderCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "OX {0:0.00} {1:0.00}", ((COrderCrossover)crossover).getCrossoverRate(), ((COrderCrossover)crossover).getCrossoverRatio()) ;
            }
            else if (crossover is COrderBasedCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "OBX {0:0.00} {1:0.00}", ((COrderBasedCrossover)crossover).getCrossoverRate(), ((COrderBasedCrossover)crossover).getCrossoverRatio()) ;
            }
            else if (crossover is CPositionBasedCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "PBX {0:0.00} {1:0.00}", ((CPositionBasedCrossover)crossover).getCrossoverRate(), ((CPositionBasedCrossover)crossover).getCrossoverRatio()) ;
            }
            else if (crossover is CCycleCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "CX {0:0.00}", ((CCycleCrossover)crossover).getCrossoverRate()) ;
            }
            else if (crossover is CPartiallyMappedCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "PMX {0:0.00}", ((CPartiallyMappedCrossover)crossover).getCrossoverRate()) ;
            }
            else if (crossover is CAlternatingPositionCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "APX {0:0.00}", ((CAlternatingPositionCrossover)crossover).getCrossoverRate()) ;
            }
            else if (crossover is CHeuristicCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "HX {0:0.00}", ((CHeuristicCrossover)crossover).getCrossoverRate()) ;
            }
            else if (crossover is CGreedyCrossover)
            {
                return String.Format(CultureInfo.InvariantCulture, "GX {0:0.00}", ((CGreedyCrossover)crossover).getCrossoverRate()) ;
            }
            else
            {
                throw new ArgumentException("An unknown (or null) crossover operator!") ;
            }
        }

        /**
         * Creates a crossover operator from the specified string.
         * 
         * @param serialization The operator's serialization to be
         *                      deserialized.
         * @param configuration The configuration of the genetic
         *                      algorithm that will use the crossover
         *                      operator.   
         *                      
         * @return The deserialized crossover operator.
         *                      
         * @throws ArgumentException If the serialization is invalid,
         *                           or represents no crossover operator.
         */
        public static ICrossoverOperator deserialize(string serialization, CConfiguration configuration)
        {
            String[] parameters = serialization.Split(' ', '\t') ;

            if (serialization.StartsWith("OX", true, null))
            {
                if (parameters.Length != 3)
                {
                    throw new ArgumentException("Invalid, or missing arguments for OX operator!") ;
                }

                return new COrderCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture), double.Parse(parameters[2], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("OBX", true, null))
            {
                if (parameters.Length != 3)
                {
                    throw new ArgumentException("Invalid, or missing arguments for OBX operator!") ;
                }

                return new COrderBasedCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture), double.Parse(parameters[2], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("PBX", true, null))
            {
                if (parameters.Length != 3)
                {
                    throw new ArgumentException("Invalid, or missing arguments for PBX operator!") ;
                }

                return new CPositionBasedCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture), double.Parse(parameters[2], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("CX", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for CX operator!") ;
                }

                return new CCycleCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("PMX", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for PMX operator!") ;
                }

                return new CPartiallyMappedCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("APX", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for APX operator!") ;
                }

                return new CAlternatingPositionCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("HX", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for HX operator!") ;
                }

                return new CHeuristicCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else if (serialization.StartsWith("GX", true, null))
            {
                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing argument for GX operator!") ;
                }

                return new CGreedyCrossover(configuration, double.Parse(parameters[1], CultureInfo.InvariantCulture)) ;
            }
            else
            {
                throw new ArgumentException("Unknown selector operator!") ;
            }
        }
    }
}
