﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The deterministic sampling selection operator implementation.
     * 
     * @author Tomas Tunys
     */
    public class CDeterministicSampling : CBaseSelectionOperator, IConfigurationDependent
    {
        /** The fraction of a chromosome to be added to the population. */
        private double[] m_chromosomeFractions ;

        /** The chromosome indices used for sorting individuals. */
        private int[] m_chromosomeIndices ;

        /** The fitness comparer determining the ordering of chromosome fractions. */
        private ChromosomeFractionsComparer m_comparer ;

        /**
         * A comparer used by deterministic sampling selector for sorting
         * chromosomes according to the decimal fractions of the expected
         * chromosome's appearances in the mating pool.
         */
        private class ChromosomeFractionsComparer : Comparer<int>
        {
            /**
             * The deterministic sampling selector using this comparer.
             */
            private CDeterministicSampling m_selector ;

            /**
             * The maximum allowed difference of two decimal considered 
             * as being the same.
             */
            private double m_tolerance ;

            /**
             * Creates a new instance of the gene remainder comparer.
             * 
             * @param An instance of DeterministicSampling selector
             *        that uses this comparer.
             */
            public ChromosomeFractionsComparer(CDeterministicSampling selector)
            {
                m_selector  = selector ;
                m_tolerance = selector.getConfiguration().getTolerance() ; 
            }

            /**
             * Compares two decimal fractions of the number of expected 
             * samples of chromosome x and y in the mating pool. It
             * sorts the indices (of chromosomes) in the descending
             * order.
             * 
             * @param x The index of the first chromosome to be compared.
             * @param y The index of the second chromosome to be compared.
             * 
             * @return -1 if the x's fraction is bigger then y's, 1 if y's
             *         fraction is bigger then x's, and 0 the other way.
             */
            public override int Compare(int x, int y)
            {
                double difference = (m_selector.m_chromosomeFractions[x] - m_selector.m_chromosomeFractions[y]) ;

                if (difference < -m_tolerance)
                {
                    return 1 ;
                }
                else if (difference > m_tolerance)
                {
                    return -1 ;
                }
                else
                {
                    return 0 ;
                }
            }
        }

        /**
         * Instantiates a new deterministic sampling selector.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this selector.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */
        public CDeterministicSampling(CConfiguration configuration) : base(configuration)
        {
            m_comparer = new ChromosomeFractionsComparer(this) ;
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this selector.
         */
        public void notify(CConfiguration configuration)
        {
            m_chromosomeFractions = new double[configuration.getPopulationSize()] ;
            m_chromosomeIndices   = new int[configuration.getPopulationSize()] ;
        }

        /**
         * Selects a chromosome from the specified population according to its
         * fitness. See the description of  the algorithm in the description
         * of this class.
         * 
         * @param population    The population the chromosome will be selected from.
         * @param matingPool    The mating pool the chromosome will be inserted into.
         * @param toSelectCount The total number of chromosomes to select.
         */
        public override void select(CPopulation population, CPopulation matingPool, int toSelectCount)
        {
            enforcePositiveFitnessScores(population) ;

            double totalFitnessScore = 0.0 ;

            for (int i = 0 ; i < population.getSize() ; i++)
            {
                totalFitnessScore += population[i].getFitnessValue() ;
            }

            for (int i = 0 ; i < population.getSize() && toSelectCount > 0 ; i++)
            {
                m_chromosomeIndices[i] = i ;

                /* Determines how many copies of this individual should go to the mating pool...*/
                m_chromosomeFractions[i] = population[i].getFitnessValue() / totalFitnessScore * toSelectCount ;

                /* Fractions of individuals are (likely) processed later. */
                int chromosomeSamplesCount = (int)Math.Truncate(m_chromosomeFractions[i]) ;

                for (int j = 0 ; j < chromosomeSamplesCount && toSelectCount > 0 ; j++)
                {
                    matingPool.addChromosome(population[i]) ;

                    --toSelectCount ;
                }

                m_chromosomeFractions[i] -= chromosomeSamplesCount ;
            }

            /* If there is still room for an individual in the mating pool. */
            if (toSelectCount > 0)
            {
                Array.Sort(m_chromosomeIndices, m_comparer) ;

                int currentChromosome = 0 ;

                while (toSelectCount > 0)
                {
                    matingPool.addChromosome(population[m_chromosomeIndices[currentChromosome]]) ;

                    if (++currentChromosome == m_chromosomeIndices.Length)
                    {
                        currentChromosome = 0 ;
                    }

                    --toSelectCount ;
                }
            }
        }
    }
}
