﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_DIVM

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The displaced inversion mutation operator. It randomly chooses two points within
     * a chromosome with uniform probability, reverses the genes within  the range defined 
     * by these two points, and finally displaces the reversed part and reinserts it back
     * into a random position. This operator may be considered as combination of inversion
     * and displacement mutation operators.
     * 
     * @author Tomas Tunys
     */
    public class CDisplacedInversionMutation : CBaseMutationOperator, IConfigurationDependent
    {
        /** Temporary memory for displaced genes. */
        private IGene[] m_genes ;

        /** The minimum span between chosen points. */
        private int minSpan ;

        /** The maximum span between chosen points. */
        private int maxSpan ;

        /**
         * Creates new displaced inversion mutation operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param mutationRate  The mutation rate.
         */
        public CDisplacedInversionMutation(CConfiguration configuration, double mutationRate) : base(configuration, mutationRate)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            maxSpan = configuration.getChromosomeSize() ;
            minSpan = Math.Min(3, maxSpan) ;
            m_genes = new IGene[maxSpan] ;
        }

        /**
         * Mutates the specified chromosome in a operator-specific way forming 
         * a new chromosome, which will then go to the next generation.
         * 
         * @param chromosome The chromosome to be mutated.
         */
        public override void apply(IChromosome chromosome)
        {
            if (shouldPerformMutation())
            {
                IRandomizer randomizer = getConfiguration().getRandomGenerator() ;

                int start, span  ;

                /* Chooses a section within the chromosome. */
                getChromosomeSection(out start, out span, minSpan, maxSpan) ;

                int ipoint = randomizer.Next(chromosome.getSize() - span + 1) ;

                /* Saves the genes within the chosen section. */
                for (int i = 0 ; i < span ; i++)
                {
                    m_genes[i] = chromosome.getGene(start + i) ;
                }

#if DEBUG_DIVM
                Console.WriteLine("Displaced Inversion Mutation: " + (start + 1) + " - " + (start + span) + " -> " + (ipoint + 1)) ;
                printOutGenome(chromosome, (int)Math.Ceiling(Math.Log10(chromosome.getSize())), true) ;
#endif
                /* Shift the genes to make place for reinsertion of the chosen genes. */
                if (ipoint < start)
                {
                    int toCopy = start - ipoint ;

                    int end  = start + span - 1 ;
                    int iend = ipoint + toCopy - 1 ;

                    for (int i = 0 ; i < toCopy ; i++)
                    {
                        chromosome.setGene(chromosome.getGene(iend - i), end - i) ;
                    }
                }
                else
                {
                    int toCopy = ipoint - start ;
                    int end    = start + span ;

                    for (int i = 0 ; i < toCopy ; i++)
                    {
                        chromosome.setGene(chromosome.getGene(end + i), start + i) ;
                    }
                }

                /* Reinserts the genes into the chromosome. */
                for (int i = 0 ; i < span ; i++)
                {
                    chromosome.setGene(m_genes[span - i - 1], ipoint + i) ;
                }
#if DEBUG_DIVM
                printOutGenome(chromosome, (int)Math.Ceiling(Math.Log10(chromosome.getSize())), false) ;
#endif
            }
        }
    }
}
