﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The base public class for implementation of a genetic operator.
     *
     * @author Tomas Tunys
     */
    public abstract class CGeneticOperator
    {
        /**
         * A configuration of a genetic algorithm that uses this 
         * crossover operator. 
         */
        private CConfiguration m_configuration ;

        /**
         * Creates and initializes the base crossover operator.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this operator.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */
        public CGeneticOperator(CConfiguration configuration)
        {
            if (configuration == null) 
            {
                throw new NullReferenceException("The configuration must not be null!") ;
            }
            else
            {
                setConfiguration(configuration) ;
            }
        }

        /**
         * Sets the configuration of a genetic algorithm
         * that uses this operator.
         * 
         * @param configuration The configuration of a genetic
         *                      algorithm that uses this operator.
         */
        public void setConfiguration(CConfiguration configuration)
        {
            m_configuration = configuration ;
        }

        /**
         * Returns a configuration of a genetic algorithm
         * that uses this operator.
         * 
         * @return A configuration of a genetic algorithm
         *         that uses this operator.
         */
        public CConfiguration getConfiguration()
        {
            return m_configuration ;
        }

        /**
         * Performes a swap of the i-th gene in the first specified 
         * chromosome with the j-th gene in the second.
         * 
         * @param x The first chromosome whose gene will be swapped.
         * @param y The second chromosome whose gene will be swapped.
         * @param i The position of the gene in the first chromosome
         *          that will be swapped.
         * @param j The position of the gene in the second chromosome
         *          that will be swapped.
         */
        public static void swapGenes(IChromosome x, IChromosome y, int i, int j)
        {
            IGene z = x.getGene(i) ; x.setGene(y.getGene(j), i) ; y.setGene(z, j) ;
        }

        /**
         * Performes a swap of the i-th gene with the j-th gene 
         * in the specified chromosome.
         * 
         * @param x The chromosome whose genes will be swapped.
         * @param i The position of the first gene that will be swapped.
         * @param j The position of the second gene that will be swapped.
         */
        public static void swapGenes(IChromosome x, int i, int j)
        {
            if (i != j)
            {
                swapGenes(x, x, i, j) ;
            }
        }

        /**
         * Swaps the contents of the specified variables, i.e. assigns
         * the value of x to y, and vice versa.
         * 
         * @param x The first variable whose content will be swapped.
         * @param y The second variable whose content will be swapped.
         */
        public static void swap<T>(ref T x, ref T y)
        {
            T z = x ; x = y ; y = z ;
        }

        /**
         * Determines a random start point and span to a section given 
         * a minimum and maximum span size.
         * 
         * @param start   The section's start point.
         * @param span    The section's span.
         * @param minSpan The minimum span of the created section.
         * @param maxSpan The maximum span of the created section.
         */
        protected void getChromosomeSection(out int start, out int span, int minSpan, int maxSpan)
        {
            span  = getConfiguration().getRandomGenerator().Next(minSpan, maxSpan + 1) ;
            start = getConfiguration().getRandomGenerator().Next(maxSpan - span + 1) ;
        }

#if DEBUG
        /**
         * Debugging method. It prints out the specified permutation chromosome.
         * 
         * @param chromosome  The chromosome to be print out.
         * @param numOfDigits The max. number of digits of the header column.
         * @param printHeader If set to true a header will be printed.
         */
        protected void printOutGenome(IChromosome chromosome, int numOfDigits, bool printHeader)
        {
            if (printHeader)
            {
                Console.Write(" ") ;
                for (int i = 1 ; i <= chromosome.getSize() ; i++)
                {
                    Console.Write(" ") ;
                    Console.Write(String.Format("{0:D}", i).PadLeft(numOfDigits)) ;
                }
                Console.WriteLine() ;
            }

            Console.Write(" ") ;
            for (int i = 0 ; i < chromosome.getSize() ; i++)
            {
                Console.Write(" ") ;
                Console.Write(String.Format("{0:D}", chromosome.getGene(i).getAllele()).PadLeft(numOfDigits)) ;
            }
            Console.WriteLine() ;
        }
#endif
    }
}
