﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_HX

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The heuristic crossover operator implementation.
     *
     * @author Tomas Tunys.
     */
    public class CHeuristicCrossover : CBaseCrossoverOperator, IConfigurationDependent
    {
        /** The distances list of the neighbouring genes. */
        private double[,] m_geneDistances ;

        /** The total distance of the genes from the neighbour list of the gene with given allele value. */
        private double[] m_totalDistances ;

        /** 
         * The list of neighbouring genes sorted according to the distance from the gene
         * with the given allele value.
         */
        private IGene[,] m_geneNeighbours ;

        /** The number of neighbouring genes for the gene with the given allele value. */
        private int[] m_geneNeighboursCount ;

        /** Array of neighbour indices: m_neighbourIndices[i] = i + 1 % m_chromosomeSize. */
        private int[] m_neighbourIndices ;

        /** The pool for genes. */
        private IGene[] m_genePool ;

        /** The size of the gene pool. */
        private int m_genePoolSize ;

        /** The mapping between allele values and the genes' positions in the pool. */
        private int[] m_alleleToPoolIndex ;

        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /** The chromosome pool that is used to acquire chromosome instances. */
        private IChromosomePool m_chromosomePool ;

        /** The gene distance measurer. */
        private IGeneDistanceMeasurer m_measurer ;

        /** The chromosome size. */
        private int m_chromosomeSize ;

        /**
         * Creates new cycle crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         */
        public CHeuristicCrossover(CConfiguration configuration, double crossoverRate) : base(configuration, crossoverRate)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_chromosomePool      = configuration.getChromosomePool() ;
            m_chromosomeSize      = configuration.getChromosomeSize() ;
            m_randomizer          = configuration.getRandomGenerator() ;
            m_geneNeighbours      = new IGene[m_chromosomeSize, 2] ;
            m_geneDistances       = new double[m_chromosomeSize, 2] ;
            m_geneNeighboursCount = new int[m_chromosomeSize] ;
            m_neighbourIndices    = new int[m_chromosomeSize] ;
            m_totalDistances      = new double[m_chromosomeSize] ;
            m_genePoolSize        = 0 ;
            m_genePool            = new IGene[m_chromosomeSize] ;
            m_alleleToPoolIndex   = new int[m_chromosomeSize] ;            
            m_measurer            = getConfiguration().getGeneDistanceMeasurer() ;

            m_neighbourIndices[m_chromosomeSize - 1] = 0 ;

            for (int i = 0 ; i < m_chromosomeSize - 1 ; i++)
            {
                m_neighbourIndices[i] = i + 1 ;
            }
        }

        /**
         * Fills the gene pool (this.m_genePool) with genes from
         * the specified chromosome, and creates the mapping from 
         * allele values to the genes in the pool (this.m_alleleToPoolIndex).
         * 
         * @param chromosome The chromosome whose genes will fill
         *                   the gene pool.
         */
        private void fillGenePool(IChromosome chromosome)
        {
#if DEBUG_HX
            if (m_genePoolSize != 0)
            {
                throw new System.InvalidOperationException("The pool has to be empty before calling fill!") ;
            }
#endif
            while (m_genePoolSize < m_chromosomeSize)
            {
                m_alleleToPoolIndex[(int)chromosome.getGene(m_genePoolSize).getAllele()] = m_genePoolSize ;
                m_genePool[m_genePoolSize] = chromosome.getGene(m_genePoolSize) ;
                ++m_genePoolSize ;
            }
        }

        /**
         * Checks whether a gene with the specified allele value
         * is present in the gene pool or not. 
         * 
         * @param allele The allele value of the gene to be checked.
         * 
         * @return True if a gene with the specified allele is in
         *         the pool, otherwise false.
         */
        private bool isGeneInPool(int allele)
        {
            return (m_alleleToPoolIndex[allele] != -1) ;
        }

        /**
         * Removes the gene with the specified allele value
         * from the gene pool. It also returns the gene
         * as an output value.
         * 
         * @param allele The gene's to be removed allele value.
         * 
         * @return The removed gene.
         */
        private IGene removeGeneFromPool(int allele)
        {
            IGene gene = m_genePool[m_alleleToPoolIndex[allele]] ;

            m_alleleToPoolIndex[(int)m_genePool[--m_genePoolSize].getAllele()] = m_alleleToPoolIndex[allele] ;
            m_genePool[m_alleleToPoolIndex[allele]] = m_genePool[m_genePoolSize] ;
            
            m_alleleToPoolIndex[allele] = -1 ;

            return gene ;
        }

        /**
         * Removes a random gene from the gene pool and returns
         * it.
         * 
         * @param randomizer A random number generator.
         * 
         * @return A randomly selected gene.
         */
        private IGene removeRandomGeneFromPool() 
        {
            return removeGeneFromPool((int)m_genePool[m_randomizer.Next(m_genePoolSize)].getAllele()) ;
        }

        public override void apply(IChromosome daddy, IChromosome mommy, CPopulation population)
        {
            /* If the crossover should not be performed. Return. */
            if (!shouldPerformCrossover())
            {
                IChromosome son = m_chromosomePool.acquire() ;
            
                if (son != null)
                {
                    son.setGenes(daddy.getGenes()) ;
                }
                else
                {
                    son = (IChromosome)daddy.Clone() ;
                }

                IChromosome daughter = m_chromosomePool.acquire() ;
            
                if (daughter != null)
                {
                    daughter.setGenes(mommy.getGenes()) ;
                }
                else
                {
                    daughter = (IChromosome)mommy.Clone() ;
                }

                population.addChromosome(son) ;
                population.addChromosome(daughter) ;

                return ;
            }

            /* The result of the crossover. */
            IChromosome offspring = m_chromosomePool.acquire() ;
            
            if (offspring != null)
            {
                offspring.setGenes(daddy.getGenes()) ;
            }
            else
            {
                offspring = (IChromosome)daddy.Clone() ;
            }

            /* Determines the connections between neigbouring genes, i.e.
             * creates a list of neighbours for each gene and counts 
             * the lengths of each of the connections (edges).
             */
            for (int i = 0 ; i < m_chromosomeSize ; i++)
            {
                IGene gene = daddy.getGene(i) ;

                int allele = (int)gene.getAllele() ;

                IGene neighbourGeneOne = daddy.getGene(m_neighbourIndices[i]) ;
                IGene neighbourGeneTwo = mommy.getGene(m_neighbourIndices[i]) ;

                m_geneNeighbours[allele, m_geneNeighboursCount[allele]]  = neighbourGeneOne ;
                m_geneDistances[allele, m_geneNeighboursCount[allele]++] = m_measurer.distance(gene, neighbourGeneOne) ;

                if (!neighbourGeneOne.Equals(neighbourGeneTwo))
                {
                    m_geneNeighbours[allele, m_geneNeighboursCount[allele]]  = neighbourGeneTwo ;
                    m_geneDistances[allele, m_geneNeighboursCount[allele]++] = m_measurer.distance(gene, neighbourGeneTwo) ;
                }
            }

            /* Calculates the cumulative sum of distances for each
             * gene and scales the distances in a way that the shorter
             * the distance the better the connection (edge) between
             * the particular genes, i.e. it has higher probability
             * to be passed on the offspring.
             */
            for (int i = 0 ; i < m_chromosomeSize ; i++)
            {
                for (int j = 0 ; j < m_geneNeighboursCount[i] ; j++)
                {
                    m_totalDistances[i] += m_geneDistances[i, j] ;
                }

                for (int j = 0 ; j < m_geneNeighboursCount[i] ; j++)
                {
                    m_geneDistances[i, j] = m_totalDistances[i] / (m_geneDistances[i, j] + 0.5) ;

                    m_totalDistances[i] = 0.0 ;
                }

                for (int j = 0 ; j < m_geneNeighboursCount[i] ; j++)
                {
                    m_totalDistances[i] += m_geneDistances[i, j] ;
                }
            }

            /* Fills the gene pool with genes from the daddy chromosome,
             * and creates the mapping from allele values to the genes
             * in the pool.
             */
            fillGenePool(daddy) ;

            int iGeneLocusToSet = 0 ;

            int geneAllele = m_randomizer.Next(m_chromosomeSize) ;

            offspring.setGene(removeGeneFromPool(geneAllele), iGeneLocusToSet++) ;

            while (iGeneLocusToSet < m_chromosomeSize)
            {
                /* Erases genes that have been already passed on the offspring 
                 * from the neighbours list of the currently selected gene.
                 */
                for (int i = 0 ; i < m_geneNeighboursCount[geneAllele] ; i++)
                {
                    if (!isGeneInPool((int)m_geneNeighbours[geneAllele, i].getAllele()))
                    {
                        m_totalDistances[geneAllele] -= m_geneDistances[geneAllele, i] ;

                        m_geneNeighbours[geneAllele, i] = m_geneNeighbours[geneAllele, --m_geneNeighboursCount[geneAllele]] ;
                        --i ;
                    }
                }

                /* If there is at least one neighbouring gene to choose
                 * from, then the next gene is determined randomly, but
                 * the shorter the distance between the current gene
                 * and a possibly next gene, the higher the probability
                 * it will become really the next gene and will be passed
                 * on the offspring.
                 */
                if (m_geneNeighboursCount[geneAllele] > 0)
                {
                    /* Determines the next gene to be passed on the offspring
                     * according to distribution defined by the distances of
                     * the neighbouring genes from the current gene.
                     */
                    double distance   = m_randomizer.NextDouble() * m_totalDistances[geneAllele] ;
                    int    ineighbour = 0 ;

                    while (ineighbour < m_geneNeighboursCount[geneAllele] && m_geneDistances[geneAllele, ineighbour] < distance)
                    {
                        distance -= m_geneDistances[geneAllele, ineighbour++] ;
                    }

                    geneAllele = (int)m_geneNeighbours[geneAllele, ineighbour].getAllele() ;

                    offspring.setGene(removeGeneFromPool(geneAllele), iGeneLocusToSet++) ;
                }
                else
                {
                    /* There is no neighbouring gene to the previously set gene
                     * that would not be already present in the offspring. Random
                     * gene is selected then.
                     */
                    IGene nextGene = removeRandomGeneFromPool() ;

                    offspring.setGene(nextGene, iGeneLocusToSet++) ;

                    geneAllele = (int)nextGene.getAllele() ;
                }
            }
            
            for (int i = 0 ; i < m_chromosomeSize ; i++)
            {
                m_totalDistances[i] = 0.0 ;
                m_geneNeighboursCount[i] = 0 ;
            }

            population.addChromosome(offspring) ;

#if DEBUG_HX
            printCrossoverDebugInfo("Heuristic Crossover:", daddy, mommy, offspring, null) ;
#endif
        }
    }
}