﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_PMX

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The partially mapped crossover operator implementation.
     *
     * @author Tomas Tunys.
     */
    public class CPartiallyMappedCrossover : CBaseCrossoverOperator, IConfigurationDependent
    {
        /** The mapping between the parts of the parent chromosomes. */
        private int[] m_mapping ;

        /** The position of a gene with the give allele value. */
        private int[] m_inverse ;

        /** The chromosome pool that is used to acquire chromosome instances. */
        private IChromosomePool m_chromosomePool ;

        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /** The chromosome size. */
        private int m_chromosomeSize ;

        /**
         * Creates new partially mapped crossover operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param crossoverRate The crossover rate.
         */
        public CPartiallyMappedCrossover(CConfiguration configuration, double crossoverRate) : base(configuration, crossoverRate)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_randomizer     = configuration.getRandomGenerator() ;
            m_chromosomePool = configuration.getChromosomePool() ;
            m_chromosomeSize = configuration.getChromosomeSize() ;
            m_mapping        = new int[m_chromosomeSize] ;
            m_inverse        = new int[m_chromosomeSize] ;
        }

        /**
         * Crossovers the specified couple of chromosomes in a operator-specific way
         * forming a new chromosomes that should be added to the specified population.
         * 
         * @param daddy      The first chromosome entering into the crossover.
         * @param mommy      The second chromosome entering into the crossover.
         * @param population The population for the next generation.
         */
        public override void apply(IChromosome daddy, IChromosome mommy, CPopulation population)
        {
            IChromosome son = m_chromosomePool.acquire() ;
            
            if (son != null)
            {
                son.setGenes(daddy.getGenes()) ;
            }
            else
            {
                son = (IChromosome)daddy.Clone() ;
            }

            IChromosome daughter = m_chromosomePool.acquire() ;
            
            if (daughter != null)
            {
                daughter.setGenes(mommy.getGenes()) ;
            }
            else
            {
                daughter = (IChromosome)mommy.Clone() ;
            }

            if (shouldPerformCrossover() && !daddy.Equals(mommy))
            {
                int start = m_randomizer.Next(m_chromosomeSize) ;
                int end   = m_randomizer.Next(m_chromosomeSize) ;

                if (end < start)
                {
                    swap(ref start, ref end) ;
                }

                /* Creates the identity permutation. */
                for (int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    m_mapping[i] = i ;
                }

                /* Build mapping between daddy's gene alleles and mommy's gene
                 * alleles found in the randomly chosen section. 
                 */
                for (int i = start ; i <= end ; i++)
                {
                    swap(ref m_mapping[(int)daddy.getGene(i).getAllele()], ref m_mapping[(int)mommy.getGene(i).getAllele()]) ;    
                }

                /* Creates and mapping between allele values and gene positions
                 * that bear these values of the daddy chromosome.
                 */
                for (int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    m_inverse[(int)daddy.getGene(i).getAllele()] = i ;
                }

                /* Transfers the part from the mommy chromosome into daddy
                 * chromosome in a permutation-consistent way.*/
                for (int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    int allele = (int)son.getGene(i).getAllele() ;

                    if (m_mapping[allele] < allele)
                    {
                        swapGenes(son, i, m_inverse[m_mapping[allele]]) ;
                    }
                }

                /* Creates and mapping between allele values and gene positions
                 * that bear these values of the mommy chromosome.
                 */
                for (int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    m_inverse[(int)daughter.getGene(i).getAllele()] = i ;
                }

                /* Transfers the part from the daddy chromosome into mommy
                 * chromosome in a permutation-consistent way.*/
                for (int i = 0 ; i < m_chromosomeSize ; i++)
                {
                    int allele = (int)daughter.getGene(i).getAllele() ;

                    if (m_mapping[allele] < allele)
                    {
                        swapGenes(daughter, i, m_inverse[m_mapping[allele]]) ;
                    }
                }
            }

            population.addChromosome(son) ;
            population.addChromosome(daughter) ;
#if DEBUG_PMX
            printCrossoverDebugInfo("Partially Mapped Crossover: " + (start + 1) + " - " + (end + 1), daddy, mommy, son, daughter) ;
#endif
        }
    }
}
