﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A population consist of a individual chromosomes that forms potential
     * solutions to a problem.
     *
     * @author Tomas Tunys
     */
    public class CPopulation : ICloneable
    {
        /**
         * The list of individuals (chromosomes) forming the population.
         */
        private List<IChromosome> m_chromosomes ;

        /**
         * The fittest chromosome in the population.
         */
        private IChromosome m_fittestChromosome ;

        /**
         * The worst fit chromosome in the population.
         */
        private IChromosome m_worstFitChromosome ;

        /**
         * Indicates whether at least one of the chromosomes has been changed
         * (deleted, added, modified) non-externally. This value is used to 
         * speed up the determination of the fittest chromosome in the population.
         */
        private bool m_changed ;

        /**
         * Indicates that the list of chromosomes has been sorted according
         * to their fitness evaluation.
         */
        private bool m_sorted ;

        /**
         * The configuration of the genetic algorithm used to evolve 
         * this population of chromosomes.
         */
        private CConfiguration m_configuration ;

        /** The maximum population size (if set). */
        private int m_populationSizeMaximum ;

        /**
         * Creates a new empty population.
         * 
         * @param configuration The configuration of the genetic algorithm used
         *                      to evolve this population.
         */
        public CPopulation(CConfiguration configuration) : this(configuration, 64)
        {
        }

        /**
         * Creates a new population from the specified list of chromosomes.
         * <p>
         * <strong>Note:</strong>
         * Individual chromosomes are not cloned! This means that the structure
         * of the specified chromosomes may be changed during the evolution!
         * </p>
         *
         * @param configuration The configuration of the genetic algorithm used
         *                      to evolve this population.
         * @param chromosomes   The list of chromosomes forming the new population.
         */
        public CPopulation(CConfiguration configuration, IChromosome[] chromosomes) : this(configuration, chromosomes.Length)
        {
            m_chromosomes.AddRange(chromosomes) ;
        }

        /**
         * Creates a new population formed by the specified chromosome.
         * <p>
         * <strong>Note:</strong>
         * Individual chromosomes are not cloned! This means that the structure
         * of the specified chromosomes may be changed during the evolution!
         * </p>
         *
         * @param configuration The configuration of the genetic algorithm used
         *                      to evolve this population.
         * @param chromosome    The single chromosome forming the new population.
         */
        public CPopulation(CConfiguration configuration, IChromosome chromosome) : this(configuration, 1)
        {
            m_chromosomes.Add(chromosome) ;
        }

        /**
         * Creates a new empty population. The size argument is used for optimalization
         * reason (lesser memory reallocation).
         * 
         * @param configuration The configuration of the genetic algorithm used
         *                      to evolve this population.
         *                      
         * @param size          The number of preallocated slots for the individuals
         *                      of the initial population.
         */
        public CPopulation(CConfiguration configuration, int size)
        {
            if (configuration == null) 
            {
                throw new NullReferenceException("The configuration must not be null!") ;
            }
            else
            {
                m_configuration = configuration ;
            
                m_chromosomes = new List<IChromosome>(size) ;

                setMaximumSize(int.MaxValue) ;
            
                setChanged(true) ;
            }
        }

        /**
         * Gets or sets the chromosome at the specified index.
         * 
         * @param index The zero-based index of the chromosome
         *              to get or set.
         * @return      The chromosome at the specified index.
         * 
         * @throws ArgumentOutOfRangeException If the index
         *         is less then 0 or greater-or-equal to 
         *         <em>this.size()</em>.
         */
        public IChromosome this[int index]
        {
            get { return getChromosome(index) ; }
            set { setChromosome(index, value) ; }
        }

        /**
         * Returns the configuration of the genetic algorithm used
         * to evolve this population.
         * 
         * @return The configuration of the genetic algorithm used
         *         to evolve this population.
         */
        public CConfiguration getConfiguration() 
        {
            return m_configuration ;
        }

        /**
         * Adds a chromosome to this Population.
         *
         * @param chromosome The chromosome to be added.
         */
        public void addChromosome(IChromosome chromosome)
        {
            if (chromosome != null && getSize() < m_populationSizeMaximum) 
            {
                m_chromosomes.Add(chromosome) ;
                setChanged(true) ;
            }
        }

        /**
        * Adds 'all' the chromosomes in the specified population,
         * if the population has size maximum set, then some of
         * the chromosomes will be ommited.
        *
        * @param population The population of chromosomes to be added.
        */
        public void addChromosomes(CPopulation population) 
        {
            if (population != null) 
            {
                if (getSize() + population.getSize() < m_populationSizeMaximum)
                {
                    m_chromosomes.AddRange(population.getChromosomes()) ;
                }
                else
                {
                    for (int i = 0 ; getSize() < m_populationSizeMaximum ; i++)
                    {
                        m_chromosomes.Add(population[i]) ;
                    }
                }
                        
                setChanged(true) ;
            }
        }

        /**
         * Replaces all chromosomes in the population with the specified
         * list of chromosomes.
         *
         * @param chromosomes The chromosomes forming the population.
         */
        public void setChromosomes(IEnumerable<IChromosome> chromosomes) 
        {
            m_chromosomes.Clear() ;
            m_chromosomes.AddRange(chromosomes) ;
            setChanged(true) ;
        }

        /**
         * Replaces the chromosome at the specified index with the specified
         * chromosome. If the index is greater or equal to the population size
         * the chromosome is added to the population.
         *
         * @param index      The position of the chromosome to be replaced.
         * @param chromosome The chromosome to be set.
         */
        public void setChromosome(int index, IChromosome chromosome)
        {
            if (getSize() <= index) 
            {
                addChromosome(chromosome) ;
            }
            else 
            {
                m_chromosomes[index] = chromosome ;
                setChanged(true) ;
            }
        }

        /**
         * Returns the chromosome at the specified position.
         * 
         * @param index The index of the chromosome to be returned.
         * @return      A chromosome at specified index in the population.
         */
        public IChromosome getChromosome(int index)
        {
            return m_chromosomes[index] ;
        }

        /**
         * Removes a chromosome at the specified index in the population.
         *
         * @param index The index of the chromosome to be removed.
         */
        public void removeChromosome(int index) 
        {
            setChanged(m_fittestChromosome.Equals(m_chromosomes[index])) ;
            m_chromosomes.RemoveAt(index) ;
        }

        /**
         * Returns the unmodifiable list of chromosomes forming this population.
         * 
         * @return A readonly list of chromosomes forming this population.
         */
        public IList<IChromosome> getChromosomes()
        {
            return m_chromosomes.AsReadOnly() ;
        }

        /**
         * Returns the population size.
         * 
         * @return The number of chromosomes in the population.
         */
        public int getSize()
        {
            return m_chromosomes.Count ;
        }

        /**
         * Removes all the chromosomes from the population.
         */
        public void clear()
        {
            m_changed = true ;
            m_sorted  = true ;

            m_fittestChromosome  = null ;
            m_worstFitChromosome = null ;

            m_chromosomes.Clear() ;
        }

        /**
         * Returns the population of the chromosomes in an array.
         * 
         * @return The population in the form of an array.
         */
        public IChromosome[] toChromosomes() 
        {
            return m_chromosomes.ToArray() ;
        }

        /**
         * Determines the best and worst fit chromosome in the population
         * and assigns them to the respective fields: this.m_fittestChromosome,
         * and this.m_worstFitChromosome. 
         */
        private void determineBestAndWorstFitChromosome()
        {
            IFitnessComparer comparer = getConfiguration().getFitnessComparer() ;

            List<IChromosome>.Enumerator e = m_chromosomes.GetEnumerator() ;

            if (e.MoveNext())
            {
                m_fittestChromosome  = e.Current ;
                m_worstFitChromosome = e.Current ;
            }
            else
            {
                m_fittestChromosome  = null ;
                m_worstFitChromosome = null ;
            }

            while (e.MoveNext())
            {
                if (comparer.Compare(m_fittestChromosome, e.Current) < 0)
                {
                    m_fittestChromosome = e.Current ;
                }
                else if (comparer.Compare(m_worstFitChromosome, e.Current) > 0)
                {
                    m_worstFitChromosome = e.Current ;
                }
            }

            setChanged(false) ;
        }

        /**
         * Returns the fittest chromosome in the population. The comparison
         * of the fitness is made by the <em>IFitnessComparer</em> set in
         * the configuration.
         *
         * @return The fittest chromosome of the population.
         */
        public IChromosome getFittestChromosome()
        {
            if (!m_changed && m_fittestChromosome != null)
            {
                return m_fittestChromosome ;
            }
            else
            {
                determineBestAndWorstFitChromosome() ;
            }

            return m_fittestChromosome ;
        }

        /**
         * Returns the worst fit chromosome in the population. The comparison
         * of the fitness is made by the <em>IFitnessComparer</em> set in
         * the configuration.
         *
         * @return The worst fit chromosome of the population.
         */
        public IChromosome getWorstFitChromosome()
        {
            if (!m_changed && m_worstFitChromosome != null)
            {
                return m_worstFitChromosome ;
            }
            else
            {
                determineBestAndWorstFitChromosome() ;
            }

            return m_worstFitChromosome ;
        }


        /**
         * Informs the population that its fittest chromosome may have changed.
         *
         * @param changed If true, the population's fittest chromosome may have changed,
         *                if false, the current fittest chromosome is still valid.
         */
        protected void setChanged(bool changed) 
        {
            m_changed = changed ;
            setSorted(false) ;
        }

        /**
         * Returns true if the population has been changed, otherwise it returns
         * false.
         * 
         * @return True if the population has been changed, false otherwise.
         */
        public bool isChanged()
        {
            return m_changed ;
        }

        /**
        * Marks the population as sorted.
        *
        * @param sorted If true the population is sorted.
        */
        protected void setSorted(bool sorted) 
        {
            m_sorted = sorted ;
        }

        /**
         * Returns true if the population is sorted, false otherwise.
         */
        public bool isSorted() 
        {
            return m_sorted ;
        }

        /**
         * Returns true if the maximum capacity of the population is
         * set; otherwise, it returns false. 
         */
        public bool hasMaximumSize()
        {
            return (m_populationSizeMaximum != int.MaxValue) ;
        }

        /**
         * Sets the maximum size of the population. Any chromosome
         * added to the population that would make it larger then
         * this value is ignored.
         * 
         * @param maximumSize The maximum size of the population.
         */
        public void setMaximumSize(int maximumSize)
        {
            m_populationSizeMaximum = maximumSize ;
        }

        /**
         * Determines whether the specified chromosome is present in this population.
         *
         * @param chromosome The chromosome to be checked.
         * 
         * @return True if the chromosome is in this population, false otherwise.
         */
        public bool contains(IChromosome chromosome) 
        {
            return m_chromosomes.Contains(chromosome) ;
        }
        
        /**
         * Sorts the chromosomes in the population according to their fitness
         * value using <em>this.sortByfitness()</em>, and then returns 
         * the specified number of best fit chromosomes.
         *
         * @param count The number of best fit chromosomes in the population
         *             to be returned.
         *
         * @return An enumerator of the best fit chromosomes of the population. The
         *         number of items the enumerator enumerates is equal to the count 
         *         parameter or to the size of the population in case <em>this.size() 
         *         &lt; count</em>.
         */
        public IEnumerator<IChromosome> getFittestChromosomes(int count) 
        {
            count = Math.Min(count, getSize()) ;

            if (count <= 0) 
            {
                return null ;
            }
            else if (m_changed || !m_sorted) 
            {
                sortByFitness() ;
            }

            return new FittestChromosomeEnumerator(this, count) ;
        }

        /**
         * FittestChromosomeEnumerator may be used to enumerate fittest
         * chromosomes of a population in memory efficient way.
         */
        protected class FittestChromosomeEnumerator : IEnumerator<IChromosome>
        {
            private CPopulation m_population ;
            private int         m_minIndex ;
            private int         m_index ;

            /**
             * Creates a fitness chromosome enumerator for the specified population
             * that enumerates only first count number of fittest chromosomes.
             * 
             * @param population The population for which an enumerator is created.
             * @param count      The maximum number of enumerated chromosomes of
             *                   the population.
             */
            internal FittestChromosomeEnumerator(CPopulation population, int count)
            {
                m_index      = population.getSize() ;
                m_minIndex   = m_index - count ;
                m_population = population ;
            }

            /**
             * Returns the current chromosome.
             * 
             * @return The current chromosome.
             */
            public IChromosome Current
            {
                get { return m_population[m_index] ; }
            }

            /**
             * Destroys the enumerator.
             */
            public void Dispose()
            {
                m_population = null ;
            }

            /**
             * Returns the current chromosome.
             * 
             * @return The current chromosome.
             */
            object System.Collections.IEnumerator.Current
            {
                get { return m_population[m_index] ; }
            }

            /**
             * Moves the enumerator to the next less fit chromosome.
             * 
             * @return True, if there is another chromosome, otherwise,
             *         it returns false.
             */
            public bool MoveNext()
            {
                return (--m_index >= m_minIndex) ;
            }

            /**
             * Resets the enumerator.
             */
            public void Reset()
            {
                m_index = m_population.getSize() ;
            }
        }

        /**
         * Sorts the chromosomes in the population according to their fitness
         * value using the IFitnessComparer set in the configuration.
         * The chromosomes are sorted from the fittest to the worst fit,
         * which results in the fittest chromosome being positioned at index 0.
         */
        public void sortByFitness() 
        {
            m_chromosomes.Sort(getConfiguration().getFitnessComparer()) ;
            setChanged(false) ;
            setSorted(true) ;

            m_fittestChromosome  = m_chromosomes[getSize() - 1] ;
            m_worstFitChromosome = m_chromosomes[0] ; 
        }
        
        /**
         * Determines whether the specified object is equal to 
         * this population.
         *
         * @param population The object to be compared.
         * 
         * @return True if the specified object is equal to this 
         *         population, otherwise false.
         */
        public override bool Equals(Object population) 
        {
            try 
            {
                CPopulation other = (CPopulation)population ;

                if (other == null || getSize() != other.getSize())
                {
                    return false ;
                }

                foreach (IChromosome chromosome in m_chromosomes)
                {
                    if (other.contains(chromosome))
                    {
                        return false ;
                    }
                }
            }
            catch (InvalidCastException) 
            {
                return false ;
            }

            return true ;
        }

        /**
         * Returns the hashcode of this population.
         * 
         * @return The population's hash code.
         */
        public override int GetHashCode()
        {
            int hash = 13 ;

            foreach (IChromosome chromosome in m_chromosomes)
            {
                hash ^= chromosome.GetHashCode() ;
            }

            return hash ;
        }

        /**
         * Creates a deep copy of this population.
         * 
         * @return A new population that is a copy of this one.
         */
        public object Clone() 
        {
            CPopulation result = new CPopulation(getConfiguration(), getSize()) ;

            foreach (IChromosome chromosome in m_chromosomes)
            {
                result.addChromosome((IChromosome)chromosome.Clone()) ;
            }

            return result ;
        }
    }
}
