﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A priority queue implementation based on a binary heap
     * structure, which guarantees that all the operations define
     * has logarithmic time complexity.
     * 
     * @author Tomas Tunys
     */
    class CPriorityQueue<T>
    {
        /** The binary heap structure represented by an array. */
        private List<T> m_queue ;

        /** The number of elements in the queue. */
        private int m_length ;

        /**
         * The maximum length of the queue. Only relevant for queues
         * of fixed length.
         */
        private int m_maximumLength ;

        /** The comparer used to determine the ordering of the elements. */
        private IComparer<T> m_comparer ;

        /**
         * Creates a new priority queue with initial capacity
         * of 64 elements.
         * 
         * @param comparer    The comparer used to determine
         *                    the ordering of the elements 
         *                    within the queue.
         *                    
         */
        public CPriorityQueue(IComparer<T> comparer) : this(comparer, 64, false)
        {
        }

        /**
         * Creates a new priority queue with the specified
         * initial capacity, which is used only for performace
         * reasons.
         * 
         * @param initialCapacity The initial capacity of the queue.
         * @param comparer        The comparer used to determine
         *                        the ordering of the elements 
         *                        within the queue.
         * @param fixedLength     If true, the queue will have fixed
         *                        length given by initialCapacity.
         */
        public CPriorityQueue(IComparer<T> comparer, int initialCapacity, bool fixedLength)
        {
            m_queue = new List<T>(initialCapacity + 1) ;
            m_comparer = comparer ;
            m_maximumLength = (fixedLength ? initialCapacity : int.MaxValue) ;
        }

        /**
         * Swaps the content of the two elements specified 
         * by their position in the queue.
         * 
         * @param i The first element's position within the queue.
         * @param j The second element's position within the queue.
         */
        private void swapQueueElements(int i, int j)
        {
           T t = m_queue[i] ; m_queue[i] = m_queue[j] ; m_queue[j] = t ;
        }

        /**
         * Fixes the heap upwards from the i-th element
         * in the heap.
         * 
         * @param The position of the element in the heap
         *        to start its reconstruction.
         */
        private void fixUp(int i)
        {
            int j = (i - 1) / 2 ;

            while (i > 0 && m_comparer.Compare(m_queue[i], m_queue[j]) > 0)
            {
                swapQueueElements(i, j) ;

                i = j ;
                j = (i - 1) / 2 ;
            }
        }

        /**
         * Fixes the heap downwards from the i-th element
         * in the heap.
         * 
         * @param The position of the element in the heap
         *        to start its reconstruction.
         */
        private void fixDown(int i)
        {
            int j = 2 * i + 1 ;

            while (j < m_length)
            {
                if (j + 1 < m_length && m_comparer.Compare(m_queue[j], m_queue[j + 1]) < 0)
                {
                    ++j ;
                }

                if (m_comparer.Compare(m_queue[i], m_queue[j]) < 0)
                {
                    swapQueueElements(i, j) ;
                }
                else
                {
                    break ;
                }

                i = j ;
                j = 2 * i + 1 ;
            }
        }

        /**
         * Inserts the specified element into the queue. This increases
         * the size of the queue by .
         * 
         * @param element The element to be inserted into the queue.
         */
        public void push(T element)
        {
            if (m_length < m_maximumLength)
            {
                /* To allow inflation of the queue array. */
                m_queue.Add(element) ;

                fixUp(m_length++) ;
            }
            else
            {
                int i = 0 ;

                while (i < m_length)
                {
                    if (m_comparer.Compare(m_queue[i], element) < 0)
                    {
                        T swapee   = m_queue[i] ;
                        m_queue[i] = element ;
                        element    = swapee ;
                    }

                    i = 2 * i + 1 ;

                    if (i + 1 < m_length && m_comparer.Compare(m_queue[i], m_queue[i + 1]) > 0)
                    {
                        ++i ;
                    }
                }

            }
        }

        /**
         * Returns an element with the maximum key value, and
         * removes it from the queue.
         * 
         * @return An element with the greatest key value.
         * 
         * @throws InvalidOperationException If the queue is empty,
         *                                  and there is no element
         *                                  to return.
         */
        public T pop()
        {
            if (m_length == 0)
            {
                throw new InvalidOperationException("The priority queue is empty!") ;
            }

            T element  = m_queue[0] ;
            m_queue[0] = m_queue[--m_length] ;

            /* To allow deflation of the queue array. */
            m_queue.RemoveAt(m_length) ;

            fixDown(0) ;

            return element ;
        }

        /**
         * Returns an element with the maximum key value,
         * contrary to the pop method, the element is not
         * removed from the queue.
         * 
         * @return An element with the greatest key value.
         * 
         * @throws InvalidOperationException If the queue is empty,
         *                                  and there is no element
         *                                  to return.
         */
        public T peek()
        {
            if (m_length == 0)
            {
                throw new InvalidOperationException("The priority queue is empty!") ;
            }

            return m_queue[0] ;
        }

        /**
         * Returns the length of the queue, i.e. the number
         * of elemnts in the queue.
         * 
         * @return The number of elements in the queue.
         */
        public int length()
        {
            return m_length ;
        }

        /**
         * Returns read-only list of the elements 
         * in the queue.
         * 
         * @return A read-only list of elements in the queue.
         */
        public ReadOnlyCollection<T> elements()
        {
            return m_queue.AsReadOnly() ;
        }
    }
}
