﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The random tournament selection operator implementation.
     * 
     * @author Tomas Tunys
     */
    public class CRandomTournamentSelector : CBaseSelectionOperator, IConfigurationDependent
    {
        /** The positions of the chromosomes selected for the tournament. */
        private int[] m_selected ;

        /** The probability the fitter chromosome will win the match. */
        private double m_fitterWinProbability ;

        /** The size of the tournament. */
        private int m_tournamentSize ;

        /** The population size. */
        private int m_populationSize ;

        /** The fitness comparer used to evaluate fitter individual of the match. */
        IFitnessComparer m_fitnessComparer ;

        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /**
         * Instantiates a new random tournament selector.
         * 
         * @param configuration        The configuration of the genetic algorithm
         *                             that will utilize this selector.                    
         * @param tournamentSize       The size of the tournament.
         * @param fitterWinProbability The probability of fitter chromosome winning
         *                             a match in the tournament.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */
        public CRandomTournamentSelector(CConfiguration configuration, int tournamentSize, double fitterWinProbability) : base(configuration)
        {
            setTournamentSize(tournamentSize) ;
            setFitterWinProbability(fitterWinProbability) ;
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this selector.
         */
        public void notify(CConfiguration configuration)
        {
            m_randomizer      = configuration.getRandomGenerator() ;
            m_populationSize  = configuration.getPopulationSize() ;
            m_fitnessComparer = configuration.getFitnessComparer() ;
        }

        /**
         * Sets the probability of a fitter chromosome winning a match.
         * 
         * @param fitterWinProbability The probability that a fitter chromosome
         *                             wins a match.
         *                             
         * @throws ArgumentOutOfRangeException If fitterWinProbability is not 
         *                                     in the [0,1] interval.
         */
        public void setFitterWinProbability(double fitterWinProbability)
        {
            if (fitterWinProbability < 0.0 || fitterWinProbability > 1.0)
            {
                throw new ArgumentOutOfRangeException("The probability must be decimal value between 0.0 and 1.0 (inclusive)!") ;
            }
            m_fitterWinProbability = fitterWinProbability ;
        }

        /**
         * Returns the probability of a fitter chromosome winning
         * a match.
         * 
         * @return The probability of a fitter chromosome winning
         *         a match.
         */
        public double getFitterWinProbability()
        {
            return m_fitterWinProbability ;
        }

        /**
         * Sets the size of the tournament.
         * 
         * @param tournamentSize The size of the tournament.
         */
        public void setTournamentSize(int tournamentSize)
        {
            if (m_tournamentSize != tournamentSize)
            {
                m_tournamentSize = tournamentSize ;
                m_selected       = new int[tournamentSize] ;
            }
        }

        /**
         * Returns the tournament size.
         * 
         * @return The tournament size.
         */
        public int getTournamentSize()
        {
            return m_tournamentSize ;
        }

        /**
         * Fills the array this.m_selected with indices of individuals in the population
         * that have been selected for the tournament. 
         */
        public void getCompetitorIndices()
        {
            int numberOfCompetitors = 0, highestCompetitorIndex = m_populationSize - m_tournamentSize ;

            /* Randomly selects competitors of a tournament. */
            while (highestCompetitorIndex < m_populationSize)
            {
                int iSelected = m_randomizer.Next(highestCompetitorIndex + 1) ;

                if (Array.BinarySearch<int>(m_selected, 0, numberOfCompetitors, iSelected) >= 0)
                {
                    iSelected = highestCompetitorIndex ;
                }

                m_selected[numberOfCompetitors++] = iSelected ;

                ++highestCompetitorIndex ;
            }
        }

        /**
         * Selects a chromosome from the specified population according to its
         * fitness. See the description of  the algorithm in the description
         * of this class.
         * 
         * @param population    The population the chromosome will be selected from.
         * @param matingPool    The mating pool the chromosome will be inserted into.
         * @param toSelectCount The total number of chromosomes to select.
         */
        public override void select(CPopulation population, CPopulation matingPool, int toSelectCount)
        {
            for (int tournament = 0 ; tournament < toSelectCount ; tournament++)
            {
                /* Selects competitors - indices of individuals selected for the tournament. */
                getCompetitorIndices() ;

                /* Determines the winner of the tournament. */
                for (int step = 1 ; step < m_tournamentSize ; step++)
                {
                    for (int i = 0 ; i + step < m_tournamentSize ; i += step + 1)
                    {
                        if (m_fitnessComparer.Compare(population[m_selected[i]], population[m_selected[i + step]]) < 0)
                        {
                            if (m_randomizer.NextDouble() <= m_fitterWinProbability)
                            {
                                m_selected[i] = m_selected[i + step] ;
                            }
                        }
                        else
                        {
                            if (m_randomizer.NextDouble() > m_fitterWinProbability)
                            {
                                m_selected[i] = m_selected[i + step] ;
                            }
                        }
                    }
                }

                /* Adds the winner of the tournament into the mating pool. */
                matingPool.addChromosome(population[m_selected[0]]) ;
            }
        }
    }
}
