﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The remainder stochastic sampling selection operator implementation.
     * 
     * @author Tomas Tunys
     */
    public class CRemainderStochasticSampling : CBaseSelectionOperator, IConfigurationDependent
    {
        /** The fraction of a chromosome to be added to the population. */
        private double[] m_chromosomeFractions ;

        /** With or without replacement flag. */
        private bool m_withReplacement ;
        
        /** The pseudo-random number generator. */
        IRandomizer m_randomizer ;

        /**
         * Instantiates a new deterministic sampling selector.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this selector.
         *                      
         * @param withReplacement if true, the decimal fractions of the number
         *                        of expected samples of each chromosome in
         *                        the mating pool is used in weighted roulette
         *                        wheel selection, if false, the fractions are
         *                        treated as probabilities for selection.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */
        public CRemainderStochasticSampling(CConfiguration configuration, bool withReplacement) : base(configuration)
        {
            m_withReplacement = withReplacement ;
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this selector.
         */
        public void notify(CConfiguration configuration)
        {
            m_chromosomeFractions = new double[configuration.getPopulationSize()] ;
            m_randomizer          = getConfiguration().getRandomGenerator() ;
        }

        /**
         * Returns true if the selector is with replacement,
         * otherwise, false.
         * 
         * @return True if the selection is with replacement.
         */
        public bool isWithReplacement()
        {
            return m_withReplacement ;
        }

        /**
         * Selects a chromosome from the specified population according to its
         * fitness. See the description of  the algorithm in the description
         * of this class.
         * 
         * @param population    The population the chromosome will be selected from.
         * @param matingPool    The mating pool the chromosome will be inserted into.
         * @param toSelectCount The total number of chromosomes to select.
         */
        public override void select(CPopulation population, CPopulation matingPool, int toSelectCount)
        {
            enforcePositiveFitnessScores(population) ;

            double totalFitnessScore = 0.0 ;

            for (int i = 0 ; i < population.getSize() ; i++)
            {
                totalFitnessScore += population[i].getFitnessValue() ;
            }

            for (int i = 0 ; i < population.getSize() && toSelectCount > 0 ; i++)
            {
                m_chromosomeFractions[i] = population[i].getFitnessValue() / totalFitnessScore * toSelectCount ;

                int chromosomeSamplesCount = (int)Math.Truncate(m_chromosomeFractions[i]) ;

                for (int j = 0 ; j < chromosomeSamplesCount && toSelectCount > 0 ; j++)
                {
                    matingPool.addChromosome(population[i]) ;

                    --toSelectCount ;
                }

                m_chromosomeFractions[i] -= chromosomeSamplesCount ;
            }

            if (toSelectCount > 0)
            {
                int currentChromosome = 0 ;

                while (toSelectCount > 0)
                {
                    if (m_chromosomeFractions[currentChromosome] > 0.0)
                    {
                        if (m_randomizer.NextDouble() <= m_chromosomeFractions[currentChromosome])
                        {
                            matingPool.addChromosome(population[currentChromosome]) ;

                            if (!m_withReplacement)
                            {
                                m_chromosomeFractions[currentChromosome] -= 1.0 ;
                            }

                            --toSelectCount ;
                        }
                    }

                    if (++currentChromosome == m_chromosomeFractions.Length)
                    {
                        currentChromosome = 0 ;
                    }
                }
            }
        }
    }
}
