﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The roulette wheel selection is one of the standard fitness-proportional
     * selection methods. It chooses chromosomes according to their fitness 
     * value in such a way that the fitter the chromosome the more probability 
     * it has of being selected. It does not guarantee, tough, that the fittest 
     * chromosome is being selected, it is just more likely it will happen, and
     * that is why it is a good idea to use elitism along with this selector.
     * <p>
     * The selector works in this way: the population's total fitness score is 
     * represented by a pie chart - a roulette wheel - a slice of the wheel is
     * assigned to an individual chromosome of the population, and the size
     * of the slice is proportional to the chromosome's fitness, i.e. the fitter
     * the chromosome, the bigger the slice. Then the wheel is spun, and a ball
     * tossed in it. The chromosome whose slice the ball stops on is then selected
     * and inserted into so called mating pool. This process is repeated until
     * the mating pool is full (the desired number of selected chromosomes has
     * been reached).
     * </p>
     */
    public class CRouletteWheelSelector : CBaseSelectionOperator, IConfigurationDependent
    {
        /** Memory for cumulative fitness score values. */
        private double[] m_cumulativeFitness ;

        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /**
         * Instantiates a new roulette wheel selector.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this selector.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */
        public CRouletteWheelSelector(CConfiguration configuration) : base(configuration)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this selector.
         */
        public void notify(CConfiguration configuration)
        {
            m_cumulativeFitness = new double[configuration.getPopulationSize()] ;
            m_randomizer        = configuration.getRandomGenerator() ;
        }

        /**
         * Selects a chromosome from the specified population according to its
         * fitness. See the description of  the algorithm in the description
         * of this class.
         * 
         * @param population   The population the chromosome will be selected from.
         * @param matingPool    The mating pool the chromosome will be inserted into.
         * @param toSelectCount The total number of chromosomes to select.
         */
        public override void select(CPopulation population, CPopulation matingPool, int toSelectCount)
        {
            enforcePositiveFitnessScores(population) ;

            m_cumulativeFitness[0] = population[0].getFitnessValue() ;
            
            for (int i = 1 ; i < population.getSize() ; i++)
            {
                m_cumulativeFitness[i] = m_cumulativeFitness[i - 1] + population[i].getFitnessValue() ;
            }

            double maxCumulativeFitness = m_cumulativeFitness[m_cumulativeFitness.Length - 1] ;

            for (int selectedCount = 0 ; selectedCount < toSelectCount ; selectedCount++)
            {
                int iselected = Array.BinarySearch(m_cumulativeFitness, m_randomizer.NextDouble() * maxCumulativeFitness) ;

                /* Happens most of the time - a negative index is the lowest index of an individual with higher cumulative fitness 
                 * then the randomly generated value - with an appropriate conversion, of course. */
                if (iselected < 0)
                {
                    iselected = -(iselected + 1) ;
                }

                matingPool.addChromosome(population[iselected]) ;
            }
        }
    }
}
