﻿/* Uncomment following line to turn on debugging of this operator. */
//#define DEBUG_SM

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * The scramble mutation operator. It randomly chooses two points within
     * a chromosome with uniform probability, and randomly shuffles the genes
     * within the range defined by these two points.
     * 
     * @author Tomas Tunys
     */
    public class CScrambleMutation : CBaseMutationOperator, IConfigurationDependent
    {
        /** The pseudo-random number generator. */
        private IRandomizer m_randomizer ;

        /** The minimum span between chosen points. */
        private int minSpan ;

        /** The maximum span between chosen points. */
        private int maxSpan ;

        /**
         * Creates new scramble mutation operator for the specified configuration.
         * 
         * @param configuration The configuration of a genetic algorithm that uses
         *                      this operator.
         * @param mutationRate  The mutation rate.
         */
        public CScrambleMutation(CConfiguration configuration, double mutationRate) : base(configuration, mutationRate)
        {
        }

        /**
         * Invoked when the genetic algorithm is about to be executed.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that uses this operator.
         */
        public void notify(CConfiguration configuration)
        {
            m_randomizer = getConfiguration().getRandomGenerator() ;
            maxSpan      = configuration.getChromosomeSize() ;
            minSpan      = Math.Min(3, maxSpan) ;
        }

        /**
         * Mutates the specified chromosome in a operator-specific way forming 
         * a new chromosome, which will then go to the next generation.
         * 
         * @param chromosome The chromosome to be mutated.
         */
        public override void apply(IChromosome chromosome)
        {
            if (shouldPerformMutation())
            {
                int start, end, span ;

                /* Chooses a random section within the chromosome. */
                getChromosomeSection(out start, out span, minSpan, maxSpan) ;

                end = start + span ;
#if DEBUG_SM
                Console.WriteLine("Scramble Mutation: " + (start + 1) + " - " + (end)) ;
                printOutGenome(chromosome, (int)Math.Ceiling(Math.Log10(chromosome.getSize())), true) ;
#endif
                /* Scrambles the genes. */
                for (int i = 0 ; i < span ; i++)
                {
                    swapGenes(chromosome, m_randomizer.Next(start, end), m_randomizer.Next(start, end)) ;
                }

#if DEBUG_SM
                printOutGenome(chromosome, (int)Math.Ceiling(Math.Log10(chromosome.getSize())), false) ;
#endif
            }
        }
    }
}
