﻿using System ;
using System.Linq ;
using System.Text ;
using System.Globalization ;
using System.Collections.Generic ;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * A serializer for selector genetic operators.
     */
    class CSelectorOperatorSerializer
    {
        /**
         * Returns the System.String representing the serialization
         * of the specified selector operator.
         * 
         * @param selector The operator to be serialized.
         * 
         * @return The string representing the serialization of the
         *         specified selector operator.
         * 
         * @throws ArgumentException If selector is null or unknown
         *                           by this serializer.
         */
        public static string serialize(ISelectionOperator selector)
        {
            if (selector is CRouletteWheelSelector)
            {
                return "RWS" ;
            }
            else if (selector is CRandomTournamentSelector)
            {
                CRandomTournamentSelector ts = (CRandomTournamentSelector)selector ;

                return String.Format(CultureInfo.InvariantCulture, "RTS {0} {1:0.##}", ts.getTournamentSize(), ts.getFitterWinProbability()) ;
            }
            else if (selector is CStochasticUniversalSampling)
            {
                return "SUS" ;
            }
            else if (selector is CDeterministicSampling)
            {
                return "DS" ;
            }
            else if (selector is CRemainderStochasticSampling)
            {
                return String.Format("RSS {0}", (((CRemainderStochasticSampling)selector).isWithReplacement() ? "1" : "0")) ;
            }
            else
            {
                throw new ArgumentException("An unknown (or null) selector operator!") ;
            }
        }

        /**
         * Creates a selector operator from the specified string.
         * 
         * @param serialization The operator's serialization to be
         *                      deserialized.
         * @param configuration The configuration of the genetic
         *                      algorithm that will use the selector.  
         *                      
         * @return The deserialized selector operator.
         *                      
         * @throws ArgumentException If the serialization is invalid,
         *                           or represents no selector.
         */
        public static ISelectionOperator deserialize(string serialization, CConfiguration configuration)
        {
            if (String.Compare(serialization, "RWS", true) == 0)
            {
                return new CRouletteWheelSelector(configuration) ;
            }
            else if (serialization.StartsWith("RTS", true, null))
            {
                String[] parameters = serialization.Split(' ', '\t') ;

                if (parameters.Length != 3)
                {
                    throw new ArgumentException("Invalid, or missing arguments of RTS selector!") ;
                }

                return new CRandomTournamentSelector(configuration, int.Parse(parameters[1]), double.Parse(parameters[2], CultureInfo.InvariantCulture)) ;
            }
            else if (String.Compare(serialization, "SUS", true) == 0)
            {
                return new CStochasticUniversalSampling(configuration) ;
            }
            else if (String.Compare(serialization, "DS", true) == 0)
            {
                return new CDeterministicSampling(configuration) ;
            }
            else if (serialization.StartsWith("RSS", true, null))
            {
                String[] parameters = serialization.Split(' ', '\t') ;

                if (parameters.Length != 2)
                {
                    throw new ArgumentException("Invalid, or missing arguments of RRS selector!") ;
                }

                return new CRemainderStochasticSampling(configuration, parameters[1].Trim().Equals("1")) ;
            }
            else
            {
                throw new ArgumentException("Unknown selector operator!") ;
            }
        }
    }
}
