﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{
    /**
     * Sigma scaling may be used to force a constant selection
     * pressure over many generations. It works like this:
     * First, the average fitness score - averageFitness - is
     * calculated, and then the standard deviation of the fitness 
     * scores - sigma - is calculated. Eventually a new fitness 
     * score for each individual in the specified population is 
     * assigned according to this equation:
     * <em>newFitnessScore = (oldFitnessScore - averageFitness) 
     *                       / (2 * sigma)</em>
     * If the sigma is 0, then all the fitness scores are set to 1.0.
     */
    public class CSigmaScaling : IFitnessScaler
    {
        /**
         * Scales the fitness scores of the individuals 
         * within the specified population.
         * <p>
         * For more information look at the description 
         * of the CSigmaScaling class.
         * </p>
         * 
         * @param population The population of individuals
         *                   to be scaled.
         */
        public void scale(CPopulation population)
        {
            double fitnessAverage = 0.0 ;

            for (int i = 0 ; i < population.getSize() ; i++)
            {
                fitnessAverage += population[i].getFitnessValue() ;
            }

            /* The average fitness score of the population. */
            fitnessAverage /= population.getSize() ;

            double fitnessVariance = 0.0 ;

            for (int i = 0 ; i < population.getSize() ; i++)
            {
                fitnessVariance += (population[i].getFitnessValue() - fitnessAverage) *
                                   (population[i].getFitnessValue() - fitnessAverage) ;
            }

            /* The standard deviation of fitness scores of the population. */
            double sigma = Math.Sqrt(fitnessVariance / population.getSize()) ;

            /* If the sigma is 0, then all the fitness values are set to 1...*/
            if (sigma < 0.000001)
            {
                for (int i = 0 ; i < population.getSize() ; i++)
                {
                    population[i].setFitnessValue(1.0) ;
                }
            }
            else /* ...otherwise the fitness scores are scaled. */
            {
                for (int i = 0 ; i < population.getSize() ; i++)
                {
                    population[i].setFitnessValue((population[i].getFitnessValue() - fitnessAverage) / (2 * sigma)) ;
                }
            }
        }
    }
}