﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.GeneticAlgorithm
{ 
    /** 
     * The stochastic universal sampling selector implementation.
     * 
     * @author Tomas Tunys
     */
    public class CStochasticUniversalSampling : CBaseSelectionOperator
    {
        /**
         * Instantiates a new stochastic universal sampling selector.
         * 
         * @param configuration The configuration of the genetic algorithm
         *                      that will utilize this selector.
         *
         * @throws NullReferenceException If confuguration is <em>null</em>.
         */
        public CStochasticUniversalSampling(CConfiguration configuration) : base(configuration)
        {
        }

        /**
         * Selects a chromosome from the specified population according to its
         * fitness. See the description of  the algorithm in the description
         * of this class.
         * 
          * @param population   The population the chromosome will be selected from.
         * @param matingPool    The mating pool the chromosome will be inserted into.
         * @param toSelectCount The total number of chromosomes to select.
         */
        public override void select(CPopulation population, CPopulation matingPool, int toSelectCount)
        {
            enforcePositiveFitnessScores(population) ;

            double totalFitnessScore = 0.0 ;
            
            for (int i = 0 ; i < population.getSize() ; i++)
            {
                totalFitnessScore += population[i].getFitnessValue() ;
            }

            double pointerGap = totalFitnessScore / toSelectCount ;
            double pointer    = getConfiguration().getRandomGenerator().NextDouble() * pointerGap ;

            double cumulativeFitness = 0 ;
            int    currentGeneIndex  = 0 ;
            
            while (toSelectCount > 0)
            {
                cumulativeFitness += population[currentGeneIndex].getFitnessValue() ;

                while (cumulativeFitness > pointer && toSelectCount-- > 0)
                {
                    matingPool.addChromosome(population[currentGeneIndex]) ;
                    pointer += pointerGap ;
                }

                ++currentGeneIndex ;
            }
        }
    }
}
