﻿using System ;
using System.IO ;
using System.Data ;
using System.Linq ;
using System.Text ;
using System.Drawing ;
using System.Threading ;
using System.Windows.Forms ;
using System.ComponentModel ;
using System.Collections.Generic ;
using MCScheduling.MixedCriticality ;
using MCScheduling.GeneticAlgorithm ;
using MCScheduling.MixedCriticality.GA ;
using MCScheduling.MixedCriticality.SA ;
using MCScheduling.MixedCriticality.DP ;
using MCScheduling.MixedCriticality.MIP ;
using MCScheduling.MixedCriticality.CEDF ;

namespace MCScheduling
{
    /**
     * The main window of the application.
     */
    public partial class MainWindowForm : Form
    {
        /** The name of the last loaded or generated collection of tests. */
        private string m_lastTestCollectionPath ;

        /** Flag to ensure denial of infinite recursion calls when changing probabilities in Random Instance Generator. */
        private bool m_changingProbabilities ;

        #region >>> Test Instances and Solutions Lists <<<

        /** The mixed-criticality instances for testing. */
        private BindingList<CMixedCriticalityInstance> m_testInstancesList ;

        /**
         * The list of solutions for the current mixed-criticality instance
         * created by MIP solver.
         */
        private BindingList<CMixedCriticalityInstance> m_mipSolutionsList ;

        /**
         * The list of solutions for the current mixed-criticality instance
         * created by GA solver.
         */
        private BindingList<CMixedCriticalityInstance> m_gaSolutionsList ;

        /**
         * The list of solutions for the current mixed-criticality instance
         * created by SA solver.
         */
        private BindingList<CMixedCriticalityInstance> m_saSolutionsList ;

        /**
         * The list of solutions for the current mixed-criticality instance
         * created by CEDF solver.
         */
        private BindingList<CMixedCriticalityInstance> m_cedfSolutionsList ;

        /**
         * The list of solutions for the current mixed-criticality instance
         * created by DP solver.
         */
        private BindingList<CMixedCriticalityInstance> m_dpSolutionsList ;

        #endregion >>> Test Instances and Solutions Lists <<<

        #region >>> The Mixed-criticality Solvers <<<

        /** The MIP solver. */
        private CMixedIntegerProgrammingSolver m_mipSolver ;

        /** A backgronud thread used for execution of MIP solver. */
        private CBackgroundThreadSolver m_mipSolverThread ;

        /** The Genetic Algorithm solver. */
        private CGeneticAlgorithmSolver m_gaSolver ;

        /** A backgronud thread used for execution of GA solver. */
        private CBackgroundThreadSolver m_gaSolverThread ;

        /** The Simulated Annealing solver. */
        private CSimulatedAnnealingSolver m_saSolver ;

        /** A backgronud thread used for execution of SA solver. */
        private CBackgroundThreadSolver m_saSolverThread ;

        /** The Clairvoyant EDF solver. */
        private CClairvoyantEDFSolver m_cedfSolver ;

        /** A backgronud thread used for execution of CEDF solver. */
        private CBackgroundThreadSolver m_cedfSolverThread ;

        /** The DP solver. */
        private CDPSolver m_dpSolver ;

        /** A backgronud thread used for execution of DP solver. */
        private CBackgroundThreadSolver m_dpSolverThread ;

        #endregion >>> The Mixed-criticality Solvers <<<

        #region >>> The Solvers' Configurations <<<

        /** The configuration of a genetic algorithm. */
        private CMixedCriticalityConfiguration m_geneticAlgorithmConfiguration ;

        /** The configuration of a genetic algorithm. */
        private CSimulatedAnnealingConfiguration m_simulatedAnnealingConfiguration ;

        #endregion >>> The Solvers' Configurations <<<

        #region >>> The SolverFinishedHandler Definitions <<<<

        /** The handler for the stopping of MIP solver. */
        private CBackgroundThreadSolver.SolverFinishedHandler m_mipSolverStopHandler ;

        /** The handler for the stopping of GA solver. */
        private CBackgroundThreadSolver.SolverFinishedHandler m_gaSolverStopHandler ;

        /** The handler for the stopping of SA solver. */
        private CBackgroundThreadSolver.SolverFinishedHandler m_saSolverStopHandler ;

        /** The handler for the stopping of CEDF solver. */
        private CBackgroundThreadSolver.SolverFinishedHandler m_cedfSolverStopHandler ;

        /** The handler for the stopping of DP solver. */
        private CBackgroundThreadSolver.SolverFinishedHandler m_dpSolverStopHandler ;

        #endregion >>> The SolverFinishedHandler Definitions <<<<
        
        #region >>> The Background Thread Solver Definition <<<

        /** 
         * CSolverThread runs a specified mixed-criticality solver 
         * in the background, and allows the user to abort it 
         * at will.
         */
        private class CBackgroundThreadSolver
        {
            /** A thread in which the solver is executed. */
            private Thread m_thread ;

            /** A solver that will be executed in the separated thread. */
            private CBaseMixedCriticalitySolver m_solver ;

            /** Instances to be solved. */
            private IList<CMixedCriticalityInstance> m_instances ;

            /** Has the thread been aborted? */
            private bool m_aborted ;

            /** The delegate type for hooking up solver finish notification. */
            public delegate void SolverFinishedHandler(object sender, EventArgs e) ;

            /** 
             * The listeners registered with the solver that will be notified when
             * the solver stopes.
             */
            private event SolverFinishedHandler m_notifyWhenFinished ;

            /**
             * Creates a new background solving thread for the specified solver.
             * 
             * @param solver    The solver to be run in a background thread.
             * @param instances The list containing the instances to solve.
             */
            public CBackgroundThreadSolver(CBaseMixedCriticalitySolver solver, IList<CMixedCriticalityInstance> instances, SolverFinishedHandler finishHandler)
            {
                m_solver    = solver ;
                m_instances = instances ;
                m_notifyWhenFinished += finishHandler ;
            }

            /** 
             * Launches the solver in a background thread.
             */
            public void run()
            {
                m_thread  = new Thread(solving) ;
                m_aborted = false ;

                m_thread.Start() ;
            }

            /**
             * Method in which the solver is executed on each of
             * the mixed-criticality instances that has been given.
             */
            private void solving()
            {
                int instanceIndex = 0 ;

                /* Resets all the instances' solution variables. */
                foreach (CMixedCriticalityInstance instance in m_instances)
                {
                    instance.setMakeSpan(0.0) ;
                    instance.setLateness(0.0) ;
                    
                    foreach (CMixedCriticalityJob job in instance)
                    {
                        job.setStartTime(double.NaN) ;
                    }
                }

                try
                {
                    /* Untill all the instances has been solved or the solver has been interrupted. */
                    while (instanceIndex < m_instances.Count && !m_aborted)
                    {
                        /* The mixed-criticality instance to solve. */
                        CMixedCriticalityInstance instance = m_instances[instanceIndex] ;

                        System.Diagnostics.Stopwatch stopwatch = new System.Diagnostics.Stopwatch() ;

                        stopwatch.Start() ;

                        /* Solves the current instance. */
                        m_solver.solve(instance) ;

                        stopwatch.Stop() ;

                        instance.setCPUTime((int)stopwatch.ElapsedMilliseconds) ;

                        /* According to the status of the solution appropriate output is print out. */
                        if (m_solver.getStatus() == CMixedIntegerProgrammingSolver.Status.OPTIMAL)
                        {
                            if (!CMixedCriticalityInstance.validateInstance(instance))
                            {
                                throw new Exception("Solver generated an invalid solution!") ;
                            }
                            instance.MakespanDataView = Math.Round(instance.getMakeSpan(), 0).ToString() ;
                            instance.LatenessDataView = "0" ;
                            instance.SolutionStatusDataView = "optimal" ;
                        }
                        else if (m_solver.getStatus() == CMixedIntegerProgrammingSolver.Status.SUBOPTIMAL)
                        {
                            if (!CMixedCriticalityInstance.validateInstance(instance))
                            {
                                throw new Exception("Solver generated an invalid solution!") ;
                            }
                            instance.MakespanDataView = Math.Round(instance.getMakeSpan(), 0).ToString() ;
                            instance.LatenessDataView = "0" ;
                            instance.SolutionStatusDataView = "suboptimal" ;
                        }
                        else if (m_solver.getStatus() == CMixedIntegerProgrammingSolver.Status.HEURISTIC)
                        {
                            if (!CMixedCriticalityInstance.validateInstance(instance))
                            {
                                throw new Exception("Solver generated an invalid solution!") ;
                            }
                            instance.MakespanDataView = Math.Round(instance.getMakeSpan(), 0).ToString() ;
                            instance.LatenessDataView = Math.Round(instance.getLateness(), 0).ToString() ;
                            instance.SolutionStatusDataView = "heuristic" ;
                        }
                        else if (m_solver.getStatus() == CMixedIntegerProgrammingSolver.Status.INFEASIBLE)
                        {
                            instance.MakespanDataView = "unknown" ;
                            instance.LatenessDataView = "unknown" ;
                            instance.SolutionStatusDataView = "infeasible" ;
                        }
                        else if (m_solver.getStatus() == CMixedIntegerProgrammingSolver.Status.TIME_LIMIT)
                        {
                            instance.MakespanDataView = "unknown" ;
                            instance.LatenessDataView = "unknown" ;
                            instance.SolutionStatusDataView = "time limit" ;
                        }
                        else if (m_solver.getStatus() == CMixedIntegerProgrammingSolver.Status.ITERATION_LIMIT)
                        {
                            instance.MakespanDataView = "unknown" ;
                            instance.LatenessDataView = "unknown" ;
                            instance.SolutionStatusDataView = "iteration limit" ;
                        }
                        else 
                        {
                            instance.MakespanDataView = "unknown" ;
                            instance.LatenessDataView = "unknown" ;
                            instance.SolutionStatusDataView = "aborted" ;
                        }

                        /* Forces the data grid view component to update the row holding this solution. */
                        m_instances[instanceIndex] = instance ;

                        ++instanceIndex ;
                    }
                }
                catch(Exception exception)
                {
                    MessageBox.Show(exception.Message, "Solver", MessageBoxButtons.OK, MessageBoxIcon.Error) ;
                }
                finally
                {
                    /* Executed in case the solver has been aborted. */
                    while (instanceIndex < m_instances.Count)
                    {
                        m_instances[instanceIndex].MakespanDataView = "Unknown" ;
                        m_instances[instanceIndex].LatenessDataView = "" ;

                        /* Forces the data grid view component to update the row holding this solution. */
                        m_instances[instanceIndex] = m_instances[instanceIndex] ;

                        ++instanceIndex ;
                    }
                }

                m_thread = null ;

                m_notifyWhenFinished(this, EventArgs.Empty) ;
            }

            /**
             * Aborts the exectuon of the current solver.
             */
            public void abort()
            {
                m_aborted = true ;

                m_solver.interrupt() ;
            }
        }

        #endregion >>> The Background Thread Solver Definition <<<

        #region >>> Main Window Form Constructor <<<

        /**
         * Creates and initializes the components of
         * the main window.
         */
        public MainWindowForm()
        {
            InitializeComponent() ;

            testInstancesTable.AutoGenerateColumns = false ;
            mipSolutionsTable.AutoGenerateColumns = false ;
            gaSolutionsTable.AutoGenerateColumns = false ;
            saSolutionsTable.AutoGenerateColumns = false ;
            cedfSolutionsTable.AutoGenerateColumns = false ;
            dpSolutionsTable.AutoGenerateColumns = false ;

            /* Creates an empty list of mixed-criticality instances. */
            m_testInstancesList = new BindingList<CMixedCriticalityInstance>() ;

            /* Sets the test instances list as the data source for the table. */
            testInstancesTable.DataSource = m_testInstancesList ;

            /* Cretes the MIP Solver. */
            m_mipSolver = new CMixedIntegerProgrammingSolver() ;

            /* Creates an empty list of solutions for MIP solver. */
            m_mipSolutionsList = new BindingList<CMixedCriticalityInstance>() ;

            /* Sets the list of MIP solutions as the data source for the table. */
            mipSolutionsTable.DataSource = m_mipSolutionsList ;

            /* When the MIP solver finishes that it simulates stop button click. */
            m_mipSolverStopHandler = new CBackgroundThreadSolver.SolverFinishedHandler(mipStopButtonClick) ;

            /* Creates the MIP solver thread. */
            m_mipSolverThread = new CBackgroundThreadSolver(m_mipSolver, m_mipSolutionsList, m_mipSolverStopHandler) ;

            /* Creates untitled genetic algorithm configuration. */
            m_geneticAlgorithmConfiguration = new CMixedCriticalityConfiguration("untitled") ;

            /* Creates the GA Solver. */
            m_gaSolver = new CGeneticAlgorithmSolver(m_geneticAlgorithmConfiguration) ;

            /* Creates an empty list of solutions for GA solver. */
            m_gaSolutionsList = new BindingList<CMixedCriticalityInstance>() ;

            /* Sets the list of GA solutions as the data source for the table. */
            gaSolutionsTable.DataSource = m_gaSolutionsList ;

            /* When the GA solver finishes it simulates stop button being pressed. */
            m_gaSolverStopHandler = new CBackgroundThreadSolver.SolverFinishedHandler(gaStopButtonClick) ;

            /* Creates the MIP solver thread. */
            m_gaSolverThread = new CBackgroundThreadSolver(m_gaSolver, m_gaSolutionsList, m_gaSolverStopHandler) ;

            /* Creates the SA Solver. */
            m_saSolver = new CSimulatedAnnealingSolver() ;

            /* Creates configuration for simulated annealing solver. */
            m_simulatedAnnealingConfiguration = new CSimulatedAnnealingConfiguration() ;

            /* Creates an empty list of solutions for SA solver. */
            m_saSolutionsList = new BindingList<CMixedCriticalityInstance>() ;

            /* Sets the list of SA solutions as the data source for the table. */
            saSolutionsTable.DataSource = m_saSolutionsList ;

            /* When the GA solver finishes it simulates stop button being pressed. */
            m_saSolverStopHandler = new CBackgroundThreadSolver.SolverFinishedHandler(saStopButtonClick) ;

            /* Creates the MIP solver thread. */
            m_saSolverThread = new CBackgroundThreadSolver(m_saSolver, m_saSolutionsList, m_saSolverStopHandler) ;

            /* Creates the CEDF Solver. */
            m_cedfSolver = new CClairvoyantEDFSolver() ;        
            
            /* Creates an empty list of solutions for CEDF solver. */
            m_cedfSolutionsList = new BindingList<CMixedCriticalityInstance>() ;

            /* Sets the list of CEDF solutions as the data source for the table. */
            cedfSolutionsTable.DataSource = m_cedfSolutionsList ;

            /* When the CEDF solver finishes it simulates stop button being pressed. */
            m_cedfSolverStopHandler = new CBackgroundThreadSolver.SolverFinishedHandler(cedfStopButtonClick) ;

            /* Creates the MIP solver thread. */
            m_cedfSolverThread = new CBackgroundThreadSolver(m_cedfSolver, m_cedfSolutionsList, m_cedfSolverStopHandler) ;

            /* Creates the DP Solver. */
            m_dpSolver = new CDPSolver() ;        
            
            /* Creates an empty list of solutions for DP solver. */
            m_dpSolutionsList = new BindingList<CMixedCriticalityInstance>() ;

            /* Sets the list of DP solutions as the data source for the table. */
            dpSolutionsTable.DataSource = m_dpSolutionsList ;

            /* When the DP solver finishes it simulates stop button being pressed. */
            m_dpSolverStopHandler = new CBackgroundThreadSolver.SolverFinishedHandler(dpStopButtonClick) ;

            /* Creates the MIP solver thread. */
            m_dpSolverThread = new CBackgroundThreadSolver(m_dpSolver, m_dpSolutionsList, m_dpSolverStopHandler) ;

            /** Fills the checklists for genetic operators. */
            fillCrossoverOperatorChecklist() ;
            fillMutationOperatorChecklist() ;
        }

        #endregion >>> Main Window Form Constructor <<<

        #region >>> Test Instance Generator Tab <<<

        /**
         * Invoked when the number of jobs of some criticality
         * level has changed - by using NumericUpDown control.
         */
        private void criticalityJobsCountValueChanged(object sender, EventArgs e)
        {
            int maxCriticality = 0 ;

            maxCriticality = (criticalityOneJobsCount.Value > 0 ? 1 : maxCriticality) ;
            maxCriticality = (criticalityTwoJobsCount.Value > 0 ? 2 : maxCriticality) ;
            maxCriticality = (criticalityThreeJobsCount.Value > 0 ? 3 : maxCriticality) ;
            maxCriticality = (criticalityFourJobsCount.Value > 0 ? 4 : maxCriticality) ;

            decimal jobsCount = criticalityOneJobsCount.Value + criticalityTwoJobsCount.Value + criticalityThreeJobsCount.Value + criticalityFourJobsCount.Value ;

            currentMaxCriticalityLabel.Text = maxCriticality.ToString() ;
            currentJobsCountLabel.Text      = jobsCount.ToString() ;
        }

        /**
         * Invoked when worst-case execution times has changed - by using 
         * NumericUpDown control.
         */
        private void maximumExecutionTimesValueChanged(object sender, EventArgs e)
        {
            /* This whole machinary just ensures that the worst-case execution
             * times are non-decreasing with increasing criticality.
             */
            decimal one   = maxLvlOneWCETUpDown.Value ;
            decimal two   = maxLvlTwoWCETUpDown.Value ;
            decimal three = maxLvlThreeWCETUpDown.Value ;
            decimal four  = maxLvlFourWCETUpDown.Value ;

            if (one > two)
            {
                maxLvlTwoWCETUpDown.Value = one ;
            }

            if (two > three)
            {
                maxLvlThreeWCETUpDown.Value = two ;
            }

            if (three > four)
            {
                maxLvlFourWCETUpDown.Value = three ;
            }
        }

        /**
         * Invoked when the button "Generate!" of feasible instance
         * generator has been pressed. First the input data are
         * validated, if they are correct a new collection of mixed-
         * criticality instances is generated, otherwise, a message 
         * box with warning is displayed.
         */ 
        private void generateButtonClick(object sender, EventArgs e)
        {
            generateButton.Enabled = false ;

            string collectionName = collectionNameTextBox.Text.Trim() ;

            if (collectionName.Length == 0)
            {
                MessageBox.Show(this, "The collection name has to be specified!", "Collection Name", MessageBoxButtons.OK) ;
                collectionNameTextBox.Focus() ;
                generateButton.Enabled = true ;
                return ;
            }

            int instanceCount = (int)instanceCountUpDown.Value ;

            int jobCount = int.Parse(currentJobsCountLabel.Text) ;

            if (jobCount == 0)
            {
                MessageBox.Show(this, "There has to be at least on job per instance!", "Job Count", MessageBoxButtons.OK) ;
                criticalityOneJobsCount.Focus() ;
                generateButton.Enabled = true ;
                return ;
            }

            int[] jobsPerCriticality = new int[4] ;

            jobsPerCriticality[0] = (int)criticalityOneJobsCount.Value ;
            jobsPerCriticality[1] = (int)criticalityTwoJobsCount.Value ;
            jobsPerCriticality[2] = (int)criticalityThreeJobsCount.Value ;
            jobsPerCriticality[3] = (int)criticalityFourJobsCount.Value ;

            int[] maxExecutionTimes = new int[4] ;

            maxExecutionTimes[0] = (int)maxLvlOneWCETUpDown.Value ;
            maxExecutionTimes[1] = (int)maxLvlTwoWCETUpDown.Value ;
            maxExecutionTimes[2] = (int)maxLvlThreeWCETUpDown.Value ;
            maxExecutionTimes[3] = (int)maxLvlFourWCETUpDown.Value ;

            int maxIdleTime = (int)maxIdleTimeUpDown.Value ;

            double timeWindowSpan = (enableWindowSpanCheckbox.Checked ? (double)windowSpanFactor.Value : double.NaN) ;

            m_testInstancesList.Clear() ;
 
            foreach (CMixedCriticalityInstance instance in CDefaultMixedCriticalityInstanceGenerator.generate(collectionName, instanceCount, jobsPerCriticality, maxExecutionTimes, maxIdleTime, timeWindowSpan))
            {
                m_testInstancesList.Add(instance) ;
            }
            
            updateTestInstancesLists() ;

            generateButton.Enabled = true ;
        }

        /**
         * Invoked when the button "Generate!" of worst-case instance
         * generator has been pressed. First the input data are
         * validated, if they are correct a new collection of mixed-
         * criticality instances is generated, otherwise, a message 
         * box with warning is displayed.
         */ 
        private void generateWorstCaseButton_Click(object sender, EventArgs e)
        {
            generateWorstCaseButton.Enabled = false ;

            string collectionName = collectionNameTextBox.Text.Trim() ;

            if (collectionName.Length == 0)
            {
                MessageBox.Show(this, "The collection name has to be specified!", "Collection Name", MessageBoxButtons.OK) ;
                collectionNameTextBox.Focus() ;
                generateWorstCaseButton.Enabled = true ;
                return ;
            }

            int instanceCount = (int)instanceCountUpDown.Value ;

            int numberOfClusters = (int)clusterCountUpDown.Value ;

            double clusterMakespan = (double)clusterMakespanUpDown.Value - 1.0 ;

            m_testInstancesList.Clear() ;
 
            foreach (CMixedCriticalityInstance instance in CWorstCaseMixedCriticalityInstanceGenerator.generate(collectionName, instanceCount, numberOfClusters, clusterMakespan))
            {
                m_testInstancesList.Add(instance) ;
            }
            
            updateTestInstancesLists() ;

            generateWorstCaseButton.Enabled = true ;
        }

        /**
         * Invoked when the "Load instances..." button has been pressed.
         */
        private void loadInstancesButtonClick(object sender, EventArgs e)
        {
            if (loadInstancesDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    m_testInstancesList.Clear() ;

                    foreach (CMixedCriticalityInstance instance in CMixedCriticalityInstancesSerializer.loadCollection(loadInstancesDialog.SelectedPath))
                    {
                        m_testInstancesList.Add(instance) ;
                    }

                    updateTestInstancesLists() ;

                    testInstancesTable.Refresh() ;

                    m_lastTestCollectionPath = loadInstancesDialog.SelectedPath ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error has occured during the process!", "Error", MessageBoxButtons.OK) ;
                }
            }
        }

        /**
         * Invoked when the "Save instances..." button has been pressed.
         */
        private void saveInstancesButtonClick(object sender, EventArgs e)
        {
            if (saveInstancesDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    CMixedCriticalityInstancesSerializer.saveCollection(saveInstancesDialog.SelectedPath, collectionNameTextBox.Text, m_testInstancesList) ;

                    updateTestInstancesLists() ;

                    m_lastTestCollectionPath = loadInstancesDialog.SelectedPath ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error has occured during the process!", "Error", MessageBoxButtons.OK) ;
                }
            }
        }

        /**
         * Restores the test instances data grid views nad list for the solvers.
         * This method should be called whenever the list of test instances 
         * has changed.
         */
        private void updateTestInstancesLists()
        {            
            m_mipSolutionsList.Clear() ;
            m_gaSolutionsList.Clear() ;
            m_saSolutionsList.Clear() ;
            m_cedfSolutionsList.Clear() ;
            m_dpSolutionsList.Clear() ;

            foreach (CMixedCriticalityInstance testInstance in m_testInstancesList)
            {
                m_mipSolutionsList.Add((CMixedCriticalityInstance)testInstance.Clone()) ;
                m_gaSolutionsList.Add((CMixedCriticalityInstance)testInstance.Clone()) ;
                m_saSolutionsList.Add((CMixedCriticalityInstance)testInstance.Clone()) ;
                m_cedfSolutionsList.Add((CMixedCriticalityInstance)testInstance.Clone()) ;
                m_dpSolutionsList.Add((CMixedCriticalityInstance)testInstance.Clone()) ;
            }
        }

        private void maxWindowSpanCheckboxCheckedChanged(object sender, EventArgs e)
        {
            windowSpanFactor.Enabled = enableWindowSpanCheckbox.Checked ;
        }

        private void criticalityLevelProbabilityUpDownValueChanged(object sender, EventArgs e)
        {
            int     changedLevel    = -1 ;
            decimal restProbability = decimal.One ;
            decimal sumOfOtherProb  = decimal.Zero ;

            if (!m_changingProbabilities)
            {
                m_changingProbabilities = true ;

                if (object.ReferenceEquals(sender, critLvl1ProbUpDown))
                {
                    changedLevel     = 1 ;
                    restProbability -= critLvl1ProbUpDown.Value ;
                    sumOfOtherProb   = critLvl2ProbUpDown.Value + critLvl3ProbUpDown.Value + critLvl4ProbUpDown.Value ;
                }
                else if (object.ReferenceEquals(sender, critLvl2ProbUpDown))
                {
                    changedLevel     = 2 ;
                    restProbability -= critLvl2ProbUpDown.Value ;
                    sumOfOtherProb   = critLvl1ProbUpDown.Value + critLvl3ProbUpDown.Value + critLvl4ProbUpDown.Value ;
                }
                else if (object.ReferenceEquals(sender, critLvl3ProbUpDown))
                {
                    changedLevel     = 3 ;
                    restProbability -= critLvl3ProbUpDown.Value ;
                    sumOfOtherProb   = critLvl1ProbUpDown.Value + critLvl2ProbUpDown.Value + critLvl4ProbUpDown.Value ;
                }
                else
                {
                    changedLevel     = 4 ;
                    restProbability -= critLvl4ProbUpDown.Value ;
                    sumOfOtherProb   = critLvl1ProbUpDown.Value + critLvl2ProbWCETUpDown.Value + critLvl3ProbUpDown.Value ;
                }

                switch (changedLevel)
                {
                case 1:
                    if (sumOfOtherProb > (decimal)10e-2)
                    {
                        critLvl2ProbUpDown.Value = (restProbability * critLvl2ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl3ProbUpDown.Value = (restProbability * critLvl3ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl4ProbUpDown.Value = (restProbability * critLvl4ProbUpDown.Value) / sumOfOtherProb ;
                    }
                    else
                    {
                        critLvl2ProbUpDown.Value = restProbability ;
                        critLvl3ProbUpDown.Value = 0 ;
                        critLvl4ProbUpDown.Value = 0 ;
                    }
                    break ;

                case 2:
                    if (sumOfOtherProb > (decimal)10e-2)
                    {
                        critLvl1ProbUpDown.Value = (restProbability * critLvl1ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl3ProbUpDown.Value = (restProbability * critLvl3ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl4ProbUpDown.Value = (restProbability * critLvl4ProbUpDown.Value) / sumOfOtherProb ;
                    }
                    else
                    {
                        critLvl1ProbUpDown.Value = restProbability ;
                        critLvl3ProbUpDown.Value = 0 ;
                        critLvl4ProbUpDown.Value = 0 ;
                    }
                    break ;

                case 3:
                    if (sumOfOtherProb > (decimal)10e-2)
                    {
                        critLvl1ProbUpDown.Value = (restProbability * critLvl1ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl2ProbUpDown.Value = (restProbability * critLvl2ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl4ProbUpDown.Value = (restProbability * critLvl4ProbUpDown.Value) / sumOfOtherProb ;
                    }
                    else
                    {
                        critLvl1ProbUpDown.Value = restProbability ;
                        critLvl2ProbUpDown.Value = 0 ;
                        critLvl4ProbUpDown.Value = 0 ;
                    }
                    break ;

                case 4:
                    if (sumOfOtherProb > (decimal)10e-2)
                    {
                        critLvl1ProbUpDown.Value = (restProbability * critLvl1ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl2ProbUpDown.Value = (restProbability * critLvl2ProbUpDown.Value) / sumOfOtherProb ;
                        critLvl3ProbUpDown.Value = (restProbability * critLvl3ProbUpDown.Value) / sumOfOtherProb ;
                    }
                    else
                    {
                        critLvl1ProbUpDown.Value = restProbability ;
                        critLvl2ProbUpDown.Value = 0 ;
                        critLvl3ProbUpDown.Value = 0 ;
                    }
                    break ;

                default:
                    break ;
                }

                m_changingProbabilities = false ;
            }
        }

        /**
         * Invoked when worst-case execution times has changed - by using 
         * NumericUpDown control.
         */
        private void criticalityLevelProbabilityWCETUpDownValueChanged(object sender, EventArgs e)
        {
            /* This whole machinary just ensures that the worst-case execution
             * times are non-decreasing with increasing criticality.
             */
            decimal one   = critLvl1ProbWCETUpDown.Value ;
            decimal two   = critLvl2ProbWCETUpDown.Value ;
            decimal three = critLvl3ProbWCETUpDown.Value ;
            decimal four  = critLvl4ProbWCETUpDown.Value ;

            if (one > two)
            {
                critLvl2ProbWCETUpDown.Value = one ;
            }

            if (two > three)
            {
                critLvl3ProbWCETUpDown.Value = two ;
            }

            if (three > four)
            {
                critLvl4ProbWCETUpDown.Value = three ;
            }
        }

        private void GenerateRandomButtonClick(object sender, EventArgs e)
        {
            generateButton.Enabled = false ;

            string collectionName = collectionNameTextBox.Text.Trim() ;

            if (collectionName.Length == 0)
            {
                MessageBox.Show(this, "The collection name has to be specified!", "Collection Name", MessageBoxButtons.OK) ;
                collectionNameTextBox.Focus() ;
                generateButton.Enabled = true ;
                return ;
            }

            int instanceCount = (int)instanceCountUpDown.Value ;

            int jobsCount = (int)jobsCountUpDown.Value ;

            double[] jobsPerCriticalityProb = new double[4] ;

            jobsPerCriticalityProb[0] = (double)critLvl1ProbUpDown.Value ;
            jobsPerCriticalityProb[1] = (double)critLvl2ProbUpDown.Value ;
            jobsPerCriticalityProb[2] = (double)critLvl3ProbUpDown.Value ;
            jobsPerCriticalityProb[3] = (double)critLvl4ProbUpDown.Value ;

            int[] maxExecutionTimes = new int[4] ;

            maxExecutionTimes[0] = (int)critLvl1ProbWCETUpDown.Value ;
            maxExecutionTimes[1] = (int)critLvl2ProbWCETUpDown.Value ;
            maxExecutionTimes[2] = (int)critLvl3ProbWCETUpDown.Value ;
            maxExecutionTimes[3] = (int)critLvl4ProbWCETUpDown.Value ;

            int releaseTime = (int)releaseTimeSpanUpDown.Value ;

            int deadline = (int)deadlineSpanUpDown.Value ;

            m_testInstancesList.Clear() ;
 
            foreach (CMixedCriticalityInstance instance in CRandomInstanceGenerator.generate(collectionName, instanceCount, jobsCount, jobsPerCriticalityProb, maxExecutionTimes, releaseTime, deadline))
            {
                m_testInstancesList.Add(instance) ;
            }
            
            updateTestInstancesLists() ;

            generateButton.Enabled = true ;
        }

        #endregion >>> Test Instance Generator Tab <<<

        #region >>> The Genetic Algorithm Configuration Tab <<<

        #region >>> Elitism Checkbox Control Handler <<<
        
        /**
         * Invoked when the elitism checkbox's state has changed.
         */
        private void elitismCheckboxCheckedChanged(object sender, EventArgs e)
        {
            elitismUpDown.Enabled = elitismCheckbox.Checked ;
        }

        #endregion >>> Elitism Checkbox Control Handler <<<

        #region >>> Fitness Function Factors Controls Handlers <<<

        /**
         * Invoked when the makespan factor value has changed.
         */
        private void makespanFactorUpDownValueChanged(object sender, EventArgs e)
        {
            latenessFactorUpDown.Value = (decimal)1.0 - makespanFactorUpDown.Value ;
        }

        /**
         * Invoked when the lateness factor value has changed.
         */
        private void latenessFactorUpDownValueChanged(object sender, EventArgs e)
        {
            makespanFactorUpDown.Value = (decimal)1.0 - latenessFactorUpDown.Value ;
        }

        #endregion >>> Fitness Function Factors Controls Handlers <<<

        #region >>> Selector Operators Configuration Controls <<<

        /**
         * Invoked when the state of the Random Tournament Selection radio button
         * has changed.
         */
        private void selectorRTSRadioButtonCheckedChanged(object sender, EventArgs e)
        {
            tournamentSizeUpDown.Enabled = selectorRTSRadioButton.Checked ;
            tournamentWinProbUpDown.Enabled = selectorRTSRadioButton.Checked ;
        }

        #endregion >>> Selector Operators Configuration Controls <<<

        #region >>> Fitness Scaling Configuration Controls <<<

        /**
         * Invoked when the state of Boltzmann Scaling radio button has changed.
         */
        private void scalingBoltzmannRadioButtonCheckedChanged(object sender, EventArgs e)
        {
            boltzmannStartTempUpDown.Enabled = scalingBoltzmannRadioButton.Checked ;
            boltzmannMinTempUpDown.Enabled = scalingBoltzmannRadioButton.Checked ;
            boltzmannDropStepUpDown.Enabled = scalingBoltzmannRadioButton.Checked ;
        }

        #endregion >>> Fitness Scaling Configuration Controls <<<

        #region >>> Crossover Operators Configuration Controls <<<

        /**
         * Fills the crossover operator checklist with items
         * representing each of the implemented crossover
         * operator.
         */
        private void fillCrossoverOperatorChecklist()
        {
            ListBox.ObjectCollection crossoverItems = new ListBox.ObjectCollection(crossoverOperatorChecklist) ;

            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Order Crossover", CCrossoverOperatorChecklistItem.Type.OrderCrossover, true)) ;
            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Order-Based Crossover", CCrossoverOperatorChecklistItem.Type.OrderBasedCrossover, true)) ;
            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Position-Based Crossover", CCrossoverOperatorChecklistItem.Type.PositionBasedCrossover, true)) ;
            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Cycle Crossover", CCrossoverOperatorChecklistItem.Type.CycleCrossover, false)) ;
            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Partially Mapped Crossover", CCrossoverOperatorChecklistItem.Type.PartiallyMappedCrossover, false)) ;
            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Alternating Position Crossover", CCrossoverOperatorChecklistItem.Type.AlternatingPositionCrossover, false)) ;
            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Heuristic Crossover", CCrossoverOperatorChecklistItem.Type.HeuristicCrossover, false)) ;
            crossoverItems.Add(new CCrossoverOperatorChecklistItem("Greedy Crossover", CCrossoverOperatorChecklistItem.Type.GreedyCrossover, false)) ;

            crossoverOperatorChecklist.Items.AddRange(crossoverItems) ;
        }

        /**
         * Invoked when a selection in crossover operator checklist has changed.
         */
        private void crossoverOperatorChecklistSelectedIndexChanged(object sender, EventArgs e)
        {
            crossoverMoveUpButton.Enabled = crossoverOperatorChecklist.SelectedIndex > 0 ;
            crossoverMoveDownButton.Enabled = crossoverOperatorChecklist.SelectedIndex < crossoverOperatorChecklist.Items.Count - 1 ;

            /* Is an operator selected? */
            if (crossoverOperatorChecklist.SelectedIndex >= 0)
            {
                CCrossoverOperatorChecklistItem item = (CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.SelectedItem ;

                crossoverRateUpDown.Enabled = true ;
                crossoverRateUpDown.Value   = (decimal)item.getCrossoverRate() ;

                if ((crossoverRatioUpDown.Enabled = item.getNeedCrossoverRatio()))
                {
                    crossoverRatioUpDown.Value = (decimal)item.getCrossoverRatio() ;
                }                
            }
            else
            {
                crossoverRateUpDown.Enabled  = false ;
                crossoverRatioUpDown.Enabled = false ;
            }
        }

        /**
         * Invoked when the selected crossover operator should be moved up
         * the checklist. If no operator is selected, then nothing happens.
         */
        private void crossoverMoveUpButtonClick(object sender, EventArgs e)
        {
            if (crossoverOperatorChecklist.SelectedIndex > 0)
            {
                CheckState prevItemState = crossoverOperatorChecklist.GetItemCheckState(crossoverOperatorChecklist.SelectedIndex - 1) ;
                CheckState currItemState = crossoverOperatorChecklist.GetItemCheckState(crossoverOperatorChecklist.SelectedIndex) ;

                object tempItem = crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex] ;
                crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex] = crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex - 1] ;
                crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex - 1] = tempItem ;

                crossoverOperatorChecklist.SetItemCheckState(crossoverOperatorChecklist.SelectedIndex - 1, currItemState) ;
                crossoverOperatorChecklist.SetItemCheckState(crossoverOperatorChecklist.SelectedIndex, prevItemState) ;

                --crossoverOperatorChecklist.SelectedIndex ;

                crossoverOperatorChecklist.Refresh() ;
            }
        }

        /**
         * Invoked when the selected crossover operator should be moved down
         * the checklist. If no operator is selected, then nothing happens.
         */
        private void crossoverMoveDownButtonClick(object sender, EventArgs e)
        {
            if (crossoverOperatorChecklist.SelectedIndex < crossoverOperatorChecklist.Items.Count - 1)
            {
                CheckState nextItemState = crossoverOperatorChecklist.GetItemCheckState(crossoverOperatorChecklist.SelectedIndex + 1) ;
                CheckState currItemState = crossoverOperatorChecklist.GetItemCheckState(crossoverOperatorChecklist.SelectedIndex) ;

                object tempItem = crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex] ;
                crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex] = crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex + 1] ;
                crossoverOperatorChecklist.Items[crossoverOperatorChecklist.SelectedIndex + 1] = tempItem ;

                crossoverOperatorChecklist.SetItemCheckState(crossoverOperatorChecklist.SelectedIndex + 1, currItemState) ;
                crossoverOperatorChecklist.SetItemCheckState(crossoverOperatorChecklist.SelectedIndex, nextItemState) ;

                ++crossoverOperatorChecklist.SelectedIndex ;

                crossoverOperatorChecklist.Refresh() ;
            }
        }

        /**
         * Invoked when the crossover rate has been changed.
         */
        private void crossoverRateUpDownValueChanged(object sender, EventArgs e)
        {
            ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.SelectedItem).setCrossoverRate(decimal.ToDouble(crossoverRateUpDown.Value)) ;
        }

        /**
         * Invoked when the crossover ratio has been changed.
         */
        private void crossoverRatioUpDownValueChanged(object sender, EventArgs e)
        {
            ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.SelectedItem).setCrossoverRatio(decimal.ToDouble(crossoverRatioUpDown.Value)) ;
        }

        #region CCrossoverOperatorChecklistItem class definition.

        /**
         * A representation of a crossover operator item
         * in the crossover operator checklist.
         */
        private class CCrossoverOperatorChecklistItem
        {
            /** The crossover operator ID. */
            public enum Type { OrderCrossover, OrderBasedCrossover, PositionBasedCrossover, CycleCrossover, PartiallyMappedCrossover, AlternatingPositionCrossover, HeuristicCrossover, GreedyCrossover } ;

            /** The name of the operator as it appears in the list. */
            private string m_name ;

            /** The type of the operator. */
            private Type m_type ;

            /** The crossover rate for the operator. */
            private double m_crossoverRate ;

            /** The crossover ratio for the operator (if needed). */
            private double m_crossoverRatio ;

            /** Flag for some operators as they need a crossover ratio. */
            private bool m_needCrossoverRatio ;

            /**
             * Creates a new crossover operator checklist item.
             * 
             * @param name The name of the operator as it appears
             *             in the checklist.
             * @param type The ID type of the operator.
             * @param needCrossoverRatio True if the created operator
             *                           uses the crossover ratio.
             */
            public CCrossoverOperatorChecklistItem(string name, Type type, bool needCrossoverRatio)
            {
                m_name = name ;
                m_type = type ;
                m_crossoverRate = 0.7 ;
                m_crossoverRatio = 0.5 ;
                m_needCrossoverRatio = needCrossoverRatio ;
            }

            /**
             * Creates a crossover checklist item of the specified
             * type.
             * 
             * @param type The type of the crossover operator.
             */
            public CCrossoverOperatorChecklistItem(Type type)
            {
                m_type = type ;
            }

            public override bool Equals(object obj)
            {
                return (m_type == ((CCrossoverOperatorChecklistItem)obj).getType()) ;
            }

            public override int GetHashCode()
            {
                return m_type.GetHashCode() ;
            }

            public void setNeedCrossoverRatio(bool needRatio)
            {
                m_needCrossoverRatio = needRatio ;
            }

            public bool getNeedCrossoverRatio()
            {
                return m_needCrossoverRatio ;
            }

            public void setCrossoverRatio(double ratio)
            {
                m_crossoverRatio = ratio ;
            }

            public double getCrossoverRatio()
            {
                return m_crossoverRatio ;
            }

            public void setCrossoverRate(double rate)
            {
                m_crossoverRate = rate ;
            }

            public double getCrossoverRate()
            {
                return m_crossoverRate ;
            }

            public Type getType()
            {
                return m_type ;
            }

            public override string ToString()
            {
                return m_name ;
            }
        }
        #endregion

        #endregion >>> Crossover Operators Configuration Controls <<<

        #region >>> Mutation Operators Configuration Controls <<<

        /**
         * Fills the mutation operator checklist with items
         * representing each of the implemented mutation
         * operator.
         */
        private void fillMutationOperatorChecklist()
        {
            ListBox.ObjectCollection mutationItems = new ListBox.ObjectCollection(crossoverOperatorChecklist) ;

            mutationItems.Add(new CMutationOperatorChecklistItem("Insertion Mutation", CMutationOperatorChecklistItem.Type.InsertionMutation)) ;
            mutationItems.Add(new CMutationOperatorChecklistItem("Inversion Mutation", CMutationOperatorChecklistItem.Type.InversionMutation)) ;
            mutationItems.Add(new CMutationOperatorChecklistItem("Exchange Mutation", CMutationOperatorChecklistItem.Type.ExchangeMutation)) ;
            mutationItems.Add(new CMutationOperatorChecklistItem("Scramble Mutation", CMutationOperatorChecklistItem.Type.ScrambleMutation)) ;
            mutationItems.Add(new CMutationOperatorChecklistItem("Displacement Mutation", CMutationOperatorChecklistItem.Type.DisplacementMutation)) ;
            mutationItems.Add(new CMutationOperatorChecklistItem("Displaced Inversion Mutation", CMutationOperatorChecklistItem.Type.DisplacedInversionMutation)) ;

            mutationOperatorChecklist.Items.AddRange(mutationItems) ;
        }

        /**
         * Invoked when a selection in mutation operator checklist has changed.
         */
        private void mutationOperatorChecklistSelectedIndexChanged(object sender, EventArgs e)
        {
            mutationMoveUpButton.Enabled = mutationOperatorChecklist.SelectedIndex > 0 ;
            mutationMoveDownButton.Enabled = mutationOperatorChecklist.SelectedIndex < mutationOperatorChecklist.Items.Count - 1 ;

            /* Is an operator selected? */
            if ((mutationRateUpDown.Enabled = (mutationOperatorChecklist.SelectedIndex >= 0)))
            {
                mutationRateUpDown.Value = (decimal)((CMutationOperatorChecklistItem)mutationOperatorChecklist.SelectedItem).getMutationRate() ;
            }
        }

        /**
         * Invoked when the selected mutation operator should be moved up
         * the checklist. If no operator is selected, then nothing happens.
         */
        private void mutationUpButtonClick(object sender, EventArgs e)
        {
            if (mutationOperatorChecklist.SelectedIndex > 0)
            {
                CheckState prevItemState = mutationOperatorChecklist.GetItemCheckState(mutationOperatorChecklist.SelectedIndex - 1) ;
                CheckState currItemState = mutationOperatorChecklist.GetItemCheckState(mutationOperatorChecklist.SelectedIndex) ;

                object tempItem = mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex] ;
                mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex] = mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex - 1] ;
                mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex - 1] = tempItem ;

                mutationOperatorChecklist.SetItemCheckState(mutationOperatorChecklist.SelectedIndex - 1, currItemState) ;
                mutationOperatorChecklist.SetItemCheckState(mutationOperatorChecklist.SelectedIndex, prevItemState) ;

                --mutationOperatorChecklist.SelectedIndex ;

                mutationOperatorChecklist.Refresh() ;
            }
        }

        /**
         * Invoked when the selected mutation operator should be moved down
         * the checklist. If no operator is selected, then nothing happens.
         */
        private void mutationDownButtonClick(object sender, EventArgs e)
        {
            if (mutationOperatorChecklist.SelectedIndex < mutationOperatorChecklist.Items.Count - 1)
            {
                CheckState nextItemState = mutationOperatorChecklist.GetItemCheckState(mutationOperatorChecklist.SelectedIndex + 1) ;
                CheckState currItemState = mutationOperatorChecklist.GetItemCheckState(mutationOperatorChecklist.SelectedIndex) ;

                object tempItem = mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex] ;
                mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex] = mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex + 1] ;
                mutationOperatorChecklist.Items[mutationOperatorChecklist.SelectedIndex + 1] = tempItem ;

                mutationOperatorChecklist.SetItemCheckState(mutationOperatorChecklist.SelectedIndex + 1, currItemState) ;
                mutationOperatorChecklist.SetItemCheckState(mutationOperatorChecklist.SelectedIndex, nextItemState) ;

                ++mutationOperatorChecklist.SelectedIndex ;

                mutationOperatorChecklist.Refresh() ;
            }
        }

        /**
         * Invoked when the mutation rate has been changed.
         */
        private void mutationRateUpDownValueChanged(object sender, EventArgs e)
        {
            ((CMutationOperatorChecklistItem)mutationOperatorChecklist.SelectedItem).setMutationRate(decimal.ToDouble(mutationRateUpDown.Value)) ;
        }

        #region CMutationOperatorChecklistItem class definition.
        /**
         * A representation of a mutation operator item
         * in the mutation operator checklist.
         */
        private class CMutationOperatorChecklistItem
        {
            /** The mutation operator ID. */
            public enum Type { DisplacementMutation, InsertionMutation, InversionMutation, ScrambleMutation, ExchangeMutation, DisplacedInversionMutation } ;

            /** The name of the operator as it appears in the list. */
            private string m_name ;

            /** The type of the operator. */
            private Type m_type ;

            /** The crossover rate for the operator. */
            private double m_mutationRate ;

            /**
             * Creates a new mutation operator checklist item.
             * 
             * @param name The name of the operator as it appears
             *             in the checklist.
             * @param type The ID type of the operator.
             */
            public CMutationOperatorChecklistItem(string name, Type type)
            {
                m_name         = name ;
                m_type         = type ;
                m_mutationRate = 0.1 ;
            }

            /**
             * Creates the specified type of the mutation operator
             * checklist item.
             * 
             * @param type The type of the mutation operator.
             */
            public CMutationOperatorChecklistItem(Type type)
            {
                m_type = type ;
            }

            public override bool Equals(object obj)
            {
                return (m_type == ((CMutationOperatorChecklistItem)obj).getType()) ;
            }

            public override int GetHashCode()
            {
                return m_type.GetHashCode() ;
            }

            public void setMutationRate(double rate)
            {
                m_mutationRate = rate ;
            }

            public double getMutationRate()
            {
                return m_mutationRate ;
            }

            public Type getType()
            {
                return m_type ;
            }

            public override string ToString()
            {
                return m_name ;
            }
        }
        #endregion CMutationOperatorChecklistItem class definition.

        #endregion >>> Mutation Operators Configuration Controls <<<

        #region >>> The Genetic Algorithm Configuration Controls Setters <<<

        /**
         * Makes the opposite operation then controlsToConfiguration. It
         * fill the controls according to the values found in the inner
         * configuration file.
         */
        private void configurationToControls()
        {
            populationSizeUpDown.Value = m_geneticAlgorithmConfiguration.getPopulationSize() ;

            if ((elitismCheckbox.Checked = m_geneticAlgorithmConfiguration.getNumberOfFittestToCopy() > 0))
            {
                elitismUpDown.Value = m_geneticAlgorithmConfiguration.getNumberOfFittestToCopy() ;
            }    

            chooseFitnessEvaluatorFactors() ;

            chooseSelectionOperator() ;

            chooseScalingMethod() ;

            chooseCrossoverOperator() ;

            chooseMutationOperator() ;
        }

        /**
         * Sets the fitness evaluator factors according
         * to the configuration.
         */
        private void chooseFitnessEvaluatorFactors()
        {
            if (m_geneticAlgorithmConfiguration.getFitnessEvaluator() is CMinimizeMakespanFitnessEvaluator)
            {
                CMinimizeMakespanFitnessEvaluator evaluator = (CMinimizeMakespanFitnessEvaluator)m_geneticAlgorithmConfiguration.getFitnessEvaluator() ;

                makespanFactorUpDown.Value = (decimal)evaluator.getMakespanFactor() ;
                latenessFactorUpDown.Value = (decimal)evaluator.getLatenessFactor() ;
            }
            else
            {
                throw new InvalidOperationException("The configuration does not contain CMinimizeMakespanFitnessEvaluator!") ;
            }
        }

        /**
         * Sets the scaling radio buttons' state 
         * according to the configuration.
         */
        private void chooseScalingMethod()
        {
            if (m_geneticAlgorithmConfiguration.getFitnessScaler() is CNullFitnessScaler)
            {
                scalingNoneRadioButton.Checked = true ;
            }
            else if (m_geneticAlgorithmConfiguration.getFitnessScaler() is CRankScaling)
            {
                scalingRankRadioButton.Checked = true ;
            }
            else if (m_geneticAlgorithmConfiguration.getFitnessScaler() is CSigmaScaling)
            {
                scalingSigmaRadioButton.Checked = true ;
            }
            else if (m_geneticAlgorithmConfiguration.getFitnessScaler() is CBoltzmannScaling)
            {
                scalingBoltzmannRadioButton.Checked = true ;

                CBoltzmannScaling scaler = (CBoltzmannScaling)m_geneticAlgorithmConfiguration.getFitnessScaler() ;

                boltzmannStartTempUpDown.Value = (decimal)scaler.getStartTemperature() ;
                boltzmannMinTempUpDown.Value = (decimal)scaler.getMinimumTemperature() ;
                boltzmannDropStepUpDown.Value = (decimal)scaler.getCoolingStepSize() ;
            }
        }

        /**
         * Sets the selection radio buttons' state 
         * according to the configuration.
         */ 
        private void chooseSelectionOperator()
        {
            if (m_geneticAlgorithmConfiguration.getNaturalSelector() is CRouletteWheelSelector)
            {
                selectorRWSRadioButton.Checked = true ;
            }
            else if (m_geneticAlgorithmConfiguration.getNaturalSelector() is CRandomTournamentSelector)
            {
                selectorRTSRadioButton.Checked = true ;

                CRandomTournamentSelector selector = (CRandomTournamentSelector)m_geneticAlgorithmConfiguration.getNaturalSelector() ;

                tournamentSizeUpDown.Value = selector.getTournamentSize() ;
                tournamentWinProbUpDown.Value = (decimal)selector.getFitterWinProbability() ;
            }
            else if (m_geneticAlgorithmConfiguration.getNaturalSelector() is CStochasticUniversalSampling)
            {
                selectorSUSRadioButton.Checked = true ;
            }
            else if (m_geneticAlgorithmConfiguration.getNaturalSelector() is CDeterministicSampling)
            {
                selectorDSRadioButton.Checked = true ;
            }
            else if (m_geneticAlgorithmConfiguration.getNaturalSelector() is CRemainderStochasticSampling)
            {
                selectorRSSRadioButton.Checked = true ;

                CRemainderStochasticSampling selector = (CRemainderStochasticSampling)m_geneticAlgorithmConfiguration.getNaturalSelector() ;

                useSUSReplacementCheckbox.Checked = selector.isWithReplacement() ;
            }            
        }

        /**
         * Sets up the crossover operator checklist according
         * to the configuration.
         */
        private void chooseCrossoverOperator() 
        {
            /* Unchecks all items in the checklist. */
            for (int i = crossoverOperatorChecklist.Items.Count - 1 ; i >= 0 ; i--)
            {
                crossoverOperatorChecklist.SetItemChecked(i, false) ;
            }
                
            List<ICrossoverOperator> crossoverOperators = m_geneticAlgorithmConfiguration.getCrossoverOperators() ;

            /* Finds the crossover operator of the configuration in the checklist,
             * and moves them up the list in the right order.
             */
            for (int k = 0, i = 0 ; k < crossoverOperators.Count ; k++, i++)
            {
                int j = i ;

                if (crossoverOperators[k] is COrderCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.OrderCrossover)) ;
                    ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.Items[j]).setCrossoverRatio(((COrderCrossover)crossoverOperators[k]).getCrossoverRatio()) ;
                    ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.Items[j]).setNeedCrossoverRatio(true) ;
                }
                else if (crossoverOperators[k] is COrderBasedCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.OrderBasedCrossover)) ;
                    ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.Items[j]).setCrossoverRatio(((COrderBasedCrossover)crossoverOperators[k]).getCrossoverRatio()) ;
                    ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.Items[j]).setNeedCrossoverRatio(true) ;
                }
                else if (crossoverOperators[k] is CPositionBasedCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.PositionBasedCrossover)) ;
                    ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.Items[j]).setCrossoverRatio(((CPositionBasedCrossover)crossoverOperators[k]).getCrossoverRatio()) ;
                    ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.Items[j]).setNeedCrossoverRatio(true) ;
                }
                else if (crossoverOperators[k] is CCycleCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.CycleCrossover)) ;
                }
                else if (crossoverOperators[k] is CPartiallyMappedCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.PartiallyMappedCrossover)) ;
                }
                else if (crossoverOperators[k] is CAlternatingPositionCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.AlternatingPositionCrossover)) ;
                }
                else if (crossoverOperators[k] is CHeuristicCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.HeuristicCrossover)) ;
                }
                else if (crossoverOperators[k] is CGreedyCrossover)
                {
                    j = crossoverOperatorChecklist.Items.IndexOf(new CCrossoverOperatorChecklistItem(CCrossoverOperatorChecklistItem.Type.GreedyCrossover)) ;
                }

                ((CCrossoverOperatorChecklistItem)crossoverOperatorChecklist.Items[j]).setCrossoverRate(((CBaseCrossoverOperator)crossoverOperators[k]).getCrossoverRate()) ;
                
                object tmp = crossoverOperatorChecklist.Items[j] ;
                crossoverOperatorChecklist.Items[j] = crossoverOperatorChecklist.Items[i] ;
                crossoverOperatorChecklist.Items[i] = tmp ;
            }

            /* Checks the crossover operators that are in the configuration. */
            for (int i = 0 ; i < crossoverOperators.Count ; i++)
            {
                crossoverOperatorChecklist.SetItemChecked(i, true) ;
            }
        }

        /**
         * Sets up the mutation operator checklist according
         * to the configuration.
         */
        private void chooseMutationOperator()
        {
            /* Unchecks all items in the checklist. */
            for (int i = mutationOperatorChecklist.Items.Count - 1 ; i >= 0 ; i--)
            {
                mutationOperatorChecklist.SetItemChecked(i, false) ;
            }
                
            List<IMutationOperator> mutationOperators = m_geneticAlgorithmConfiguration.getMutationOperators() ;

            /* Finds the mutation operator of the configuration in the checklist,
             * and moves them up the list in the right order.
             */
            for (int k = 0, i = 0 ; k < mutationOperators.Count ; k++, i++)
            {
                int j = i ;

                if (mutationOperators[k] is CInsertionMutation)
                {
                    j = mutationOperatorChecklist.Items.IndexOf(new CMutationOperatorChecklistItem(CMutationOperatorChecklistItem.Type.InsertionMutation)) ;
                }
                else if (mutationOperators[k] is CInversionMutation)
                {
                    j = mutationOperatorChecklist.Items.IndexOf(new CMutationOperatorChecklistItem(CMutationOperatorChecklistItem.Type.InversionMutation)) ;
                }
                else if (mutationOperators[k] is CExchangeMutation)
                {
                    j = mutationOperatorChecklist.Items.IndexOf(new CMutationOperatorChecklistItem(CMutationOperatorChecklistItem.Type.ExchangeMutation)) ;
                }
                else if (mutationOperators[k] is CDisplacementMutation)
                {
                    j = mutationOperatorChecklist.Items.IndexOf(new CMutationOperatorChecklistItem(CMutationOperatorChecklistItem.Type.DisplacementMutation)) ;
                }
                else if (mutationOperators[k] is CScrambleMutation)
                {
                    j = mutationOperatorChecklist.Items.IndexOf(new CMutationOperatorChecklistItem(CMutationOperatorChecklistItem.Type.ScrambleMutation)) ;
                }
                else if (mutationOperators[k] is CDisplacedInversionMutation)
                {
                    j = mutationOperatorChecklist.Items.IndexOf(new CMutationOperatorChecklistItem(CMutationOperatorChecklistItem.Type.DisplacedInversionMutation)) ;
                }

                ((CMutationOperatorChecklistItem)mutationOperatorChecklist.Items[j]).setMutationRate(((CBaseMutationOperator)mutationOperators[k]).getMutationRate()) ;
                
                object tmp = mutationOperatorChecklist.Items[j] ;
                mutationOperatorChecklist.Items[j] = mutationOperatorChecklist.Items[i] ;
                mutationOperatorChecklist.Items[i] = tmp ;
            }

            /* Checks the mutation operators that are in the configuration. */
            for (int i = 0 ; i < mutationOperators.Count ; i++)
            {
                mutationOperatorChecklist.SetItemChecked(i, true) ;
            }
        }

        #endregion >>> The Genetic Algorithm Configuration Controls Setters <<<

        #region >>> The Genetic Algorithm Configuration Controls Getters <<<

        /**
         * Updates the inner configuration file according to the values 
         * of the controls on the genetic algorithm configuration tab 
         * GAConf.
         */
        private void controlsToConfiguration()
        {
            /* Sets up the population size. */
            m_geneticAlgorithmConfiguration.setPopulationSize(decimal.ToInt32(populationSizeUpDown.Value)) ;

            if (elitismCheckbox.Checked)
            {
                /* Sets up the elitism. */
                m_geneticAlgorithmConfiguration.setNumberOfFittestToCopy(decimal.ToInt32(elitismUpDown.Value)) ;
            }

            /* Sets up the fitness evaluator. */
            m_geneticAlgorithmConfiguration.setFitnessEvaluator(determineFitnessEvaluator()) ;

            /* Sets up the fitness scaler. */
            m_geneticAlgorithmConfiguration.setFitnessScaler(determineFitnessScaler()) ;

            /* Sets up the natural selector. */
            m_geneticAlgorithmConfiguration.setNaturalSelector(determineSelectionOperator()) ;

            /* Removes all previously registered crossover operators. */
            m_geneticAlgorithmConfiguration.removeAllCrossoverOperators() ;

            /* Sets up the crossover operators. */
            foreach (ICrossoverOperator crossover in determineCrossoverOperators())
            {
                m_geneticAlgorithmConfiguration.addCrossoverOperator(crossover) ;
            }

            /* Removes all previously registered mutation operators. */
            m_geneticAlgorithmConfiguration.removeAllMutationOperators() ;
            
            /* Sets up the mutation operators. */
            foreach (IMutationOperator mutation in determineMutationOperators())
            {
                m_geneticAlgorithmConfiguration.addMutationOperator(mutation) ;
            }          
        }

        /**
         * Returns the natural selector from the Selection radio button group.
         * 
         * @return The natural selector from the Selection radio button group.
         */
        private ISelectionOperator determineSelectionOperator()
        {
            if (selectorRWSRadioButton.Checked)
            {
                return new CRouletteWheelSelector(m_geneticAlgorithmConfiguration) ;
            }
            else if (selectorRTSRadioButton.Checked)
            {
                return new CRandomTournamentSelector(m_geneticAlgorithmConfiguration, decimal.ToInt32(tournamentSizeUpDown.Value), decimal.ToDouble(tournamentWinProbUpDown.Value)) ;
            }
            else if (selectorSUSRadioButton.Checked)
            {
                return new CStochasticUniversalSampling(m_geneticAlgorithmConfiguration) ;
            }
            else if (selectorDSRadioButton.Checked)
            {
                return new CDeterministicSampling(m_geneticAlgorithmConfiguration) ;
            }
            else if (selectorRSSRadioButton.Checked)
            {
                return new CRemainderStochasticSampling(m_geneticAlgorithmConfiguration, useSUSReplacementCheckbox.Checked) ;
            }
            else 
            {
                return null ;
            }
        }

        /**
         * Returns the list of crossover operators from the Crossover
         * Operator Checklist.
         * 
         * @return The list of selected crossover operators.
         */
        private List<ICrossoverOperator> determineCrossoverOperators()
        {
            List<ICrossoverOperator> crossoverOperators = new List<ICrossoverOperator>() ;

            foreach (object item in crossoverOperatorChecklist.CheckedItems)
            {
                CCrossoverOperatorChecklistItem crossoverItem = (CCrossoverOperatorChecklistItem)item ;

                switch (crossoverItem.getType())
                {
                case CCrossoverOperatorChecklistItem.Type.OrderCrossover:
                    crossoverOperators.Add(new COrderCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate(), crossoverItem.getCrossoverRatio())) ;
                    break ;

                case CCrossoverOperatorChecklistItem.Type.OrderBasedCrossover:
                    crossoverOperators.Add(new COrderBasedCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate(), crossoverItem.getCrossoverRatio())) ;
                    break ;

                case CCrossoverOperatorChecklistItem.Type.PositionBasedCrossover:
                    crossoverOperators.Add(new CPositionBasedCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate(), crossoverItem.getCrossoverRatio())) ;
                    break ;

                case CCrossoverOperatorChecklistItem.Type.CycleCrossover:
                    crossoverOperators.Add(new CCycleCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate())) ;
                    break ;

                case CCrossoverOperatorChecklistItem.Type.PartiallyMappedCrossover:
                    crossoverOperators.Add(new CPartiallyMappedCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate())) ;
                    break ;

                case CCrossoverOperatorChecklistItem.Type.AlternatingPositionCrossover:
                    crossoverOperators.Add(new CAlternatingPositionCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate())) ;
                    break ;

                case CCrossoverOperatorChecklistItem.Type.HeuristicCrossover:
                    crossoverOperators.Add(new CHeuristicCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate())) ;
                    break ;

                case CCrossoverOperatorChecklistItem.Type.GreedyCrossover:
                    crossoverOperators.Add(new CGreedyCrossover(m_geneticAlgorithmConfiguration, crossoverItem.getCrossoverRate())) ;
                    break ;
                }
            }

            return crossoverOperators ;
        }

        /**
         * Returns the list of mutation operators from the Mutation
         * Operator Checklist.
         * 
         * @return The list of selected mutation operators.
         */
        private List<IMutationOperator> determineMutationOperators()
        {
            List<IMutationOperator> mutationOperators = new List<IMutationOperator>() ;

            foreach (object item in mutationOperatorChecklist.CheckedItems)
            {
                CMutationOperatorChecklistItem mutationItem = (CMutationOperatorChecklistItem)item ;

                switch (mutationItem.getType())
                {
                case CMutationOperatorChecklistItem.Type.InsertionMutation:
                    mutationOperators.Add(new CInsertionMutation(m_geneticAlgorithmConfiguration, mutationItem.getMutationRate())) ;
                    break ;

                case CMutationOperatorChecklistItem.Type.InversionMutation:
                    mutationOperators.Add(new CInversionMutation(m_geneticAlgorithmConfiguration, mutationItem.getMutationRate())) ;
                    break ;

                case CMutationOperatorChecklistItem.Type.ExchangeMutation:
                    mutationOperators.Add(new CExchangeMutation(m_geneticAlgorithmConfiguration, mutationItem.getMutationRate())) ;
                    break ;

                case CMutationOperatorChecklistItem.Type.DisplacementMutation:
                    mutationOperators.Add(new CDisplacementMutation(m_geneticAlgorithmConfiguration, mutationItem.getMutationRate())) ;
                    break ;

                case CMutationOperatorChecklistItem.Type.ScrambleMutation:
                    mutationOperators.Add(new CScrambleMutation(m_geneticAlgorithmConfiguration, mutationItem.getMutationRate())) ;
                    break ;

                case CMutationOperatorChecklistItem.Type.DisplacedInversionMutation:
                    mutationOperators.Add(new CDisplacedInversionMutation(m_geneticAlgorithmConfiguration, mutationItem.getMutationRate())) ;
                    break ;
                }
            }

            return mutationOperators ;
        }

        /**
         * Returns the fitness evaluator set according to the values
         * of makespan, and lateness controls.
         * 
         * @return The fitness evaluator.
         */
        private IFitnessEvaluator determineFitnessEvaluator()
        {
            return new CMinimizeMakespanFitnessEvaluator(decimal.ToDouble(makespanFactorUpDown.Value), decimal.ToDouble(latenessFactorUpDown.Value)) ;
        }

        /**
         * Return the fitness scaler from the Scaling radio button group.
         * 
         * @param The fitness scaler from the Scaling radio button group.
         */
        private IFitnessScaler determineFitnessScaler()
        {
            if (scalingNoneRadioButton.Checked)
            {
                return new CNullFitnessScaler() ;
            }
            else if (scalingRankRadioButton.Checked)
            {
                return new CRankScaling() ;
            }
            else if (scalingSigmaRadioButton.Checked)
            {
                return new CSigmaScaling() ;
            }
            else if (scalingBoltzmannRadioButton.Checked)
            {
                return new CBoltzmannScaling(decimal.ToDouble(boltzmannStartTempUpDown.Value), decimal.ToDouble(boltzmannMinTempUpDown.Value), decimal.ToDouble(boltzmannDropStepUpDown.Value)) ;
            }
            else
            {
                return null ;
            }
        }

        #endregion >>> The Genetic Algorithm Configuration Controls Getters <<<

        #region >>> The Genetic Algorithm Configuration Load & Save Button Handlers <<<

        /**
         * Invoked when the Save Configuration... button has been pressed.
         * It opens Save File Dialog.
         */
        private void gaSaveConfigurationClick(object sender, EventArgs e)
        {
            if (saveConfigurationDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                controlsToConfiguration() ;

                m_geneticAlgorithmConfiguration.setName(Path.GetFileNameWithoutExtension(saveConfigurationDialog.FileName)) ;

                try
                {
                    m_geneticAlgorithmConfiguration.serialize(saveConfigurationDialog.FileName) ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error occured during the processing of the file!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error) ;
                }
            }
        }

        /**
         * Invoked when the Load Configuration... button has been pressed.
         * It opens Open File Dialog.
         */
        private void gaLoadConfigurationClick(object sender, EventArgs e)
        {
            if (loadConfigurationDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    m_geneticAlgorithmConfiguration.deserialize(loadConfigurationDialog.FileName) ;

                    configurationToControls() ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error occured during the processing of the file!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error) ;
                }
            }
        }

        #endregion >>> The Genetic Algorithm Configuraton Load & Save Button Handlers <<<

        #endregion >>> The Genetic Algorithm Configuration Tab <<<

        #region >>> The Mixed Integer Programming Solver Tab <<<

        /**
         * Invoked when the time limit of MIP solver is enabled / disabled.
         */
        private void mipTimeLimitCheckboxCheckedChanged(object sender, EventArgs e)
        {
            mipTimeLimitUpDown.Enabled = mipTimeLimitCheckbox.Checked ;
        }

        /**
         * Invoked when the iteration limit of MIP solver is enabled / disabled.
         */
        private void mipIterationLimitCheckboxCheckedChanged(object sender, EventArgs e)
        {
            mipIterationLimitUpDown.Enabled = mipIterationLimitCheckbox.Checked ;
        }

        /**
         * Invoked when the "Start Optimization" button for the MIP solver 
         * has been pressed.
         */
        private void mipStartButtonClick(object sender, EventArgs e)
        {
            mipStartButton.Enabled = false ;
            mipStopButton.Enabled  = true ;

            foreach (CMixedCriticalityInstance testInstance in m_mipSolutionsList)
            {
                testInstance.MakespanDataView = "Solving..." ;
                testInstance.LatenessDataView = "Solving..." ;
            }

            mipSolutionsTable.Refresh() ;

            /* Sets or Unsets the time limit. */
            if (mipTimeLimitCheckbox.Checked)
            {
                m_mipSolver.setTimeLimit((int)mipTimeLimitUpDown.Value) ;
            }
            else
            {
                m_mipSolver.setTimeLimit(-1) ;
            }

            /* Sets or unsets the iteration limit. */
            if (mipIterationLimitCheckbox.Checked)
            {
                m_mipSolver.setIterationLimit((int)mipIterationLimitUpDown.Value) ;
            }
            else
            {
                m_mipSolver.setIterationLimit(-1) ;
            }

            /* Runs the MIP solver in the background. */
            m_mipSolverThread.run() ;
        }

        /**
         * Invoked when the "Stop Optimization" button for the MIP solver 
         * has been pressed.
         */
        private void mipStopButtonClick(object sender, EventArgs e)
        {
            /* This handler will be called twice: the 1st time 
             * when handling the mouse press event... 
             */
            if (!Object.ReferenceEquals(sender, m_mipSolverThread))
            {
                mipStopButton.Enabled  = false ;
                mipStartButton.Enabled = true ;   
  
                m_mipSolverThread.abort() ;
            }
            else
            {
                /**
                 * ...the 2nd time as a result of solver being aborted/finished.
                 */
                if (mipStopButton.Enabled != false)
                {
                    mipStopButton.BeginInvoke(m_mipSolverStopHandler, new Object[] { mipStopButton, EventArgs.Empty }) ;
                }
            }
        }

        private void saveMipSolutionButtonClick(object sender, EventArgs e)
        {
            saveInstancesDialog.SelectedPath = m_lastTestCollectionPath ;

            if (saveInstancesDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    CMixedCriticalityInstancesSerializer.saveSolution(saveInstancesDialog.SelectedPath, m_mipSolutionsList) ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error has occured during the process!", "Error", MessageBoxButtons.OK) ;
                }
            }
        }

        #endregion >>> The Mixed Integer Programming Solver Tab <<<

        #region >>> The Genetic Algorithm Solver Tab <<<

        /**
         * Invoked when the time limit of GA solver is enabled / disabled.
         */
        private void gaTimeLimitCheckboxCheckedChanged(object sender, EventArgs e)
        {
            gaTimeLimitUpDown.Enabled = gaTimeLimitCheckbox.Checked ;
        }

        /**
         * Invoked when the iteration limit of GA solver is enabled / disabled.
         */
        private void gaIterationLimitCheckboxCheckedChanged(object sender, EventArgs e)
        {
            gaIterationLimitUpDown.Enabled = gaIterationLimitCheckbox.Checked ;
        }

        private void gaStartButtonClick(object sender, EventArgs e)
        {
            gaStartButton.Enabled = false ;
            gaStopButton.Enabled  = true ;

            foreach (CMixedCriticalityInstance testInstance in m_gaSolutionsList)
            {
                testInstance.MakespanDataView = "Solving..." ;
                testInstance.LatenessDataView = "Solving..." ;
            }

            gaSolutionsTable.Refresh() ;

            /* Sets or Unsets the time limit. */
            if (gaTimeLimitCheckbox.Checked)
            {
                m_gaSolver.setTimeLimit((int)gaTimeLimitUpDown.Value) ;
            }
            else
            {
                m_gaSolver.setTimeLimit(-1) ;
            }

            /* Sets or unsets the iteration limit. */
            if (gaIterationLimitCheckbox.Checked)
            {
                m_gaSolver.setIterationLimit((int)gaIterationLimitUpDown.Value) ;
            }
            else
            {
                m_gaSolver.setIterationLimit(-1) ;
            }

            /* Reads the currect configuration of the genetic algorithm. */
            controlsToConfiguration() ;

            /* Runs the GA solver in the background. */
            m_gaSolverThread.run() ;
        }

        private void gaStopButtonClick(object sender, EventArgs e)
        {
            /* This handler will be called twice: the 1st time 
             * when handling the mouse press event... 
             */
            if (!Object.ReferenceEquals(sender, m_gaSolverThread))
            {
                gaStopButton.Enabled  = false ;
                gaStartButton.Enabled = true ;   
  
                m_gaSolverThread.abort() ;
            }
            else
            {
                /**
                 * ...the 2nd time as a result of solver being aborted/finished.
                 */
                if (gaStopButton.Enabled != false)
                {
                    gaStopButton.BeginInvoke(m_gaSolverStopHandler, new Object[] { gaStopButton, EventArgs.Empty }) ;
                }
            }
        }

        private void saveGASolutionButtonClick(object sender, EventArgs e)
        {
            saveInstancesDialog.SelectedPath = m_lastTestCollectionPath ;

            if (saveInstancesDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    CMixedCriticalityInstancesSerializer.saveSolution(saveInstancesDialog.SelectedPath, m_gaSolutionsList) ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error has occured during the process!", "Error", MessageBoxButtons.OK) ;
                }
            }
        }

        #endregion >>> The Genetic Algorithm Solver Tab <<<

        #region >>> Solution Tables Mouse Event Handlers <<<

        /**
         * Invoked when a test instance result table row has been double-clicked.
         * As a result Gantt View Chart with the selected instance is displayed.
         */
        private void testInstancesResultTableDoubleClick(object sender, EventArgs e)
        {
            if (((DataGridView)sender).SelectedRows.Count != 0)
            {
                try
                {
                    (new GanttViewForm((CMixedCriticalityInstance)((DataGridView)sender).SelectedRows[0].DataBoundItem)).Show(this) ;
                }
                catch(InvalidOperationException)
                {
                    MessageBox.Show("Cannot display selected instance unless it has been solved!", "GanttView", MessageBoxButtons.OK, MessageBoxIcon.Information) ;
                }
            }
        }

        #endregion >>> Solution Tables Mouse Event Handlers <<<

        #region >>> The Simulated Annealing Solver Tab <<<

        /**
         * Invoked when the time limit of SA solver is enabled / disabled.
         */
        private void saTimeLimitCheckboxCheckedChanged(object sender, EventArgs e)
        {
            saTimeLimitUpDown.Enabled = saTimeLimitCheckbox.Checked ;
        }

        /**
         * Invoked when the iteration limit of SA solver is enabled / disabled.
         */
        private void saIterationLimitCheckboxCheckedChanged(object sender, EventArgs e)
        {
            saIterationLimitUpDown.Enabled = saIterationLimitCheckbox.Checked ;
        }

        /**
         * Invoked when the makespan factor value has changed.
         */
        private void saMakespanFactorUpDownValueChanged(object sender, EventArgs e)
        {
            saLatenessFactorUpDown.Value = (decimal)1.0 - saMakespanFactorUpDown.Value ;
        }

        /**
         * Invoked when the lateness factor value has changed.
         */
        private void saLatenessFactorUpDownValueChanged(object sender, EventArgs e)
        {
            saMakespanFactorUpDown.Value = (decimal)1.0 - saLatenessFactorUpDown.Value ;
        }

        /**
         * Invoked when the "Start Optimization" button for the MIP solver 
         * has been pressed.
         */
        private void saStartButtonClick(object sender, EventArgs e)
        {
            saStartButton.Enabled = false ;
            saStopButton.Enabled  = true ;

            foreach (CMixedCriticalityInstance testInstance in m_saSolutionsList)
            {
                testInstance.MakespanDataView = "Solving..." ;
                testInstance.LatenessDataView = "Solving..." ;
            }

            saSolutionsTable.Refresh() ;

            /* Sets or Unsets the time limit. */
            if (saTimeLimitCheckbox.Checked)
            {
                m_saSolver.setTimeLimit((int)saTimeLimitUpDown.Value) ;
            }
            else
            {
                m_saSolver.setTimeLimit(-1) ;
            }

            /* Sets or unsets the iteration limit. */
            if (saIterationLimitCheckbox.Checked)
            {
                m_saSolver.setIterationLimit((int)saIterationLimitUpDown.Value) ;
            }
            else
            {
                m_saSolver.setIterationLimit(-1) ;
            }

            m_simulatedAnnealingConfiguration.setMakespanFactor((double)saMakespanFactorUpDown.Value) ;
            m_simulatedAnnealingConfiguration.setLatenessFactor((double)saLatenessFactorUpDown.Value) ;
            m_simulatedAnnealingConfiguration.setMaximumTemperature((double)saMaxTemperatureUpDown.Value) ;
            m_simulatedAnnealingConfiguration.setMinimumTemperature((double)saMinTemperatureUpDown.Value) ;
            m_simulatedAnnealingConfiguration.setCoolingFactor((double)saCoolingFactorUpDown.Value) ;

            m_saSolver.setConfiguration(m_simulatedAnnealingConfiguration) ;
            
            /* Runs the MIP solver in the background. */
            m_saSolverThread.run() ;
        }

        /**
         * Invoked when the "Stop Optimization" button for the SA solver 
         * has been pressed.
         */
        private void saStopButtonClick(object sender, EventArgs e)
        {
            /* This handler will be called twice: the 1st time 
             * when handling the mouse press event... 
             */
            if (!Object.ReferenceEquals(sender, m_saSolverThread))
            {
                saStopButton.Enabled  = false ;
                saStartButton.Enabled = true ;   
  
                m_saSolverThread.abort() ;
            }
            else
            {
                /**
                 * ...the 2nd time as a result of solver being aborted/finished.
                 */
                if (saStopButton.Enabled != false)
                {
                    saStopButton.BeginInvoke(m_saSolverStopHandler, new Object[] { saStopButton, EventArgs.Empty }) ;
                }
            }
        }

        private void saveSASolutionButtonClick(object sender, EventArgs e)
        {
            saveInstancesDialog.SelectedPath = m_lastTestCollectionPath ;

            if (saveInstancesDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    CMixedCriticalityInstancesSerializer.saveSolution(saveInstancesDialog.SelectedPath, m_saSolutionsList) ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error has occured during the process!", "Error", MessageBoxButtons.OK) ;
                }
            }
        }

        private void saMinTemperatureUpDownValueChanged(object sender, EventArgs e)
        {
            if (saMinTemperatureUpDown.Value > saMaxTemperatureUpDown.Value)
            {
                saMinTemperatureUpDown.Value = saMaxTemperatureUpDown.Value ;
            }
        }

        #endregion >>> The Simulated Annealing Solver Tab <<<

        #region >>> The Clairvoyant EDF Solver Tab <<<

        /**
         * Invoked when the "Start Optimization" button for the CEDF solver 
         * has been pressed.
         */
        private void cedfStartButtonClick(object sender, EventArgs e)
        {
            cedfStartButton.Enabled = false ;
            cedfStopButton.Enabled  = true ;

            foreach (CMixedCriticalityInstance testInstance in m_cedfSolutionsList)
            {
                testInstance.MakespanDataView = "Solving..." ;
                testInstance.LatenessDataView = "Solving..." ;
            }

            cedfSolutionsTable.Refresh() ;

            /* Runs the CEDF solver in the background. */
            m_cedfSolverThread.run() ;
        }

        /**
         * Invoked when the "Stop Optimization" button for the MIP solver 
         * has been pressed.
         */
        private void cedfStopButtonClick(object sender, EventArgs e)
        {
            /* This handler will be called twice: the 1st time 
             * when handling the mouse press event... 
             */
            if (!Object.ReferenceEquals(sender, m_cedfSolverThread))
            {
                cedfStopButton.Enabled  = false ;
                cedfStartButton.Enabled = true ;   
  
                m_cedfSolverThread.abort() ;
            }
            else
            {
                /**
                 * ...the 2nd time as a result of solver being aborted/finished.
                 */
                if (cedfStopButton.Enabled != false)
                {
                    cedfStopButton.BeginInvoke(m_cedfSolverStopHandler, new Object[] { cedfStopButton, EventArgs.Empty }) ;
                }
            }
        }

        private void saveCEDFSolutionButtonClick(object sender, EventArgs e)
        {
            saveInstancesDialog.SelectedPath = m_lastTestCollectionPath ;

            if (saveInstancesDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    CMixedCriticalityInstancesSerializer.saveSolution(saveInstancesDialog.SelectedPath, m_cedfSolutionsList) ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error has occured during the process!", "Error", MessageBoxButtons.OK) ;
                }
            }
        }

        #endregion >>> The Clairvoyant EDF Solver Tab <<<

        #region >>> The DP Solver Tab <<<

        private void dpStartButtonClick(object sender, EventArgs e)
        {
            dpStartButton.Enabled = false ;
            dpStopButton.Enabled  = true ;

            foreach (CMixedCriticalityInstance testInstance in m_dpSolutionsList)
            {
                testInstance.MakespanDataView = "Solving..." ;
                testInstance.LatenessDataView = "Solving..." ;
            }

            dpSolutionsTable.Refresh() ;

            /* Runs the DP solver in the background. */
            m_dpSolverThread.run() ;
        }

        private void dpStopButtonClick(object sender, EventArgs e)
        {
            /* This handler will be called twice: the 1st time 
             * when handling the mouse press event... 
             */
            if (!Object.ReferenceEquals(sender, m_dpSolverThread))
            {
                dpStopButton.Enabled  = false ;
                dpStartButton.Enabled = true ;   
  
                m_dpSolverThread.abort() ;
            }
            else
            {
                /**
                 * ...the 2nd time as a result of solver being aborted/finished.
                 */
                if (dpStopButton.Enabled != false)
                {
                    dpStopButton.BeginInvoke(m_dpSolverStopHandler, new Object[] { dpStopButton, EventArgs.Empty }) ;
                }
            }
        }

        private void saveDPSolutionButtonClick(object sender, EventArgs e)
        {
            saveInstancesDialog.SelectedPath = m_lastTestCollectionPath ;

            if (saveInstancesDialog.ShowDialog(this) == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    CMixedCriticalityInstancesSerializer.saveSolution(saveInstancesDialog.SelectedPath, m_dpSolutionsList) ;
                }
                catch (Exception)
                {
                    MessageBox.Show(this, "An error has occured during the process!", "Error", MessageBoxButtons.OK) ;
                }
            }
        }

        #endregion >>> The DP Solver Tab <<<
    }
}
