﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MCScheduling.MixedCriticality
{
    /**
     * The basic implementation of the IMixedCriticalitySolver. It should be
     * used as a base class for an algorithn solving mixed-criticality instances.
     */
    public abstract class CBaseMixedCriticalitySolver : IMixedCriticalitySolver
    {
        /**
         * A status of a solution that is returned as result of the solving algorithm.
         */
        public class Status
        {            
            /**
             * Error has occured during solving.
             */
            public const int ERROR = -1 ;

            /**
             * No problem has been optimized yet.
             */
            public const int NONE = 0 ;

            /**
             * Model was proven to be infeasible.
             */
            public const int INFEASIBLE = 1 ;
            
            /**
             * Optimization was terminated by the user.
             */
            public const int INTERRUPTED = 2 ;

            /**
             * Optimization terminated because the total number of simplex iterations performed
             * exceeded the limit specified in the IterationLimit parameter.
             */
            public const int ITERATION_LIMIT = 3 ;

            /**
             * Optimization terminated because the total elapsed time exceeded the limit
             * specified in the TimeLimit parameter.
             */
            public const int TIME_LIMIT = 4 ;

            /**
             * Model was solved to optimality (subject to tolerances) and an optimal solution
             * is available.
             */
            public const int OPTIMAL = 5 ;

            /**
             * Optimization terminated with a sub-optimal solution. The solution is feasible, yet
             * not proven optimal.
             */
            public const int SUBOPTIMAL = 6 ;

            /**
             * Infeasible solution found by meta-heuristic algorithm.
             */
            public const int HEURISTIC = 7 ;
        }

        /** 
         * The maximum number of iteration an iterative solver
         * may perform before stopping.
         */
        private int m_iterationLimit ;

        /**
         * The maximum time (in seconds) a solver may
         * run before stopping.
         */
        private int m_timeLimit ;

        /**
         * The status of the solution.
         */
        private int m_status ;

        /**
         * Constructs a new base mixed-criticality solver.
         */
        protected CBaseMixedCriticalitySolver()
        {
            m_iterationLimit = -1 ;
            m_timeLimit      = -1 ;
            m_status         = Status.NONE ;
        }

        /**
         * Limits the number of iteration an iterative algorithm
         * may perform. If the specified value is less then 0, 
         * the iteration limit is unset.
         * 
         * @param iterationLimit The maximum number of iteration
         *                       for the algorithm.
         */
        public void setIterationLimit(int iterationLimit)
        {
            if (iterationLimit <= 0)
            {
                m_iterationLimit = -1 ;
            }
            else
            {
                m_iterationLimit = iterationLimit ;
            }
        }

        /**
         * Returns the maximum number of iterations the algorithm
         * performs before stopping.
         * 
         * @return The maximum number of iteration of the algorithm.
         */
        public int getIterationLimit()
        {
            return m_iterationLimit ;
        }

        /**
         * Limits the maximum time expended by the algorithm.
         * If the specified value is less then 0, the time
         * limit is unset.
         * 
         * @param timeLimit The time limit (in seconds) for the algorithm.
         */
        public void setTimeLimit(int timeLimit)
        {
            if (timeLimit <= 0)
            {
                m_timeLimit = -1 ;
            }
            else
            {
                m_timeLimit = timeLimit ;
            }
        }

        /**
         * Returns the maximum time the algorithm may spend solving
         * a problem.
         * 
         * @return The time limit (in seconds) for the algorithm.
         */
        public int getTimeLimit()
        {
            return m_timeLimit ;
        }

        /**
         * Returns true if the algorithm has a time limit set.
         * 
         * @return True if the limit is set, false, otherwise.
         */
        public bool hasTimeLimit()
        {
            return getTimeLimit() != -1 ;
        }

        /**
         * Returns true if the algorithm has an iteration limit set.
         * 
         * @return True if the limit is set, false, otherwise.
         */
        public bool hasIterationLimit()
        {
            return getIterationLimit() != -1 ;
        }

        /**
         * Returns an integer value specifying the result of the last
         * call to <em>solve</em> method.
         * 
         * @return One of the public status codes defined in Status 
         *         class.
         */
        public int getStatus()
        {
            return m_status ;
        }

        /**
         * Sets the status of the result to call of solve method.
         * 
         * @param status The status of the last call of solve method.
         */
        protected void setStatus(int status)
        {
            m_status = status ;
        }

        /**
         * Solves the specified mixed-criticality instance.
         * 
         * @param instance The mixed-criticality instance 
         *                 to be solved.
         */
        public abstract void solve(CMixedCriticalityInstance instance) ;

        /**
         * Interrupts the process of solving.
         */
        public abstract void interrupt() ;
    }
}