﻿using System ;
using System.Collections.Generic ;
using MCScheduling.Utils ;

namespace MCScheduling.MixedCriticality
{
    /**
     * The default implemenation of the <em>IMixedCriticalityInstanceGenerator</em>
     * interface.
     */
    public class CDefaultMixedCriticalityInstanceGenerator
    {
        /**
         * Creates a set of <em>instanceCount</em> mixed-criticality instances,
         * each containing <em>jobCountPerCriticality[i]</em> mixed-criticality 
         * jobs of the i-th criticality, each is being assigned a random worst-
         * case execution time for each criticality level with its maximum 
         * value bounded by <em>maxExecutionTimes</em>. The release times and
         * deadlines of the jobs are determined randomly (if the resp. arguments 
         * are set to true), but it is guaranteed that there is at least one 
         * feasible schedule.
         * 
         * @param instanceNamePrefix     The prefix of the generated instances names.
         * @param instanceCount          The number of MC instances to generate.
         * @param jobCountPerCriticality The number of jobs per criticality level.
         * @param maximumExecutionTimes  The maximum random value of an exectution
         *                               time per level.
         * @param maxIdleTime            The maximum length of an idle time interval 
         *                               between jobs' executions.
         * @param timeWindowSpan         The length of the [release time; deadline] 
         *                               time window does not exceeds the largest 
         *                               execution time of a job times this constant.
         * 
         * @return The list of generated MC instances.
         */
        public static List<CMixedCriticalityInstance> generate(string instanceNamePrefix, int instanceCount, int[] jobCountPerCriticality, int[] maximumExecutionTimes, int maxIdleTime, double timeWindowSpan)
        {
            /** A pseudo-random number generator. */
            CUniformRandom randomizer = new CUniformRandom((ulong)DateTime.Now.Ticks) ;

            /* The total number of jobs to generate. */
            int jobIndex = 0, jobCount = 0 ;

            /* Calculates the total number of jobs. */
            foreach (int jobCountOfCriticality in jobCountPerCriticality)
            {
                jobCount += jobCountOfCriticality ;
            }

            /* A permutation used to assigne criticality levels to jobs. */
            int[] permutation = new int[jobCount] ;

            /* Creates a initial permutation. */
            for (int i = 0 ; i < permutation.Length ; i++)
            {
                permutation[i] = i ;
            }

            int[] criticalityAssignments = new int[jobCount] ;
            int   maximumCriticality     = 0 ;

            /* Creates a list of criticality assignments. */
            for (int i = 0 ; i < jobCountPerCriticality.Length ; i++)
            {
                for (int j = 0 ; j < jobCountPerCriticality[i] ; j++)
                {
                    criticalityAssignments[jobIndex++] = (i + 1) ;
                }

                if (jobCountPerCriticality[i] > 0)
                {
                    ++maximumCriticality ;
                }
            }

            /* The list of generated instances. */
            List<CMixedCriticalityInstance> instanceList = new List<CMixedCriticalityInstance>(instanceCount) ;

            /* A list of randomly generated execution times for a job. */
            double[] executionTimes = new double[jobCountPerCriticality.Length] ;

            string nameSuffixFormat = determineInstanceNameSuffix(instanceCount) ;

            /* Generates random mixed-criticality instances. */
            for (int iInstance = 1 ; iInstance <= instanceCount ; iInstance++)
            {
                /* Creates a random assignement of job criticalities: the i-th job's
                 * criticaliy is: criticalityAssignments[permutation[i]]. 
                 */
                shuffle(permutation, randomizer) ;

                /* A randomly generated instance. */
                CMixedCriticalityInstance instance = new CMixedCriticalityInstance(instanceNamePrefix + String.Format(nameSuffixFormat, iInstance)) ;

                /* Job generation loop. */
                for (jobIndex = 0 ; jobIndex < jobCount ; jobIndex++)
                {
                    /* The job's criticality level. */
                    int jobCriticality = criticalityAssignments[permutation[jobIndex]] ;

                    /* Generates execution times for this job. */
                    createExecutionTimes(executionTimes, maximumExecutionTimes, jobCriticality, randomizer) ;

                    instance.addJob(new CMixedCriticalityJob("T" + (jobIndex + 1), jobCriticality, executionTimes)) ;
                }

                int[] makespanAtCriticality = new int[maximumCriticality + 1] ;

                /* The 1st job starts at 0.0 */
                instance[0].setStartTime(0.0) ;

                for (int criticality = 1 ; criticality <= instance[0].getCriticality() ; criticality++)
                {
                    makespanAtCriticality[criticality] = (int)instance[0].getCompletionTime(criticality) ;
                }
                
                /* Calculates the temporal start times of jobs. */
                for (jobIndex = 1 ; jobIndex < instance.getJobCount() ; jobIndex++)
                {
                    int maxCritMakespan = 1 ;

                    for (int criticality = maxCritMakespan + 1 ; criticality <= instance[jobIndex].getCriticality() ; criticality++)
                    {
                        if (makespanAtCriticality[maxCritMakespan] < makespanAtCriticality[criticality] )
                        {
                            maxCritMakespan = criticality ;
                        }
                    }

                    instance[jobIndex].setStartTime(makespanAtCriticality[maxCritMakespan] + randomizer.Next(maxIdleTime + 1)) ;

                    for (int criticality = 1 ; criticality <= instance[jobIndex].getCriticality() ; criticality++)
                    {
                        makespanAtCriticality[criticality] = (int)instance[jobIndex].getCompletionTime(criticality) ;
                    }
                }

                /* The (possible) makespan of the currently generated instance. */
                double maxMakespan = double.MinValue ;

                for (jobIndex = 0 ; jobIndex < instance.getJobCount() ; jobIndex++)
                {
                    if (maxMakespan < instance[jobIndex].getLargestCompletionTime())
                    {
                        maxMakespan = instance[jobIndex].getLargestCompletionTime() ;
                    }
                }

                if (!Double.IsNaN(timeWindowSpan))
                {
                    /* Release times generation. */
                    for (jobIndex = 0 ; jobIndex < instance.getJobCount() ; jobIndex++)
                    {
                        int minimumReleaseTime = Math.Max(0, (int)(instance[jobIndex].getStartTime() - (timeWindowSpan - 1.0) * instance[jobIndex].getLargestExecutionTime())) ;
                        int maximumReleaseTime = (int)(instance[jobIndex].getStartTime()) ;

                        /* Sets the release time of the job. */
                        instance[jobIndex].setReleaseTime(randomizer.Next(minimumReleaseTime, maximumReleaseTime + 1)) ;

                        int minimumDeadline = (int)(instance[jobIndex].getLargestCompletionTime()) ;
                        int maximumDeadline = (int)(instance[jobIndex].getReleaseTime() + timeWindowSpan * instance[jobIndex].getLargestCompletionTime()) ;

                        /* Sets the deadline of the job. */
                        instance[jobIndex].setDeadline(randomizer.Next(minimumDeadline, maximumDeadline + 1)) ;
                    }
                }
                else
                {
                    /* Release times generation. */
                    for (jobIndex = 0 ; jobIndex < instance.getJobCount() ; jobIndex++)
                    {
                        /* Sets the release time of the job to be less or equal to its current start time. */
                        instance[jobIndex].setReleaseTime(randomizer.Next((int)instance[jobIndex].getStartTime() + 1)) ;
                    }

                    /* Deadlines generation. */
                    for (jobIndex = 0 ; jobIndex < instance.getJobCount() ; jobIndex++)
                    {
                        double completionTime = instance[jobIndex].getLargestCompletionTime() ;

                        /* Sets the deadline of the job to be greater or equal to its current largest completion time. */
                        instance[jobIndex].setDeadline((int)(completionTime + randomizer.NextDouble() * (maxMakespan - completionTime))) ;
                    }
                }

                /* Validates generated instance. */
                if (!validateGeneratedInstance(instance))
                {
                    throw new InvalidOperationException("The instance generator does not work correctly!!!") ;
                }

                instance.setMakeSpan(0.0) ;
                instance.setLateness(0.0) ;

                instance.MakespanDataView = "0" ;
                instance.LatenessDataView = "0" ;

                /* Adds the job to the current instance. */
                instanceList.Add(instance) ;
            }

            return instanceList ;
        }

        /**
         * Shuffles the specified permutation, i.e creating
         * random permutation out of it.
         * 
         * @param permutation The permutation to be shuffled.
         * @param randomizer  The random number generator to use.
         */
        protected static void shuffle(int[] permutation, CUniformRandom randomizer)
        {
            for (int i = permutation.Length - 1 ; i > 1 ; i--)
            {
                swap(ref permutation[i], ref permutation[randomizer.Next(i + 1)]) ;
            }
        }

        /**
         * Swaps the content of the specified variables.
         * 
         * @param x The first variable.
         * @param y The second variable.
         */
        protected static void swap(ref int x, ref int y)
        {
            int z = x ; x = y ; y = z ;
        }

        /**
         * Creates a random list of worst-case execution times
         * for a job of the specified criticality. The maximum
         * values of execution times are specified in
         * maximumExecutionTimes.
         * 
         * @param executionTimes        The resulting vector of exec.
         *                              times.
         * @param maximumExecutionTimes The maximum value of WCET per
         *                              criticality level.
         * @param criticality           The maximum (zero-based) 
         *                              criticality to be set.
         * @param randomizer            The random number generator to
         *                              use.
         */
        public static void createExecutionTimes(double[] executionTimes, int[] maximumExecutionTimes, int criticality, CUniformRandom randomizer)
        {
            if (criticality > 0)
            {
                executionTimes[0] = randomizer.Next(maximumExecutionTimes[0] + 1) ;
            }

            for (int i = 1 ; i < criticality ; i++)
            {
                executionTimes[i] = randomizer.Next((int)executionTimes[i - 1], maximumExecutionTimes[i] + 1) ;
            }
        }

        /**
         * An auxiliary method which determines a number
         * of digits in the specified number, and according
         * to that creates a format string for MC-instances
         * name suffix.
         * 
         * @param number The number whose digits are counted.
         * 
         * @return The format string for instance name suffix.
         */
        public static string determineInstanceNameSuffix(int number)
        {
            int result = (number != 0 ? 0 : 1)  ;

            while (number != 0)
            {
                number /= 10 ;
                ++result ;
            }

            return ("{0:D" + result + "}") ;
        }

        /**
         * Validates the specified mixed-criticality instance. The jobs
         * in the instance is assumed to be in order in which they appear
         * in the feasible schedule from which the instance has been generated.
         * 
         * @param instance The instance to be validated.
         * 
         * @return Returns true if the given (generated) instance is specified
         *         correctly, if not, it returns false.
         */
        public static bool validateGeneratedInstance(CMixedCriticalityInstance instance)
        {
            for (int i = 0 ; i < instance.getJobCount() ; i++)
            {
                /* Non-decreasing processing times validation. */
                for (int criticality = 2 ; criticality <= instance[i].getCriticality() ; criticality++)
                {
                    if (instance[i].getExecutionTime(criticality - 1) > instance[i].getExecutionTime(criticality))
                    {
                        return false ;
                    }
                }

                /* Release times and deadlines validations. */
                if (instance[i].getReleaseTime() > instance[i].getStartTime() || instance[i].getLargestCompletionTime() > instance[i].getDeadline())
                {
                    return false ;
                }                
            }

            int[] completionTimeAtCriticality = new int[instance.getMaximumCriticality() + 1] ;

            /* The first scheduled job. */
            CMixedCriticalityJob job = instance[0] ;

            if (job.getStartTime() < job.getReleaseTime())
            {
                return false ;
            }

            for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
            {
                completionTimeAtCriticality[criticality] = (int)job.getCompletionTime(criticality) ;
            }

            /* Start times validation - important for existance 
             * of at least one feasible schedule.
             */
            for (int jobIndex = 1 ; jobIndex < instance.getJobCount() ; jobIndex++)
            {
                int maximumCompletionTimeCriticality = 1 ;

                /* The next scheduled job. */
                job = instance[jobIndex] ;

                /* Determines the lower bound for the start time of the next job. */
                for (int criticality = 2 ; criticality <= job.getCriticality() ; criticality++)
                {
                    if (completionTimeAtCriticality[maximumCompletionTimeCriticality] < completionTimeAtCriticality[criticality] )
                    {
                        maximumCompletionTimeCriticality = criticality ;
                    }
                }

                if (job.getStartTime() < Math.Max(completionTimeAtCriticality[maximumCompletionTimeCriticality], job.getReleaseTime()))
                {
                    return false ;
                }

                /* Updates the completion times after the next job has been processed. */
                for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
                {
                    completionTimeAtCriticality[criticality] = (int)job.getCompletionTime(criticality) ;
                }
            }

            return true ;
        }
    }
}
