﻿using System ;

/**
 * Contains the implementation of a mixed-criticality solver based on the clairvoyant non-preemptive
 * earliest deadline first algorithm.
 */
namespace MCScheduling.MixedCriticality.CEDF
{
    /**
     * The mixed-criticality scheduling problem solver based on the clairvoyant 
     * non-preemptive EDF scheduling algorithm.
     * 
     * @author Tomas Tunys
     */
    class CClairvoyantEDFSolver : CBaseMixedCriticalitySolver
    {
        /** The CEDF algorithm implementation. */
        private ClairvoyantEDF.ClairvoyantEDF solver = new ClairvoyantEDF.ClairvoyantEDF() ;

        /**
         * Solves the specified mixed-criticality instance using Clairvoyant
         * Non-preemptive EDF Scheduling algorithm.
         * 
         * @param instance The mixed-criticality instance to be solved.
         */
        public override void solve(CMixedCriticalityInstance instance)
        {
            solver.solve(instance) ;

            calculateMakespanAndLateness(instance) ;
        }

        /**
         * Ignored since Clairvoyant EDF algorithm runs in O(nlogn)
         * time which is fast enough to finish up any instance of a
         * reasonable size.
         */
        public override void interrupt()
        {
        }

        /**
         * Calculates the makespan and lateness of the specified mixed-criticality
         * instance. It is assumed that it contains valid start time values.
         * 
         * @param instance The mixed-criticality instance whose makespan and lateness
         *                 is calculated.
         */
        private void calculateMakespanAndLateness(CMixedCriticalityInstance instance)
        {
            double makespan = double.MinValue ;
            double lateness = 0.0 ;

            foreach (CMixedCriticalityJob job in instance)
            {
                lateness += Math.Max(0, job.getLargestCompletionTime() - job.getDeadline()) ;
                
                if (makespan < job.getLargestCompletionTime())
                {
                    makespan = job.getLargestCompletionTime() ;
                }
            }

            /* Sets the makespan. */
            instance.setMakeSpan(makespan) ;

            /* Sets the lateness. */
            instance.setLateness(lateness) ;

            /* Feasible solution found. */
            if (lateness < 0.00001)
            {
                setStatus(Status.SUBOPTIMAL) ;
            }
            else
            {
                setStatus(Status.HEURISTIC) ;
            }
        }
    }
}
