﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Globalization;

namespace MCScheduling.MixedCriticality
{
    /**
     * Representation of a mixed-criticality instance.
     */
    public class CMixedCriticalityInstance : IEnumerable, ICloneable
    {
        /** The name of this instance. */
        private string m_name ;

        /** The jobs in this mixed-criticality instance. */
        private List<CMixedCriticalityJob> m_jobs ;

        /** The makespan of this mixed-criticality instance. */
        private double m_makespan ;

        /** The lateness of this mixed-criticality instance. */
        private double m_lateness ;

        /** The maximum criticality level of this instance. */
        private int m_maximumCriticality ;

        /** The cpu time spent on solving this instance. */
        private int m_cpuTime ;

        /**
         * Creates untitled and empty mixed-criticality instance.
         */
        public CMixedCriticalityInstance() : this("untitled")
        {
        }

        /**
         * Creates an empty mixed-criticality instance
         * with the specified name.
         * 
         * @param name The name of this instance.
         */
        public CMixedCriticalityInstance(string name)
        {
            m_name = name ;
            m_jobs = new List<CMixedCriticalityJob>() ;
        }

        /**
         * Access the <em>index</em>-job of this mixed-criticality
         * instance.
         * 
         * @param index The index of the job to be returned.
         */
        public CMixedCriticalityJob this[int index]
        {
            get { return m_jobs[index] ; }
        }

        /**
         * Adds the specified job to this mixed-criticality
         * instance.
         * 
         * @param job The job to be added.
         */
        public void addJob(CMixedCriticalityJob job)
        {
	        m_jobs.Add(job) ;

            /* Updates the maximum criticality. */
            if (job.getCriticality() > m_maximumCriticality)
            {
                m_maximumCriticality = job.getCriticality() ;
            }
        }

        /**
         * Removes the specified job from this mixed-criticality
         * instance.
         * 
         * @param job The job to be removed.
         * 
         * @return True, if item was successfully removed, otherwise, false. 
         *         This method also returns false if item was not found 
         *         in this mixed-criticality instance.
         */
        public void removeJob(CMixedCriticalityJob job)
        {
            /* Removes the job and updates the maximum criticality (if needed). */
	        if (m_jobs.Remove(job) && job.getCriticality() == m_maximumCriticality)
            {
                m_maximumCriticality = 1 ;

                foreach (CMixedCriticalityJob j in this)
                {
                    if (m_maximumCriticality < j.getCriticality())
                    {
                        m_maximumCriticality = j.getCriticality() ;
                    }
                }
            }
        }

        /**
         * Returns the maximum criticality level of 
         * this instance.
         * 
         * @return The maximum criticality level of 
         *         this instance.
         */
        public int getMaximumCriticality()
        {
            return m_maximumCriticality ;
        }

        /**
         * Sets the makespan for this instance.
         * 
         * @param makespan The makespan to be set.
         */
        public void setMakeSpan(double makespan)
        {
	        m_makespan = makespan ;
        }

        /**
         * Returns the makespan of this instance.
         * 
         * @return The makespan of this instance.
         */
        public double getMakeSpan()
        {
	        return m_makespan ;
        }

        /**
         * Sets the lateness for this instance.
         * 
         * @param lateness The lateness to be set.
         */
        public void setLateness(double lateness)
        {
            m_lateness = lateness ;
        }

        /**
         * Returns the lateness of this instance.
         * 
         * @return The lateness of this instance.
         */
        public double getLateness()
        {
            return m_lateness ;
        }

        /**
         * Returns the start times of the jobs in this
         * mixed-criticality instance.
         * 
         * @return The start times of the jobs in this instance.
         */
        public double[] getStartTimes()
        {
	        double[] startTimes = new double[m_jobs.Count] ;

	        for (int i = 0 ; i < m_jobs.Count ; i++)
	        {
		        startTimes[i] = m_jobs[i].getStartTime() ;
	        }

	        return startTimes ;
        }

        /**
         * Sets the name of this mixed-criticality instance.
         * 
         * @param name The name of this instance.
         */
        public void setName(string name)
        {
            m_name = name ;
        }

        /**
         * Returns the name of this mixed-criticality instance.
         * 
         * @return The name of this instance.
         */
        public string getName()
        {
            return m_name ;
        }

        /**
         * Returns an array of jobs in this mixed-criticality
         * instance.
         * 
         * @return The jobs' start times.
         */
        public CMixedCriticalityJob[] getJobs()
        {
	        return m_jobs.ToArray() ;
        }

        /**
         * Returns the number of jobs in this mixed-criticality
         * instance.
         * 
         * @return The number of jobs in this mixed-criticality
         *         instance.
         */
        public int getJobCount()
        {
	        return m_jobs.Count ;
        }

        /**
         * Sets the time spent on solving this instance.
         * 
         * @param miliseconds The time in miliseconds spent on solving 
         *                    this instance.
         */
        public void setCPUTime(int miliseconds)
        {
            m_cpuTime = miliseconds ;
        }

        /**
         * Returns the time in miliseconds that has been spent
         * on solving the instance.
         * 
         * @return The time in miliseconds that has been spent
         *         on solving the instance.
         */
        public int getCPUTime()
        {
            return m_cpuTime ;
        }

        /**
         * Returns a string representation of this mixed-criticality
         * instance. 
         * <p>
         * For debugging purpose only.
         * </p>
         * 
         * @return A string representing this mixed-criticality 
         *         instance.
         */
        public override string ToString()
        {
            StringBuilder output = new StringBuilder() ;

            output.Append("Mixed-criticality instance - " + m_name  + ":\n") ;

	        for (int i = 0 ; i < getJobCount() ; i++)
	        {
		        output.Append(m_jobs[i].ToString()) ;
                output.Append("\n") ;
	        }

	        return output.ToString() ;
        }

        /**
         * Returns this mixed-criticality instance
         * in output file format.
         */
        public string ToFileOutput()
        {
            int maxCriticality = int.MinValue ;

            foreach (CMixedCriticalityJob job in m_jobs)
            {
                if (maxCriticality < job.getCriticality())
                {
                    maxCriticality = job.getCriticality() ;
                }
            }

            StringBuilder output = new StringBuilder() ;

            output.Append(getJobCount()) ;
            output.Append(' ') ;
            output.Append(maxCriticality) ;
            output.Append('\n') ;

            foreach (CMixedCriticalityJob job in m_jobs)
            {
                output.Append(job.getName()) ;
                output.Append(' ') ;
                if (double.IsInfinity(job.getReleaseTime()))
                {
                    output.Append("-Inf") ;
                }
                else
                {
                    output.Append((int)job.getReleaseTime()) ;
                }
                output.Append(' ') ;
                if (double.IsInfinity(job.getDeadline()))
                {
                    output.Append("Inf") ;
                }
                else
                {
                    output.Append((int)job.getDeadline()) ;
                }
                output.Append(' ') ;
                output.Append(job.getCriticality()) ;

                int criticality = 1 ;

                while (criticality <= job.getCriticality())
                {
                    output.Append(' ') ;
                    output.Append(job.getExecutionTime(criticality)) ;
                    ++criticality ;
                }

                while (criticality <= maxCriticality)
                {
                    output.Append(' ') ;
                    output.Append(0) ;
                    ++criticality ;
                }

                output.Append(' ') ;

                if (double.IsInfinity(job.getStartTime()) || double.IsNaN(job.getStartTime()))
                {
                    output.Append(0) ;
                }
                else
                {
                    output.Append((int)job.getStartTime()) ;
                }

                output.Append('\n') ;
            }

            return output.ToString() ;
        }

        /**
         * Returns this mixed-criticality instance
         * in solution output file format.
         */
        public string ToSolutionFileOutput()
        {
            StringBuilder output = new StringBuilder() ;

            output.Append(String.Format(CultureInfo.InvariantCulture, "{0:0.##}", getMakeSpan())) ;
            output.Append(' ') ;
            output.Append(String.Format(CultureInfo.InvariantCulture, "{0:0.##}", getLateness())) ;
            output.Append('\n') ;

            foreach (CMixedCriticalityJob job in m_jobs)
            {
                output.Append(job.getName()) ;
                output.Append(' ') ;
                
                if (double.IsInfinity(job.getStartTime()) || double.IsNaN(job.getStartTime()))
                {
                    output.Append("unknown") ;
                }
                else
                {
                    output.Append((int)job.getStartTime()) ;
                }

                output.Append('\n') ;
            }

            return output.ToString() ;
        }

        /**
         * Returns the enumerator for this mixed-criticality
         * instance jobs.
         * 
         * @return The job enumberator.
         */
        IEnumerator IEnumerable.GetEnumerator()
        {
           return (IEnumerator)GetEnumerator() ;
        }

        /**
         * Returns the enumerator for this mixed-criticality
         * instance jobs.
         * 
         * @return The job enumberator.
         */
        public IEnumerator GetEnumerator()
        {
            return m_jobs.GetEnumerator() ;
        }

        /**
         * Returns the copy of this mixed-criticality instance.
         * 
         * @return The copy of this mixed-criticality instance.
         */
        public object Clone()
        {
            CMixedCriticalityInstance copy = new CMixedCriticalityInstance(getName()) ;

            foreach (CMixedCriticalityJob job in this)
            {
                copy.addJob((CMixedCriticalityJob)job.Clone()) ;
            }

            return (CMixedCriticalityInstance)copy ;
        }
 
        #region >>> The Mixed-Criticality Instance Properties. <<<

        /** The makespan property used to display in data views. */
        private string MakespanDataViewProperty = "" ;

        /** The makespan property used to display in data views. */
        private string LatenessDataViewProperty = "" ;

        /** The Name Data View property. */
        public string NameDataView
        {
            set { setName(value) ; }
            get { return getName() ; }
        }

        /** The Makespan property. */
        public string MakespanDataView
        {
            set { MakespanDataViewProperty = value ; }
            get { return MakespanDataViewProperty ; }
        }

        /** The Lateness property. */
        public string LatenessDataView
        {
            set { LatenessDataViewProperty = value ; }
            get { return LatenessDataViewProperty ; }
        }

        /** The MaximumCriticality property. */
        public int MaximumCriticalityDataView
        {
            get { return getMaximumCriticality() ; }
        }

        /** The JobCount property. */
        public int JobCountDataView
        {
            get { return getJobCount() ; }
        }

        /** The CPU time property. */
        public double CPUTimeDataView
        {
            get { return ((double)getCPUTime() / 1000)  ; }
        }

        /** The SolutionStatus property. */
        public string SolutionStatusDataView
        {
            get ; set ;
        }

        #endregion >>> The Mixed-Criticality Instance Properties. <<<

        #region >>> An Instance Solution Validator <<<

        private class StartTimeComparer : System.Collections.Generic.Comparer<int>
        {
            private CMixedCriticalityInstance instance ;

            public StartTimeComparer(CMixedCriticalityInstance instance)
            {
                this.instance = instance ;
            }

            public override int Compare(int i, int j)
            {
                return decimal.ToInt32((decimal)(Math.Round(instance[i].getStartTime() - instance[j].getStartTime()))) ;
            }
        }

        /**
         * Validates the correctness of the specified instance.
         * 
         * @param instance The instance to be validated.
         * 
         * @return True if the specified instance is correct, that is the start
         *         time of each job respects the release time and no two jobs
         *         overlap (i.e. are not being executed at the same time on the
         *         same critical level).
         */
        public static bool validateInstance(CMixedCriticalityInstance instance)
        {
            int[] jobIndices = new int[instance.getJobCount()] ;

            for (int i = 0 ; i < instance.getJobCount() ; i++)
            {
                jobIndices[i] = i ;

                /* Release times x Start times validation. */
                if (instance[i].getReleaseTime() > instance[i].getStartTime() + 10e-5)
                {
                    return false ;
                }                
            }

            Array.Sort(jobIndices, new StartTimeComparer(instance)) ;

            /* The first scheduled job. */
            CMixedCriticalityJob job = null, nextJob = null ;

            /* Start times validation. */
            for (int jobIndex = 0 ; jobIndex < instance.getJobCount() ; jobIndex++)
            {
                /* The next scheduled job. */
                job = instance[jobIndices[jobIndex]] ;

                /* Determines the lower bound for the start time of the next job. */
                for (int nextJobIndex = jobIndex + 1 ; nextJobIndex < instance.getJobCount() ; nextJobIndex++)
                {
                    nextJob = instance[jobIndices[nextJobIndex]] ;

                    if (nextJob.getStartTime() <= job.getLargestCompletionTime())
                    {
                        if (nextJob.getStartTime() + 10e-5 < job.getCompletionTime(Math.Min(job.getCriticality(), nextJob.getCriticality())) &&
                            nextJob.getExecutionTime(Math.Min(job.getCriticality(), nextJob.getCriticality())) != 0.0)
                        {
                            return false ;
                        }
                    }
                    else
                    {
                        break ;
                    }
                }
            }

            return true ;
        }
    }
    #endregion >>> An Instance Solution Validator <<<
}
