﻿using System;
using System.IO ;
using System.Linq ;
using System.Text ;
using System.Collections.Generic ;
using System.Globalization ;

namespace MCScheduling.MixedCriticality
{
    /**
     * CMixedCriticalityInstancesSerializer provides a methods to serialize
     * and deserialize mixed-criticality instances. 
     */
    class CMixedCriticalityInstancesSerializer
    {
        /** The extension of the file with a mixed-criticality instance. */
        private static string INSTANCE_FILE_EXTENSION = ".ins" ;

        /** The extension of the file with a mixed-criticality instance solution. */
        private static string SOLUTION_FILE_EXTENSION = ".sol" ;

        /**
         * Saves collection of the specified instances into the 
         * specified path. A directory with the specified name
         * of the collection will be created and filled with
         * serialized mixed-criticality instances.
         * 
         * @param directoryPath  The directory path where the collection 
         *                       is going to be saved.
         * @param collectionName The name of the collection.
         * @param instances      The instances to save.
         */
        public static void saveCollection(string directoryPath, string collectionName, IList<CMixedCriticalityInstance> instances)
        {
            string collectionDirectoryPath = Path.Combine(directoryPath, collectionName) ;

            /* Creates a directory for the collection. */
            Directory.CreateDirectory(collectionDirectoryPath) ;

            foreach (CMixedCriticalityInstance instance in instances)
            {
                using (StreamWriter fileWriter = new StreamWriter(Path.Combine(collectionDirectoryPath, instance.getName() + INSTANCE_FILE_EXTENSION)))
                {
                    fileWriter.Write(instance.ToFileOutput()) ;
                }
            }
        }

        /**
         * Loads a collection of mixed-criticality instances
         * from the specified directory path.
         */
        public static List<CMixedCriticalityInstance> loadCollection(string directoryPath)
        {
            List<CMixedCriticalityInstance> loadedInstances = new List<CMixedCriticalityInstance>() ;

            foreach (string fileName in Directory.EnumerateFiles(directoryPath, "*.ins"))
            {
                using (StreamReader fileReader = new StreamReader(fileName))
                {
                    char[] separator = new char[1] ; separator[0] = ' ' ;

                    string[] parsedLine = fileReader.ReadLine().Split(separator, 2) ;

                    /* The job count is the 1st argument on the 1st line. */
                    int jobCount = int.Parse(parsedLine[0]) ;

                    /* The maximum criticality is the 2nd argument on the 1st line. */
                    int maximumCriticality = int.Parse(parsedLine[1]) ;

                    /* A mixed-criticality instance loaded from the file. */
                    CMixedCriticalityInstance instance = new CMixedCriticalityInstance(Path.GetFileNameWithoutExtension(fileName)) ;

                    /* A list for execution times. */
                    double[] executionTimes = new double[maximumCriticality] ;
                    double   makespan       = 0.0 ;

                    /* Parses the mixed-criticality jobs from the file. */
                    for (int iJob = 0 ; iJob < jobCount ; iJob++)
                    {
                        parsedLine = fileReader.ReadLine().Split(separator, 5 + maximumCriticality) ;

                        string name        = parsedLine[0] ;
                        double releaseTime = (String.Compare(parsedLine[1], "-Inf", true) == 0 ? double.NegativeInfinity : double.Parse(parsedLine[1], CultureInfo.InvariantCulture)) ;
                        double deadline    = (String.Compare(parsedLine[2], "Inf", true) == 0 ? double.PositiveInfinity : double.Parse(parsedLine[2], CultureInfo.InvariantCulture)) ;
                        int    criticality = int.Parse(parsedLine[3]) ;

                        for (int i = 0, j = 4 ; i < criticality ; i++, j++)
                        {
                            executionTimes[i] = double.Parse(parsedLine[j], CultureInfo.InvariantCulture) ;
                        }

                        CMixedCriticalityJob job = new CMixedCriticalityJob(name, releaseTime, deadline, criticality, executionTimes) ;

                        /* The start time is the last parameter on the line. */
                        job.setStartTime(double.Parse(parsedLine[4 + maximumCriticality], CultureInfo.InvariantCulture)) ;

                        if (makespan < job.getLargestCompletionTime())
                        {
                            makespan = job.getLargestCompletionTime() ;
                        }

                        instance.addJob(job) ;
                    }

                    if (makespan > 0.0)
                    {
                        instance.MakespanDataView = makespan.ToString() ;
                    }

                    loadedInstances.Add(instance) ;
                }
            }

            return loadedInstances ;
        }

        /**
         * Saves the solutions of the specified instances to the specified directory.
         *
         * @param directoryPath The path of the directory where the solutions will be
         *                      saved.
         * @param instances     The instances whose solutions will be saved.
         */
        public static void saveSolution(string directoryPath, IList<CMixedCriticalityInstance> instances)
        {
            /* Creates a directory for the collection of solutions. */
            if (!Directory.Exists(directoryPath))
            {
                Directory.CreateDirectory(directoryPath) ;
            }

            foreach (CMixedCriticalityInstance instance in instances)
            {
                using (StreamWriter fileWriter = new StreamWriter(Path.Combine(directoryPath, instance.getName() + SOLUTION_FILE_EXTENSION)))
                {
                    fileWriter.Write(instance.ToSolutionFileOutput()) ;
                }
            }
        }

        /**
         * Deserializes, i.e., rebuilds a mixed-criticality instance from data
         * in the specified file.
         * 
         * @param fileName The file name where an MC instance will be rebuilt from.
         * 
         * @return The deserialized MC instance.
         */
        public static CMixedCriticalityInstance deserialize(string fileName)
        {
            using (StreamReader fileReader = new StreamReader(fileName))
            {
                char[] separator = new char[1] ; separator[0] = ' ' ;

                string[] parsedLine = fileReader.ReadLine().Split(separator, 2) ;

                /* The job count is the 1st argument on the 1st line. */
                int jobCount = int.Parse(parsedLine[0]) ;

                /* The maximum criticality is the 2nd argument on the 1st line. */
                int maximumCriticality = int.Parse(parsedLine[1]) ;

                /* A mixed-criticality instance loaded from the file. */
                CMixedCriticalityInstance instance = new CMixedCriticalityInstance(Path.GetFileNameWithoutExtension(fileName)) ;

                /* A list for execution times. */
                double[] executionTimes = new double[maximumCriticality] ;

                /* Parses the mixed-criticality jobs from the file. */
                for (int iJob = 0 ; iJob < jobCount ; iJob++)
                {
                    parsedLine = fileReader.ReadLine().Split(separator, 5 + maximumCriticality) ;

                    string name        = parsedLine[0] ;
                    double releaseTime = (String.Compare(parsedLine[1], "-Inf", true) == 0 ? double.NegativeInfinity : double.Parse(parsedLine[1], CultureInfo.InvariantCulture)) ;
                    double deadline    = (String.Compare(parsedLine[2], "Inf", true) == 0 ? double.PositiveInfinity : double.Parse(parsedLine[2], CultureInfo.InvariantCulture)) ;
                    int    criticality = int.Parse(parsedLine[3]) ;
                    
                    /* The start time is the last parameter on the line. */
                    double startTime   = double.Parse(parsedLine[4 + maximumCriticality], CultureInfo.InvariantCulture) ;

                    for (int i = 0, j = 4 ; i < criticality ; i++, j++)
                    {
                        executionTimes[i] = double.Parse(parsedLine[j], CultureInfo.InvariantCulture) ;
                    }

                    CMixedCriticalityJob job = new CMixedCriticalityJob(name, releaseTime, deadline, startTime, criticality, executionTimes) ;

                    instance.addJob(job) ;
                }

                return instance ;
            }
        }
    }
}
