﻿using System ;
using System.Linq ;
using System.Text ;
using System.Globalization ;
using System.Collections.Generic ;

namespace MCScheduling.MixedCriticality
{
    /**
     * A CJob instance represents a mixed-criticality job.
     */
    public class CMixedCriticalityJob : ICloneable
    {
        /** The name of the job. */
	    private string  m_name ;

        /** The job's start time. */
	    private double m_startTime ;

        /** The job's release time. */
	    private double m_releaseTime ;

        /** The job's deadline. */
	    private double m_deadline ;

        /** The job's criticality. */
	    private int m_criticality ;

        /** The job's worst-case execution times for each criticality level. */
	    private double[] m_executionTimes ;

        /**
         * Creates a new mixed-criticality job instance.
         * 
         * @param name           The job's name.
         * @param criticality    The job's criticality level.
         * @param executionTimes The worst-case exectution time for each level of criticality.
         */
	    public CMixedCriticalityJob(string name, int criticality, double[] executionTimes) : this(name, double.NegativeInfinity, double.PositiveInfinity, criticality, executionTimes)
        {
        }

        /**
         * Creates a new mixed-criticality job instance.
         * 
         * @param name           The job's name.
         * @param releaseTime    The job's release time.
         * @param deadline       The job's deadline.
         * @param criticality    The job's criticality level.
         * @param executionTimes The worst-case exectution time for each level of criticality.
         */
	    public CMixedCriticalityJob(string name, double releaseTime, double deadline, int criticality, double[] executionTimes) : this(name, releaseTime, deadline, double.NaN, criticality, executionTimes)
        {
        }

        /**
         * Creates a new mixed-criticality job instance.
         * 
         * @param name           The job's name.
         * @param releaseTime    The job's release time.
         * @param deadline       The job's deadline.
         * @param startTime      The job's start time.
         * @param criticality    The job's criticality level.
         * @param executionTimes The worst-case exectution time for each level of criticality.
         */
        public CMixedCriticalityJob(string name, double releaseTime, double deadline, double startTime, int criticality, double[] executionTimes)
        {
            setName(name) ;
            setStartTime(startTime) ;
            setReleaseTime(releaseTime) ;
            setDeadline(deadline) ;
            setCriticality(criticality) ;

            m_executionTimes = new double[criticality] ;

            setExecutionTimes(executionTimes) ;
        }

        /**
         * Returns the job's name.
         * 
         * @return The job's name.
         */
	    public string getName()
        {
            return m_name ;
	    }

        /**
         * Sets a name for the job.
         * 
         * @param name A job's new name.
         */
        void setName(string name)
        {
            m_name = name ;
        }

        /**
         * Returns the job's start time.
         * 
         * @return The job's start time.
         */
	    public double getStartTime() 
        {
            return m_startTime ;
        }

        /**
         * Sets the start time for the job.
         * 
         * @param startTime The job's start time.
         */
	    public void setStartTime(double startTime)
        {
            m_startTime = startTime ;
        }

        /**
         * Returns the job's completion time for the specified
         * criticality level. If the job has not been assigned
         * any start time, the returned value is <em>NaN</em>.
         * 
         * @param criticality The level of criticality for which
         *                    the completion time is requested.
         *
         * @return The job's completion time for the specified
         *         criticality level.
         */
	    public double getCompletionTime(int criticality)
        {
            return getStartTime() + getExecutionTime(criticality) ;
        }

        /**
         * Returns the largest completion time of this job.
         * 
         * @return The largest completion time of this job.
         */
	    public double getLargestCompletionTime() 
        {
            return getCompletionTime(getCriticality()) ;
        }
 
        /**
         * Returns the job's release time.
         * 
         * @return The job's release time.
         */
	    public double getReleaseTime()
        {
            return m_releaseTime ;
        }

        /**
         * Sets the job's release time.
         * 
         * @param releaseTime The job's release time.
         */
	    public void setReleaseTime(double releaseTime)
        {
            m_releaseTime = releaseTime ;
        }
        
        /**
         * Returns the deadline of the job.
         * 
         * @return The job's deadline.
         */
	    public double getDeadline() 
        {
            return m_deadline ;
        }

        /**
         * Sets the deadline for the job.
         * 
         * @param deadline The job's new deadline.
         */
	    public void setDeadline(double deadline)
        {
            m_deadline = deadline ;
        }
        
        /**
         * Returns the criticality of the job.
         * 
         * @return The job's criticality.
         */
	    public int getCriticality() 
        {
            return m_criticality ;
        }

        /**
         * Sets the criticality for the job.
         * 
         * @param criticality The job's criticality level.
         * 
         * @throws ArgumentOutOfRangeException If the criticality is less
         *                                     then 1.
         */
	    public void setCriticality(int criticality)
        {
            if (criticality < 1)
            {
                throw new ArgumentOutOfRangeException("The criticality has to be at least 1!") ;
            }
            m_criticality = criticality ;
        }
        
        /**
         * Returns the worst-case execution times
         * of this job.
         * 
         * @return The job's WCETs.
         */
	    public double[] getWCETs() 
        {
            return m_executionTimes ;
        }

        /**
         * Sets the worst-case execution times for this job. It is expected
         * that the job has assigned valid criticality (using this.setCriticality
         * method) prior to calling this method, and the length of the specified
         * array has to be at least as big as the criticality of the job.
         * 
         * @param executionTimes The execution times to be set for the job.
         */
        public void setExecutionTimes(double[] executionTimes)
        {
            Array.Copy(executionTimes, m_executionTimes, m_criticality) ;
        }

        /**
         * Sets the worst-case execution times for this job. It sets the WCETs 
         * from the criticality <em>fromCriticality</em> to the criticality
         * <em>toCriticality</em> inclusive. The <em>executionTimes</em> should
         * contain new WCETs, and be at least as long as 
         * (<em>toCriticality - fromCriticality + 1</em>).
         * 
         * @param fromCriticality The lowest crit. level to be set.
         * @param toCriticality   The highest crit. level to be set.
         * @param executionTimes  The new WCETs for the job.
         */
	    public void setExecutionTimes(int fromCriticality, int toCriticality, double[] executionTimes)
        {
            Array.Copy(executionTimes, 0, m_executionTimes, fromCriticality - 1, toCriticality - (fromCriticality - 1)) ;
        }            
        
        /**
         * Returns the largest WCET of this job. Since the WCETs are
         * expected to be non-decreasing, this is equivalent to
         * <em>this.getExecutionTime(this.getCriticality())</em>.
         * 
         * @return The job's largest worst-case execution time.
         */
	    public double getLargestExecutionTime()
        {
            return m_executionTimes[m_criticality - 1] ;
        }
        
        /**
         * Returns the worst-case execution time of this job 
         * on the specified criticality level.
         * 
         * @param criticality The WCET on this criticality.
         * 
         * @return The worst-case execution time on the given 
         *         criticality level.
         * 
         * @throws ArgumentOutOfRangeException If the specified criticality
         *                                     is less then 1 or greater
         *                                     then the job's criticality.
         */
	    public double getExecutionTime(int criticality)
        {
            if (criticality < 1 || criticality > getCriticality())
            {
                throw new ArgumentOutOfRangeException() ;
            }

            return m_executionTimes[criticality - 1] ;
        }

        /**
         * Sets the worst-case execution time for the specified
         * criticality level.
         * 
         * @param criticality   The criticality level of a WCET to be set.
         * @param executionTime The new WCET's value.
         * 
         * @throws ArgumentOutOfRangeException If the specified criticality
         *                                     is less then 1 or greater
         *                                     then the job's criticality.
         *
         */
	    public void setExecutionTime(int criticality, double executionTime)
        {
            if (criticality < 1 || criticality > getCriticality())
            {
                throw new ArgumentOutOfRangeException() ;
            }

            m_executionTimes[criticality - 1] = executionTime ;
        }

        /**
         * Determines whether the specified object is equal
         * to this job.
         * 
         * @param o The object to compare with this job.
         * 
         * @return True, if the specified object is equal to this job,
         *         otherwise, false.
         */
        public override bool Equals(object o)
        {
            return m_name.Equals(o) ;
        }

        /**
         * Returns a hash code of this job.
         * 
         * @return The job's hash code.
         */
        public override int GetHashCode()
        {
            return m_name.GetHashCode() ;
        }

        /**
         * Returns a copy of this job.
         * 
         * @return A copy of this job.
         */
        public object Clone()
        {
 	        return new CMixedCriticalityJob(m_name, m_releaseTime, m_deadline, m_startTime, m_criticality, (double[])m_executionTimes.Clone()) ;
        }

        /**
         * Returns a string representation of this job.
         * <p>
         * For debugging purpose only.
         * </p>
         * 
         * @return A string representation of this job.
         */
        public override string ToString()
        {
            StringBuilder output = new StringBuilder() ;

            output.Append(getName()) ;
            output.Append("=(") ;
            output.AppendFormat(CultureInfo.InvariantCulture, "{0:0.##},", getReleaseTime()) ;
            output.AppendFormat(CultureInfo.InvariantCulture, "{0:0.##},", getDeadline()) ;
            output.AppendFormat(CultureInfo.InvariantCulture, "{0},[{1:0.##}", getCriticality(), getExecutionTime(1)) ;

            for (int criticality = 2 ; criticality <= getCriticality() ; criticality++)
            {
                output.AppendFormat(CultureInfo.InvariantCulture, ",{0:0.##}", getExecutionTime(criticality)) ;
            }

            output.Append("])") ;

            return output.ToString() ;
        }
    }
}
