﻿using System ;
using System.Collections.Generic ;
using MCScheduling.Utils ;

namespace MCScheduling.MixedCriticality
{
    /**
     * Generates a list of random MC instances.
     */
    class CRandomInstanceGenerator
    {
        /**
         * Returns a list of generated MC instances. Each instance will be named with the specified instanceNamePrefix following with its ordinal number.
         * The number of MC instances to generated is instanceCount and the number of jobs in an instance is specified by jobsCount. Each parameter of
         * an MC job is assigned randomly as follows: The criticality levels of jobs are distributed according to the jobPerCriticalityProb, the execution
         * times maxima per crit. level are given in maximumExecutionTimes and the release times are selected randomly from [0, releaseTime] and deadlines
         * for each job as [ri, ri + deadline].
         *  
         * @param instanceNamePrefix    The prefix of the generated instances names.
         * @param instanceCount         The number of MC instances to generate.
         * @param jobsCount             The number of jobs in each MC instance.
         * @param jobPerCriticalityProb The distribution of the MC job criticality levels.
         * @param maximumExecutionTimes The maximum random value of an exectution time per level.
         * @param releaseTime           The release time window. Each job's release time is set to
         *                              a number in the interval [0, releaseTime].
         * @param deadline              The deadline span. Each job's deadline is se to a number
         *                              in the interval [r, r + deadline], where r is its release time.
         * 
         * @return The list of randomly generated MC instances.
         */
        public static List<CMixedCriticalityInstance> generate(string instanceNamePrefix, int instanceCount, int jobsCount, double[] jobPerCriticalityProb, int[] maximumExecutionTimes, int releaseTime, int deadline)
        {
            /** A pseudo-random number generator. */
            CUniformRandom randomizer = new CUniformRandom((ulong)DateTime.Now.Ticks) ;

            /* The total number of jobs to generate. */
            int jobIndex = 0 ;

            /* The list of generated instances. */
            List<CMixedCriticalityInstance> instanceList = new List<CMixedCriticalityInstance>(instanceCount) ;

            /* A list of randomly generated execution times for a job. */
            double[] executionTimes = new double[jobPerCriticalityProb.Length] ;

            string nameSuffixFormat = CDefaultMixedCriticalityInstanceGenerator.determineInstanceNameSuffix(instanceCount) ;
            
            /* Generates random mixed-criticality instances. */
            for (int iInstance = 1 ; iInstance <= instanceCount ; iInstance++)
            {
                /* A randomly generated instance. */
                CMixedCriticalityInstance instance = new CMixedCriticalityInstance(instanceNamePrefix + String.Format(nameSuffixFormat, iInstance)) ;

                /* Job generation loop. */
                for (jobIndex = 0 ; jobIndex < jobsCount ; jobIndex++)
                {
                    /* The generated job's criticality. */
                    int jobCriticality = -1 ;

                    /* The random number that determines the job's criticality level. */
                    double rand = randomizer.NextDouble() ;

                    /* Creates a random assignement of the job's criticality. */
                    for (int i = 0 ; jobCriticality == -1 && i < jobPerCriticalityProb.Length ; i++)
                    {
                        if (rand > jobPerCriticalityProb[i])
                        {
                            rand -= jobPerCriticalityProb[i] ;
                        }
                        else
                        {
                            jobCriticality = i + 1 ;
                        }
                    }

                    /* Generates execution times for this job. */
                    CDefaultMixedCriticalityInstanceGenerator.createExecutionTimes(executionTimes, maximumExecutionTimes, jobCriticality, randomizer) ;

                    instance.addJob(new CMixedCriticalityJob("T" + (jobIndex + 1), jobCriticality, executionTimes)) ;
                }

                /* Release times generation. */
                for (jobIndex = 0 ; jobIndex < instance.getJobCount() ; jobIndex++)
                {
                    instance[jobIndex].setReleaseTime(randomizer.Next(releaseTime + 1)) ;
                }

                /* Deadlines generation. */
                for (jobIndex = 0 ; jobIndex < instance.getJobCount() ; jobIndex++)
                {
                    instance[jobIndex].setDeadline(Math.Max((int)(instance[jobIndex].getReleaseTime() + instance[jobIndex].getLargestExecutionTime()), (int)(randomizer.NextDouble() * (instance[jobIndex].getReleaseTime() + deadline + 1)))) ;
                }
                
                instance.setMakeSpan(0.0) ;
                instance.setLateness(0.0) ;

                instance.MakespanDataView = "0" ;
                instance.LatenessDataView = "0" ;

                /* Adds the job to the current instance. */
                instanceList.Add(instance) ;
            }

            return instanceList ;
        }
    }
}
