﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using MCScheduling.GeneticAlgorithm ;
using System.ComponentModel;

namespace MCScheduling.MixedCriticality.GA
{
    /**
     * Evaluates the chromosome representing a schedule for a mixed-criticality    
     * instance according to its total lateness and maximum completion time, i.e.    
     * the schedule's makespan.   
     */   
    public class CMinimizeMakespanFitnessEvaluator : IFitnessEvaluator, IConfigurationDependent
    {
        /** Used to build up a schedule. */
        private double[] m_completionTimeAtCriticality ;
        private double   m_latenessFactor ;
        private double   m_makespanFactor ;

        /**
         * Creates the CMinimizeMakespanFitnessEvaluator with makespan
         * and lateness factors set to 0.5.
         */
        public CMinimizeMakespanFitnessEvaluator() : this(0.5, 0.5)
        {
        }

        /**
         * Creates the CMinimizeMakespanFitnessEvaluator for the specified 
         * lateness, makespan weighting factor.
         * 
         * @param makespanFactor The weight of makespan.
         * @param latenessFactor The weight of lateness.
         */
        public CMinimizeMakespanFitnessEvaluator(double makespanFactor, double latenessFactor)
        {   
            m_latenessFactor = latenessFactor ;
            m_makespanFactor = makespanFactor ;
        }

        /**
         * Returns the makespan factor.
         * 
         * @return The makespan factor.
         */
        public double getMakespanFactor()
        {
            return m_makespanFactor ;
        }

        /**
         * Sets the makespan factor.
         * 
         * @param makespanFactor The new makespan factor value.
         */
        public void setMakespanFactor(double makespanFactor)
        {
            m_makespanFactor = makespanFactor ;
        }

        /**
         * Returns the lateness factor.
         * 
         * @return The lateness factor.
         */
        public double getLatenessFactor()
        {
            return m_latenessFactor ;
        }

        /**
         * Sets the lateness factor.
         * 
         * @param latenessFactor The new lateness factor value.
         */
        public void setLatenessFactor(double latenessFactor)
        {
            m_latenessFactor = latenessFactor ;
        }

         /**
          * Invoked when the genetic algorithm that uses this operator is about 
          * to be executed. 
          * 
          * @param configuration The configuration the operator depends on.
          */
        public void notify(CConfiguration configuration)
        {
            if (configuration is CMixedCriticalityConfiguration)
            {
                m_completionTimeAtCriticality = new double[((CMixedCriticalityConfiguration)configuration).getInstanceToSolve().getMaximumCriticality() + 1] ;
            }
            else
            {
                throw new ArgumentException("This operator may be registered with CMixedCriticalityConfiguration only!") ;
            }
        }

        /**
         * Evaluate the specified individual which represents a schedule of associated
         * mixed-criticality instance. The evaluation is base on the total lateness
         * and makespan of the schedule - the fittest individuals has got the lateness
         * equal to zero and the makespan as small as possible.
         * 
         * @param c The individual - schedule to be evaluated.
         */
        public void evaluate(IChromosome c)
        {
            CMixedCriticalityChromosome chromosome = (CMixedCriticalityChromosome)c ;

            /* The completion time of the current 1 job schedule is actualized. */
            for (int criticality = 1 ; criticality < m_completionTimeAtCriticality.Length ; criticality++)
            {
                m_completionTimeAtCriticality[criticality] = 0.0 ;
            }

            /* The schedule. */
            IGene[] genes = chromosome.getGenes() ;

            /* The lateness of the schedule represented by the evaluated chromosome. */
            double lateness = 0.0 ;

            /* The makespan of the schedule represented by the evaluated chromosome. */
            double makespan = 0.0 ;

            CMixedCriticalityJob job = ((CMixedCriticalityGene)genes[0]).getScheduledJob() ;

            /* The first job starts at its release time. */
            job.setStartTime(job.getReleaseTime()) ;

            /* The completion time of the current 1 job schedule is actualized. */
            for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
            {
                m_completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
            }

            for (int geneIndex = 1 ; geneIndex < genes.Length ; geneIndex++)
            {
                int maximumCompletionTimeCriticality = 1 ;

                /* The next scheduled job. */
                job = ((CMixedCriticalityGene)genes[geneIndex]).getScheduledJob() ;

                /* Determines the start time of the next job. */
                for (int criticality = 2 ; criticality <= job.getCriticality() ; criticality++)
                {
                    if (m_completionTimeAtCriticality[maximumCompletionTimeCriticality] < m_completionTimeAtCriticality[criticality] )
                    {
                        maximumCompletionTimeCriticality = criticality ;
                    }
                }

                job.setStartTime(Math.Max(m_completionTimeAtCriticality[maximumCompletionTimeCriticality], job.getReleaseTime())) ;

                /* Updates the completion times after the next job has been processed. */
                for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
                {
                    m_completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
                }

                /* Each delayed job increases the lateness. */
                lateness += Math.Max(0.0, job.getLargestCompletionTime() - job.getDeadline()) ;

                /* The makespan is also updated with each processed job. */
                if (makespan < job.getLargestCompletionTime())
                {
                    makespan = job.getLargestCompletionTime() ;
                }
            }

            c.setFitnessValue(1.0/(1 + m_latenessFactor * lateness + m_makespanFactor * makespan)) ;  
        }

        /** 
         * Evaluates the population of schedules according to their   
         * total lateness and makespan.   
         *    
         * @param population The population of chromosomes - schedules   
         *                   to be evaluated.   
         *                      
         * @see this.evaluate(IChromosome)   
         */   
        public void evaluate(CPopulation population)
        {
            for (int i = 0 ; i < population.getSize() ; i++)
            {
                evaluate(population[i]) ;
            }
        }
    }
}
