﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using MCScheduling.GeneticAlgorithm ;

namespace MCScheduling.MixedCriticality.GA
{
    /**
     * The mixed-criticality gene represents a place in a schedule
     * for a mixed-criticality job. Which job it is depends on the
     * (integer) value of the gene allele and respective MC instance.
     */
    public class CMixedCriticalityGene : IGene
    {
        /** The associated mixed-criticality instance. */
        private CMixedCriticalityInstance m_instance ;

        /** The position of the job represented by this gene in the associated instance. */
        private object m_jobIndex ;

        /**
         * Creates a new CMixedCriticalityGene that represents a place in schedule
         * that will be occupied by the "jobIndex"-th job in the specified mixed-
         * criticality instance.
         * 
         * @param instance The associated mixed-criticality instance.
         * @param jobIndex The index of the job in the instance.
         */
        public CMixedCriticalityGene(CMixedCriticalityInstance instance, int jobIndex)
        {
            m_instance = instance ;
            m_jobIndex = jobIndex ;
        }

        /**
         * Sets the allele value, i.e. the job in the associated mixed-criticality
         * instance that will occupy the spot in the schedule represented by this
         * gene.
         * 
         * @param value The index of the job in the associated mixed-criticality
         *              instance.
         */
        public void setAllele(object value)
        {
            m_jobIndex = value ;
        }

        /**
         * Returns the allele value, i.e. the index of the job that occupies
         * the spot in the schedule represented by this gene.
         * 
         * @return The index of a job in the associated mixed-criticality
         *         instance.
         */
        public object getAllele()
        {
            return m_jobIndex ;
        }

        /**
         * Returns the mixed-criticality job that has been scheduled in 
         * the spot in the schedule that is represente by this gene.
         * 
         * @return The scheduled mixed-criticality job.
         */
        public CMixedCriticalityJob getScheduledJob()
        {
            return m_instance[(int)m_jobIndex] ;
        }

        /**
         * Returns a copy of this gene.
         * 
         * @return A copy of this gene.
         */
        public object Clone()
        {
            return new CMixedCriticalityGene(m_instance, (int)m_jobIndex) ;
        }
    }
}
