﻿using System.Globalization;
namespace MCScheduling.MixedCriticality.SA
{
    /**
     * The CSimulatedAnnealingConfiguration instance stores a configuration
     * of a simulated annealing solver settings.
     */
    public class CSimulatedAnnealingConfiguration
    {
        /** The makespan factor on the fitness of a solution. */
        private double m_makespanFactor ;

        /** The lateness factor on the fitness of a solution. */
        private double m_latenessFactor ;

        /** The factor of temperature decrease per iteration. */
        private double m_coolingFactor ;

        /** The maximum temperature at which the annealing process starts. */
        private double m_maximumTemperature ;

        /** The minimum temperature at which the annealing process stops. */
        private double m_minimumTemperature ;

        /**
         * Creates a new CSimulatedAnnealingConfiguration.
         */
        public CSimulatedAnnealingConfiguration()
        {
        }

        /**
         * Creates a CSimulatedAnnealingConfiguration from the data
         * in the specified file.
         * 
         * @param filePath The location of the configuration file.
         */
        public CSimulatedAnnealingConfiguration(string filePath)
        {
            deserialize(filePath) ;
        }

        /**
         * Returns the makespan factor.
         * 
         * @return The makespan factor.
         */
        public double getMakespanFactor()
        {
            return m_makespanFactor ;
        }

        /**
         * Sets the makespan factor.
         * 
         * @param makespanFactor The new makespan factor value.
         */
        public void setMakespanFactor(double makespanFactor)
        {
            m_makespanFactor = makespanFactor ;
        }

        /**
         * Returns the lateness factor.
         * 
         * @return The lateness factor.
         */
        public double getLatenessFactor()
        {
            return m_latenessFactor ;
        }

        /**
         * Sets the lateness factor.
         * 
         * @param latenessFactor The new lateness factor value.
         */
        public void setLatenessFactor(double latenessFactor)
        {
            m_latenessFactor = latenessFactor ;
        }

        /**
         * Returns the cooling factor.
         * 
         * @return The cooling factor.
         */
        public double getCoolingFactor()
        {
            return m_coolingFactor ;
        }

        /**
         * Sets the cooling factor.
         * 
         * @param coolingFactor The new cooling factor value.
         */
        public void setCoolingFactor(double coolingFactor)
        {
            m_coolingFactor = coolingFactor ;
        }

        /**
         * Sets the maximum temperature.
         * 
         * @param maximumTemperature The new maximum temperature value.
         */
        public void setMaximumTemperature(double maximumTemperature)
        {
            m_maximumTemperature = maximumTemperature ;
        }

        /**
         * Returns the maximum temperature.
         * 
         * @return The maximum temperature.
         */
        public double getMaximumTemperature()
        {
            return m_maximumTemperature ;
        }

        /**
         * Sets the minimum temperature.
         * 
         * @param minimumTemperature The new minimum temperature value.
         */
        public void setMinimumTemperature(double minimumTemperature)
        {
            m_minimumTemperature = minimumTemperature ;
        }

        /**
         * Returns the minimum temperature.
         * 
         * @return The minimum temperature.
         */
        public double getMinimumTemperature()
        {
            return m_minimumTemperature ;
        } 

        #region >>> Serialization Methods <<<

        /**
         * Serializes this configuration into the specified
         * file.
         * 
         * @param filePath The path to a file where this configuration
         *                 is going to be serialized.
         */ 
        public void serialize(string filePath)
        {
            using (System.IO.StreamWriter fileWriter = new System.IO.StreamWriter(filePath))
            {
                fileWriter.WriteLine(getMakespanFactor().ToString(CultureInfo.InvariantCulture)) ;
                fileWriter.WriteLine(getLatenessFactor().ToString(CultureInfo.InvariantCulture)) ;
                fileWriter.WriteLine(getCoolingFactor().ToString(CultureInfo.InvariantCulture)) ;
                fileWriter.WriteLine(getMinimumTemperature().ToString(CultureInfo.InvariantCulture)) ;
                fileWriter.WriteLine(getMaximumTemperature().ToString(CultureInfo.InvariantCulture)) ;
            }
        }

        /**
         * Deserializes the configuration in the specified 
         * file and sets the fields of this configuration
         * according to it.
         * 
         * @param filePath The path where the configuration 
         *                 is going to be deserialized from.
         */
        public void deserialize(string filePath)
        {
            using (System.IO.StreamReader fileReader = new System.IO.StreamReader(filePath))
            {
                setMakespanFactor(double.Parse(fileReader.ReadLine().Trim(), CultureInfo.InvariantCulture)) ;
                setLatenessFactor(double.Parse(fileReader.ReadLine().Trim(), CultureInfo.InvariantCulture)) ;
                setCoolingFactor(double.Parse(fileReader.ReadLine().Trim(), CultureInfo.InvariantCulture)) ;
                setMinimumTemperature(double.Parse(fileReader.ReadLine().Trim(), CultureInfo.InvariantCulture)) ;
                setMaximumTemperature(double.Parse(fileReader.ReadLine().Trim(), CultureInfo.InvariantCulture)) ;
            }
        }

        #endregion >>> Serialization Methods <<<
    }
}
