﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

/**
 * Contains the implementation of a mixed-criticality solver based on the simulated
 * annealing algorithm.
 */
namespace MCScheduling.MixedCriticality.SA
{
    /**
     * The mixed-criticality scheduling problem solver based on the simulated 
     * annealing meta-heuristic method.
     * 
     * @author Tomas Tunys
     */
    class CSimulatedAnnealingSolver : CBaseMixedCriticalitySolver
    {
        /** The simulated annealing algorithm performing the optimization. */
        CMixedCriticalitySimulatedAnnealing m_algorithm ;

        /**
         * Creates a new CSimulatedAnnealingSolver instance.
         */
        public CSimulatedAnnealingSolver()
        {
            m_algorithm = new CMixedCriticalitySimulatedAnnealing() ;
        }

        /**
         * Creates a new CSimulatedAnnealingSolver instance with the specified
         * configuration of the simulated annealing algorithm.
         * 
         * @param configuration The configuration of the simulated annealing
         *                      algorithm.
         */
        public CSimulatedAnnealingSolver(CSimulatedAnnealingConfiguration configuration)
        {
            m_algorithm = new CMixedCriticalitySimulatedAnnealing() ;

            setConfiguration(configuration) ;
        }

        /**
         * Sets the configuration for the simulated annealing algorithm.
         * 
         * @param configuration The configuration of the simulated annealing
         *                      algorithm.
         */
        public void setConfiguration(CSimulatedAnnealingConfiguration configuration)
        {
            m_algorithm.setConfiguration(configuration) ;
        }

        /**
         * Returns the configuration of the simulated annealing algorithm
         * that performs the optimization.
         * 
         * @return The configuration of the simulated annealing algorithm
         * that performs the optimization.
         */
        public CSimulatedAnnealingConfiguration getConfiguration()
        {
            return m_algorithm.getConfiguration() ;
        }

        /**
         * Solves the specified mixed-criticality instance using simulated
         * annealing.
         * 
         * @param instance The mixed-criticality instance to solve.
         */
        public override void solve(CMixedCriticalityInstance instance)
        {
            m_algorithm.setInstanceToSolve(instance) ;

            /* Time limit in miliseconds. */
            long timeLimit = (hasTimeLimit() ? getTimeLimit() * 1000 : long.MaxValue) ;
            
            /* Iteration limit. */
            long iterationLimit = (hasIterationLimit() ? getIterationLimit() : long.MaxValue) ;

            reconstructSolution(m_algorithm.solve(iterationLimit, timeLimit), instance) ;
        }

        /**
         * Sets the parameters of the solution represented by the specified scheduledJobIndices up
         * in the specified mixed-criticality instance.
         * 
         * @param state    The simulated annealing state representing a solution to the 
         *                 specified mixed-criticality instance.
         * @param instance The instance to be set up according to the solution represented 
         *                 by the specified simulated annealing state.
         */
        private void reconstructSolution(CSimulatedAnnealingState state, CMixedCriticalityInstance instance)
        {
            /* The lateness of the schedule represented by the evaluated chromosome. */
            double lateness = 0.0 ;

            /* The makespan of the schedule represented by the evaluated chromosome. */
            double makespan = 0.0 ;

            double[] completionTimeAtCriticality = new double[instance.getMaximumCriticality() + 1] ;

            CMixedCriticalityJob job = instance[state.ScheduledJobIndex[0]] ;

            /* The first job starts at its release time. */
            job.setStartTime(job.getReleaseTime()) ;

            /* The completion time of the current 1 job schedule is actualized. */
            for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
            {
                completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
            }

            for (int i = 1 ; i < state.Length ; i++)
            {
                int maximumCompletionTimeCriticality = 1 ;

                /* The next scheduled job. */
                job = instance[state.ScheduledJobIndex[i]] ;

                /* Determines the start time of the next job. */
                for (int criticality = 2 ; criticality <= job.getCriticality() ; criticality++)
                {
                    if (completionTimeAtCriticality[maximumCompletionTimeCriticality] < completionTimeAtCriticality[criticality] )
                    {
                        maximumCompletionTimeCriticality = criticality ;
                    }
                }

                job.setStartTime(Math.Max(completionTimeAtCriticality[maximumCompletionTimeCriticality], job.getReleaseTime())) ;

                /* Updates the completion times after the next job has been processed. */
                for (int criticality = 1 ; criticality <= job.getCriticality() ; criticality++)
                {
                    completionTimeAtCriticality[criticality] = job.getCompletionTime(criticality) ;
                }

                /* Each delayed job increases the lateness. */
                lateness += Math.Max(0.0, job.getLargestCompletionTime() - job.getDeadline()) ;

                /* The makespan is also updated with each processed job. */
                if (makespan < job.getLargestCompletionTime())
                {
                    makespan = job.getLargestCompletionTime() ;
                }
            }

            /* Sets the makespan. */
            instance.setMakeSpan(makespan) ;

            /* Sets the lateness. */
            instance.setLateness(lateness) ;

            /* Feasible solution found. */
            if (lateness < 0.00001)
            {
                setStatus(Status.SUBOPTIMAL) ;
            }
            else
            {
                setStatus(Status.HEURISTIC) ;
            }
        }

        /**
         * Interrupts the genetic algorithm solver.
         */
        public override void interrupt()
        {
            m_algorithm.setInterrupted(true) ;
        }
    }
}
