﻿/* Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
 * All rights reserved.                          
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. The names of its contributors may not be used to endorse or promote 
 *      products derived from this software without specific prior written 
 *      permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Contains two pseudo-random number generator implementations; one with
 * uniform distribution (based on famous Mersenne Twister) and the other 
 * one with exponential distribution.
 */
namespace MCScheduling.Utils
{
    /**
     * The (uniform random number generator) Mersenne Twister is a pseudorandom 
     * number generator developed in 1997 by Makoto Matsumoto and Takuji Nishimura.
     */
    public class CUniformRandom
    {
        private const int N = 624 ;
        private const int M = 397 ;

        private const ulong MATRIX_A = 0x9908b0dfUL ;

        private const ulong UPPER_MASK = 0x80000000UL ;
        private const ulong LOWER_MASK = 0x7fffffffUL ;

        private ulong[] mt = new ulong[N] ;

        private int mti = N + 1 ;

        private readonly ulong[] mag01 = new ulong[] { 0x0UL, MATRIX_A } ;
        
        /**
         * Creates the Mersenne Twister random number generator,
         * and initializes it with a default seed.
         */
        public CUniformRandom()
        {
            initialize(new ulong[]{ 0x123, 0x234, 0x345, 0x456 }) ;
        }

        /**
         * Creates and initializes the Mersenne Twister by the specified
         * seed value.
         * 
         * @param seed A number used to calculate a starting value for 
         *             the pseudo-random number sequence.
         */
        public CUniformRandom(ulong seed) 
        {
            initialize(seed) ;
        }

        /**
         * Creates and initializes the Mersenne Twister by the specified
         * seed value.
         * 
         * @param seed An array whose values are used to calculate 
         *             a starting value for the pseudo-random number
         *             sequence.
         */
        protected CUniformRandom(ulong[] seed) 
        {
            initialize(seed) ; 
        }
        
        /**
         * Initializes the Mersenne Twister by the specified seed value.
         * 
         * @param seed A number used to calculate a starting value for 
         *             the pseudo-random number sequence.
         */
        private void initialize(ulong seed)
        {
	        mt[0] = seed & 0xffffffffUL ;
            System.Random r= new System.Random(12) ;

	        for (mti = 1 ; mti < N ; mti++) 
            {
		        mt[mti] = (1812433253UL * (mt[mti - 1] ^ (mt[mti - 1] >> 30)) + (ulong)mti) & 0xffffffffUL ;
	        }
        }

        /**
         * Initializes the Mersenne Twister by the specified array values.
         * 
         * @param seed An array whose values are used to calculate 
         *             a starting value for the pseudo-random number
         *             sequence.
         */
        private void initialize(ulong[] seed) 
        {
	        int i = 1 ;
	        int j = 0 ;
	        int k = (N > seed.Length ? N : seed.Length) ;
	        
            for ( ; k != 0 ; k--) 
            {
		        mt[i] = ((mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1664525UL)) + seed[j] + (ulong)j) & 0xffffffffUL ;
		        
		        if (++i >= N) 
                {
                    mt[0] = mt[N - 1] ;
                    i     = 1 ;
                }

		        if (++j >= seed.Length)
                { 
                    j = 0 ; 
                }
	        }
	        
            for (k = N - 1 ; k != 0 ; k--) 
            {
		        mt[i] = ((mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1566083941UL)) - (ulong)i) & 0xffffffffUL ;
		        
		        if (++i >= N)
                {
                    mt[0] = mt[N - 1] ;
                    i     = 1 ;
                }
	        }

	        mt[0] = 0x80000000UL ;
        }

        /**
         * Generates a random non-negative 32-bit integer.
         * 
         * @return A random integer in the [0,2^32) interval.
         */
        private ulong Generate() 
        {
	        ulong y = 0 ;
	        
	        if (mti >= N) 
            {
		        if (mti == N + 1) 
                {
			        initialize(5489UL) ;
		        }

                int kk = 0 ;

		        for ( ; kk < N - M ; kk++) 
                {
			        y = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK) ;
			        mt[kk] = mt[kk + M] ^ (y >> 1) ^ mag01[y & 0x1UL] ;
		        }

		        for ( ; kk < N - 1 ; kk++) 
                {
			        y = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK) ;
			        mt[kk] = mt[kk + (M - N)] ^ (y >> 1) ^ mag01[y & 0x1UL] ;
		        }

		        y = (mt[N - 1] & UPPER_MASK) | (mt[0] & LOWER_MASK) ;
		        mt[N - 1] = mt[M - 1] ^ (y >> 1) ^ mag01[y & 0x1UL] ;

		        mti = 0 ;
	        }

	        y = mt[mti++] ;

	        y ^= (y >> 11) ;
	        y ^= (y <<  7) & 0x9d2c5680UL ;
	        y ^= (y << 15) & 0xefc60000UL ;
	        y ^= (y >> 18) ;

	        return y ;
        }

        /**
         * Generates a random non-negative 32-bit integer.
         * 
         * @return A random integer in the [0,2^32) interval.
         */
        public int Next()
        {
            return (int)(NextDouble() * ((ulong)int.MaxValue + 1)) ;
        }

        /**
         * Returns a random non-negative integer lesser then the specified value.
         * 
         * @param maxValue The exclusive upper bound of the generated random integer.
         * 
         * @return A random integer from the [0,maxValue) interval.
         */
        public int Next(int maxValue)
        {
	        return (int)(NextDouble() * maxValue) ;
        }

        /**
         * Returns a random integer bigger or equal to the specified minValue, and
         * lesser then the specified maxValue.
         * 
         * @param minValue The inclusive lower bound of the generated random integer.
         * @param maxValue The exclusive upper bound of the generated random integer.
         * 
         * @return A random integer from the [minValue,maxValue) interval.
         */
        public int Next(int minValue, int maxValue)
        {
	        return (Next(maxValue - minValue) + minValue) ;
        }

        /**
         * Returns a random decimal number in the [0,1) interval.
         * 
         * @return A random decimal number in the [0,1) interval.
         */
        public double NextDouble()
        {
	        return ((double)Generate() / ((ulong)uint.MaxValue + 1)) ;
        }
    }
}